% FUNCTION: [Xsol, Wsol, fopt, output] = mat_QUIC(Shat, rho, X0, param)
%
% PURPOSE: A MATLAB implementation of the QUIC algorithm
%          This is simply a transformation of C-code in QUIC to MATLAB code.
%
% INFORMATION:
%   By Quoc Tran Dinh, LIONS, EPFL, Lausanne, Switzerland.
%   Date 05.07.2012.
%
%

function [Xsol, Wsol, fopt, output] = mat_QUIC(Shat, rho, X0, param)

% Check the input parameters.
if nargin < 4
    param.MaxIters    = 1000;
    param.BtkMaxIters = 20;
    param.RelTolX     = 1e-6;
    param.RelTolFun   = 2.22e-16;
    param.bcdsSubTol  = 0.05;
    param.isTrace     = 0;
end
if isempty(param)
    param.MaxIters    = 1000;
    param.BtkMaxIters = 20;
    param.RelTolX     = 1e-6;
    param.RelTolFun   = 2.22e-16;
    param.bcdsSubTol  = 0.05;
    param.isTrace     = 0;
end

if nargin < 3
    X0 = diag(1./diag(Shat));
end
if isempty(X0)
    X0 = diag(1./diag(Shat));
end
if isempty(rho) || rho < 0 || isinf(rho) || isnan(rho)
    error('The regularization parameter is increct!');
end

% Make the matrix of Lambda.
p       = size(Shat);
Lambda0 = rho*ones(p);

% Call the solver.
[Xsol, Wsol, fopt, output] = QUIC_solver(Shat, Lambda0, X0, param);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [X, W, opt_fx, output] = QUIC_solver(Shat, Lambda0, X0, param)
%%% PURPOSE:  Reimplementation of QUIC in MATLAB.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Xsol, Wsol, opt_fx, output] = QUIC_solver(Shat, Lambda0, X0, param)

% Check the input parameters.
if size(Shat,1) ~= size(Shat, 2)
    error('Size of the input matrix is not consistant!');
end
if isempty(Lambda0)
    error('The reguralize parameter is empty!');
end
if nargin < 3
    error('The initial point is empty!');
end

if isempty(X0)
    error('The initial point is empty!');
end

% Other parameters.
SIGMA = 0.001;

% Problem size.
p = size(Shat, 1);

% Check the regularization parameter.
if all(all(size(Lambda0) ~= [p, p]))
    lambda = Lambda0(1, 1);
    Lambda = lambda*ones(p, p);
else
    Lambda = Lambda0;
end

% Initialize the objective value terms.
fX              = inf;
fX1             = inf;
fXprev          = inf;
output.time     = tic;
output.msg      = 'Start!';
output.chol_cnt = 0;
if param.isTrace
    output.info.fx   = [];
    output.info.sit  = [];
    output.info.serr = [];
    output.info.step = [];
else
    output.info     = [];
end

% Initialize other parameters.
l1normX = 0.0;
trSX    = 0.0;
logDetX = 0.0;
D_cur   = zeros(p, p);
U_cur   = zeros(p, p);

X_cur   = X0;
[L_cur, cholfailed3] = chol(X_cur);
if cholfailed3 ~= 0
    error('Cholesky factorization is failed!');
end
output.chol_cnt = output.chol_cnt + 1;
Ip       = eye(size(Shat));
invL_cur = L_cur\Ip;
W_cur    = invL_cur*invL_cur';

% Compute the norm(Lambda.*X, 1) and trace(Shat*X.
l1normX = sum(sum(Lambda.*abs(X_cur)));
trSX    = sum(sum(Shat.*X_cur));
cdSweep = 0;

% The main iteration loop.
for NewtonIter = 1:param.MaxIters
    
    % Initialization.
    normD   = 0.0;
    diffD   = 0.0;
    subgrad = inf;
    
    %---------------------------------------------------------------------
    % Step 1: Compute the search direction.
    %---------------------------------------------------------------------
    
    % The first iteration.
    if NewtonIter == 1 && ~(nnz(X_cur) - nnz(diag(X_cur))) %isdiagonal(X)
        
        if param.PrintLevel >= 2
            %fprintf('Newton iteration 1.\n');
            fprintf(' + X is a diagonal matrix.\n');
        end
        D_cur       = zeros(p, p);
        [fX, D_cur] = diagonalNewton(Shat, Lambda, X_cur, W_cur, D_cur);
        
        % Fill in the full matrix.
        D_cur       = tril(D_cur) + tril(D_cur,-1)';
        
    % Other iterations.
    else
        
        % Initialize 
        numActive = 0;
        U_cur     = zeros(p,p);
        D_cur     = zeros(p,p);
        subgrad   = 0.0;
        
        % Compute the active set.
        for i = 1:p
            for j = 1:i
                
                g = Shat(i,j) - W_cur(i,j);
                
                if X_cur(i,j) ~= 0.0 || abs(g) > Lambda(i,j)
                    
                    activeSet(numActive+1, 1).i = i;
                    activeSet(numActive+1, 1).j = j;
                    numActive                   = numActive + 1;
                    
                    if X_cur(i,j) > 0.0
                        g = g + Lambda(i,j);
                    elseif X_cur(i,j) < 0.0
                        g = g - Lambda(i,j);
                    else
                        g = abs(g) - Lambda(i,j);
                    end
                    
                    subgrad = subgrad + abs(g);
                end
            end
        end
        
        % Print the messages.
        if param.PrintLevel >= 2
            fprintf('Newton iteration %d.\n', NewtonIter);
            fprintf(' + Active set size %d.\n', numActive);
            fprintf(' + Sub-gradident = %6.10f, l1-norm of X = %6.10f.\n', ...
                     subgrad, l1normX);
        end
        
        % Solve the subproblem by using coordinate descent method.
        for cdSweep = 1:round(NewtonIter/3) + 2
            
            % Initialize diffD.
            diffD = 0.0;
            
            % Change the active set by some random exchanges.
            for i = 1:numActive
                
                % Generate a random coordinate.
                j = i + randi([0, max(numActive - i-1, 0)], 1);
                
                % Exchange the coordinate.
                k1               = activeSet(i,1).i;
                k2               = activeSet(i,1).j;
                activeSet(i,1).i = activeSet(j,1).i;
                activeSet(i,1).j = activeSet(j,1).j;
                activeSet(j,1).i = k1;
                activeSet(j,1).j = k2;
            end
            
            % For each coordinate, perform the coordinate descent step.
            for l = 1:numActive
                i = activeSet(l,1).i;
                j = activeSet(l,1).j;
                [U_cur, D_cur, normD, diffD] = coordinateDescentUpdate(...
                                               Shat, Lambda, X_cur, ...
                                               W_cur, U_cur, D_cur, ...
                                               i, j, normD, diffD);
            end
            
            % Print the message in the coordinate descent algorithm.
            if param.PrintLevel >= 3
                fprintf([' + Coordinate descent sweep %d. Norm of ', ...
                         'D = %6.10f, change in D = %6.10f'], cdSweep, ...
                          normD, diffD);
            end
            
            % Terminate condition.
            if diffD <= normD*param.bcdsSubTol
                break;
            end
            
        end
        % End of the coordinate descent method.
        
    end
    % End of the if statement for computing the search direction.   
    
    % Fill in the full metrix.
    U_cur     = tril(U_cur) + tril(U_cur,-1)';
	D_cur     = tril(D_cur) + tril(D_cur,-1)';

    %---------------------------------------------------------------------
    % Step 2: Evaluate the objective value.
    %---------------------------------------------------------------------
    if(fX == inf)
        
        % Compute the Cholesky factorization.
        [U_cur, cholfailed1] = chol(X_cur, 'upper');
        if cholfailed1 ~= 0
            error('Cholesky factorization is failed!');
        end
        output.chol_cnt = output.chol_cnt + 1;
        
        % Evaluate the objective value.
        logDetX = 2*sum(log(diag(U_cur)));
        fX      = (trSX + l1normX) - logDetX;
    end

    % Evaluate the direction gradient vector.
    trgradgD = sum(sum((Shat - W_cur).*D_cur)); 

    %---------------------------------------------------------------------
    % Step 3: Performing line-search procedure.
    %---------------------------------------------------------------------
    alpha    = 1;
    l1normXD = 0.0;
    fX1prev  = inf;

    % The main loop of line-search.
    for lsiter = 1:param.BtkMaxIters
        
        % Update the next trial iteration.
        X_next   = X_cur + alpha*D_cur;
        
        % Evaluate the l1-norm and trace.
        l1normX1 = sum(sum(Lambda.*abs(X_next)));
        trSX1    = sum(sum(Shat.*X_next));
        
        % Perform the Cholesky factorization of the trial point.
        [W_cur, cholfailed2] = chol(X_next, 'upper');
        output.chol_cnt = output.chol_cnt + 1;
        
        % If cholesky failed then decrease the step size and move on.
        if cholfailed2 ~= 0
            if param.PrintLevel >= 2
                fprintf([' + Line-search step size = %6.8f. Lack of ', ...
                         'positive definiteness.\n'], alpha);
            end
            alpha = 0.5*alpha;
            continue;
        end

        % Evaluate the objective value.
        logDetX1 = 2*sum(log(diag(W_cur)));
        fX1      = (trSX1 + l1normX1) - logDetX1;
        
        % If the full-step size is reached.
        if (alpha == 1.0 )
            l1normXD = l1normX1;
        end

        % Check the descent direction.
        if( fX1 <= fX + alpha*SIGMA*(trgradgD + l1normXD - l1normX) || ...
            normD == 0.0 )
            
            if param.PrintLevel >= 2
                fprintf(' + Line-search step size chosen: %6.8f\n', alpha);
            end
            
            fXprev  = fX;
            fX      = fX1;
            l1normX = l1normX1;
            logDetX = logDetX1;
            trSX    = trSX1;
            
            break;
        end

        if param.PrintLevel >= 2
            fprintf(' + Line-search step size: %6.8f\n', alpha);
            fprintf(['+ The objective value would not decrease ', ...
                     'sufficiently: %6.8f\n'], fX1 - fX);
        end

        % If the objective value is improved, then take this step.
        if fX1prev < fX1
            fXprev  = fX;
            l1normX = l1normX1;
            logDetX = logDetX1;
            trSX    = trSX1;
            
            break;
        end

        % Keep the objective value and move on.
        fX1prev = fX1;
        alpha   = 0.5*alpha;

    end
    % End of the line search.

    if param.PrintLevel >= 2
        fprintf(' + Objective value decreased by: %7.8f\n', fXprev - fX);
    end

    % Compute W = inv(X).
    [L_cur, cholfailed3] = chol(X_next, 'upper');
    output.chol_cnt = output.chol_cnt + 1;
    if cholfailed3 ~= 0
        fprintf(' + Line search iterations: %d\n', lsiter);
        error('Cholesky factorization is failed!');
    end
    Ip       = eye(size(Shat));
    invL_cur = L_cur\Ip;
    W_cur    = invL_cur*invL_cur';
    
    
    % Update X for the next iteration.
    X_cur = X_cur + alpha*D_cur;
    
    % Print the iteration.
    fprintf('-----------------------------------------------------\n');
    fprintf('Iter = %d, normD = %6.8f, fX = %6.9f\n', NewtonIter, normD, fX);
    fprintf('-----------------------------------------------------\n');
    
    % Save the information if required.
    if param.isTrace
        output.info.fx   = [output.info.fx; fX];
        output.info.sit  = [output.info.sit; cdSweep];
        output.info.serr = [output.info.serr; diffD/max(1, normD)];
        output.info.step = [output.info.step; alpha];
    end
    
    % Terminate if the time exceed the required time
    total_time = toc(output.time);
    if total_time >= 60*60*5
        output.msg = 'Exceed the required time';
        break;
    end
    
    % Check the stopping criterion.
    if (subgrad*alpha >= l1normX*param.RelTolX && ...
       (abs(fX - fXprev)/abs(fX) >= param.RelTolFun) )
        continue;
    end

    % Terminate if the stopping critterion is satisfied.
    output.msg = 'Convergence acheived!';
    break;
end
% End of the main loop.

% Exceed the maximum number of iterations.
if NewtonIter >= param.MaxIters
    output.msg = 'Exceed the maximum iterations!';
end

% Get the final solution and other information.
Xsol         = X_cur;
Wsol         = W_cur;
opt_fx       = fX;
logDetW      = projLogDet(Shat, W_cur, Lambda);

output.dGap  = -logDetW - p - logDetX + trSX + l1normX;
output.iter  = NewtonIter;
output.time  = toc(output.time);
output.error = subgrad*alpha;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [U, D, normD, diffD] = coordinateDescentUpdate(Shat, ...
%%%                                  Lambda, X, W, U, D, i, j, normD, diffD)
%%%
%%% PURPOSE: Update one coordinate at each iteration.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [U_out, D_out, normD_out, diffD_out] = coordinateDescentUpdate(...
                                                Shat, Lambda, X_cur, ...
                                                W_cur, U_cur, D_cur, ...
                                                i, j, normD, diffD)
% Compute the quantity a.
a = W_cur(i, j)*W_cur(i,j);
if(i ~= j)
   a = a + W_cur(i,i)*W_cur(j,j); 
end

% Compute the quantity b.
inva = 1.0/a;
b    = Shat(i,j) - W_cur(i,j);
b    = b + sum( W_cur(i, :)*U_cur(:, j) );


l     = Lambda(i,j)*inva;
c     = X_cur(i,j) + D_cur(i,j);
f     = b*inva;

normD = normD - abs(D_cur(i,j));

if( c > f )
    mu = -f - l;
    if (c + mu < 0.0)
        mu = -c;
        D_cur(i,j) = -X_cur(i,j);
    else
        D_cur(i,j) = D_cur(i,j) + mu;
    end
else
    mu = -f + l;
    if (c + mu > 0.0)
        mu = -c;
        D_cur(i,j) = -X_cur(i,j);
    else
        D_cur(i,j) = D_cur(i,j) + mu;
    end
end

diffD = diffD + abs(mu);
normD = normD + abs(D_cur(i,j));

if (mu ~= 0.0)
    U_cur(i,:) = U_cur(i,:)  + mu*W_cur(j,:);
    
    if (i ~= j)
        U_cur(j,:) = U_cur(j,:)  + mu*W_cur(i,:);
    end
end

% Assign the output.
diffD_out = diffD;
normD_out = normD;
U_out     = U_cur;
D_out     = D_cur;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [fX, D_out] = diagonalNewton(Shat, Lambda, X_cur, W_cur, ...
%%%                                        D_cur)
%%% PURPOSE: Compute the diagonal proximal Newton direction.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [fX, D_out] = diagonalNewton(Shat, Lambda, X_cur, W_cur, D_cur)

p = size(Shat, 1);

for i = 1:p
    for j = 1:i-1
        a    = W_cur(i,i)*W_cur(j,j);
        inva = 1.0/a;
        b    = Shat(i,j);

        l    = Lambda(i,j)*inva;
        f    = b*inva;
        x    = -b*inva;
        
        if (f < 0 )
            mu = -f - l;
            x  = x - l;
            
            if( mu < 0.0 )
                mu         = 0.0;
                D_cur(i,j) = -X_cur(i,j);
            else
                D_cur(i,j) = D_cur(i,j) + mu;
            end
        else
            mu = -f + l;
        
            if (mu > 0.0)
                mu         = 0.0;
                D_cur(i,j) = -X_cur(i,j);
            else
                D_cur(i,j) = D_cur(i,j) + mu;
            end
            
        end
    end
end

logDet  = 0.0;
l1normX = 0.0;
trSX    = 0.0;

for i = 1:p
    logDet  = logDet  + log(X_cur(i,i));
    l1normX = l1normX + Lambda(i,i)*abs(X_cur(i,i));
    trSX    = trSX    + Shat(i,i)*X_cur(i,i);
    
    a       = W_cur(i,i)*W_cur(i,i);
    inva    = 1.0/a;
    b       = Shat(i,i) - W_cur(i,i);
    l       = Lambda(i,i)*inva;
    c       = X_cur(i,i);
    f       = b*inva;
    
    if(c > f)
        mu = -f - l;
        if( c + mu < 0.0 )
            D_cur(i,i) = -X_cur(i,i);
            continue;
        end
    else
        mu = -f + l;
        if( c + mu > 0.0 )
            D_cur(i,i) = -X_cur(i,i);
            continue;
        end
    end
    D_cur(i,i) = D_cur(i,i) + mu;
end

% Evaluate the objective value.
fX    = -logDet + trSX + l1normX;
D_out = D_cur; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: logDet = projLogDet(Shat, W, Lambda)
%%% PURPOSE: Compute projection.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function logDet = projLogDet(Shat, W, Lambda)

prW = min( max(Shat - Lambda, W), Shat + Lambda);

[prW, cholfailed] = chol(prW, 'upper');
if cholfailed ~= 0, logDet = inf; return; end

logDet = 2*sum(log(diag(prW)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%







