% PURPOSE: This function tests the scopt_glearn_ls_pna solver with real data.
%          The data can be downloaded from
%                http://ima.umn.edu/~maxxa007/send_SICS/.
% NOTE:    
%   scopt_glearn_pga is an implementation of Proximal-Newton algorithm
%   for solving the following problem:
%           min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%           s.t. X > 0.
%

%%%%%%%%%%%%%%%%%%%%% Begin: Lympth data %%%%%%%%%%%%%%%%%%%%%%%%
% Input data.
rho   = 0.25;          % Regularization parameter.
pname = 'Lymph.mat';  % Data file.

% Load the data.
dataDir = which('scopt_glearn_pga_exam.m');
dataDir = [dataDir(1:end-23), 'glearn_data/'];
if exist([dataDir, pname])
    data  = load([dataDir, pname]);
    S_hat = data.S;
    p     = size(S_hat, 1);
else
    error('Cannot load the input data');
end
%%%%%%%%%%%%%%%%%%%%% End: Lympth data %%%%%%%%%%%%%%%%%%%%%%%%

% Default parameters.
param = scopt_optimset([]);

% User defined parameters.
param.AbsTol      = 1e-8;   % Absolute tolerance.
param.RelTol      = 1e-8;   % Relative tolerance.
param.MaxIters    = 200;    % Maximum number of iterations.
param.SubMaxIters = 50;     % Maximum number of iterations in subsolver.
param.SubRelTol   = 1e-9;   % Tolerance for terminating subsolver.
param.isFxEval    = true;   % Require to evaluate the objective value.
param.numFxCheck  = 4;      % Terminate if the objective value does not 
                            % change significantly after 4 successive
                            % iterations.
param.isTrace     = 0;      % Do not keep trace.
param.BtkMaxIters = 200;    % Maximum number of line-search iterations.
param.PrintLevel  = 2;      % Print 50 iterations for each line.

% Choose the method for solving the subproblem (dual-FISTA).
param.SubMethod   = 'dFISTA';

% Print problem information.
fprintf('+ Problem name: %s\n', pname);
fprintf('+ Problem size: %d\n', p);

% Initial point.
X0 = diag(1./diag(S_hat));

% Call the solver with dual-FISTA and without line-search.
param.isBtkLS     = true;
param.SubMethod   = 'dFISTA';
fprintf(' CALL THE SOLVER (dual-FISTA and without LS)...\n');
[fsol_dfista, output_dfista] = scopt_glearn_ls_pna(S_hat, rho, param);

% Parameter for QUIC.
param.BtkMaxIters = 100;
param.RelTolFun   = 2.22e-16;

% Call the QUIC solver 
[Xsol_q, Wsol, fopt, output] = mat_QUIC(S_hat, rho, X0, param);

% Recalculate the objective value.
Xsol        = fsol_dfista.x;
fval_dfista   = -sum(log(eig(Xsol))) + trace(S_hat*Xsol) + rho*norm(Xsol(:), 1);
fval_quic   = -sum(log(eig(Xsol_q))) + trace(S_hat*Xsol_q) + rho*norm(Xsol_q(:), 1);
fprintf(' + The objective value of SCOPT: %6.16f\n', fval_dfista);
fprintf(' + The objective value of QUIC:  %6.16f\n', fval_quic);
fprintf(' + Number of iteration of SCOPT: %5d.\n', output_dfista.iter);
fprintf(' + Number of iteration of QUIC:  %5d.\n', output.iter);
fprintf(' + CPU time in QUIC:  %3.6f.\n', output.time);
fprintf(' + CPU time in SCOPT: %3.6f.\n', output_dfista.time);
fprintf(' + Compare the solution: norm(X_quic-X_scopt, "fro") = %6.8f.\n\n', ...
        norm(fsol_dfista.x - Xsol_q, 'fro'));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%