% PURPOSE: This function tests the scopt_glearn_fpna solver with real data.
%          The data can be downloaded from
%                http://ima.umn.edu/~maxxa007/send_SICS/.
% NOTE:    
%   scopt_glearn_pga is an implementation of Path-following Proximal-Newton 
%   algorithm for solving the following problem:
%           min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%           s.t. X > 0.
%

% Input data.
rho_start = 0.25;  % Run from rho = 0.25 to rho = 0.2.
rho_end   = 0.20;
pname     = 'Lymph.mat'; 

% Load the data.
dataDir = which('scopt_glearn_pga_exam.m');
dataDir = [dataDir(1:end-23), 'glearn_data/'];
if exist([dataDir, pname])
    data  = load([dataDir, pname]);
    S_hat = data.S;
    p     = size(S_hat, 1);
else
    error('Cannot load the input data');
end

% Default parameters.
param = scopt_optimset([]);

% User defined parameters.
param.AbsTol      = 1e-8;   % Absolute tolerance.
param.RelTol      = 1e-8;   % Relative tolerance.
param.MaxItersP1  = 1000;   % Maximum number of iterations in Phase 1.       
param.MaxItersP2  = 3000;   % Maximum number of iterations in Phase 2.
param.SubMaxIters = 100;    % Maximum number of iterations in subsolver.
param.SubRelTol   = 1e-9;   % Tolerance for terminating subsolver.
param.isFxEval    = true;   % Require to evaluate the objective value.
param.numFxCheck  = 4;      % Terminate if the objective value does not 
                            % change significantly after 4 successive
                            % iterations.
param.isTrace     = 0;      % Do not keep trace.
param.BtkMaxIters = 20;     % Maximum number of line-search iterations.
param.PrintLevel  = 2;      % Print 50 iterations for each line.

% Choose the method for solving the subproblem (dual-FISTA).
param.SubMethod   = 'dFISTA';

% Problem information.
fprintf('+ Problem name: %s\n', pname);
fprintf('+ Problem size: %d\n', p);

% Call the path-following algorithm.
fprintf(' CALL THE PATH-FOLLOWING PROXIMAL-NEWTON SOLVER ...\n');
[final_sol, output] = scopt_glearn_pfpna(S_hat, rho_start, rho_end, param);
Xsol = final_sol.x;

% Recalculate the objective value.
fval = -sum(log(eig(Xsol))) + trace(S_hat*Xsol) ...
     + output.p2.rho*norm(Xsol(:), 1);
fprintf(' + The objective value: %6.6f\n\n', fval);

% Call the algorithm directly at rho_end.
fprintf(' CALL THE PROXIMAL-NEWTON SOLVER AT Rho_end ...\n');
param.isBtkLS         = true;
[final_solf, outputf] = scopt_glearn_ls_pna(S_hat, output.p2.rho, param);

% Recalculate the objective value.
Xsolf = final_solf.x;
fvalf = -sum(log(eig(Xsolf))) + trace(S_hat*Xsolf) + rhof*norm(Xsolf(:), 1);
fprintf(' + The objective value: %6.6f\n\n', fvalf);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%