% PURPOSE: This function tests the scopt_glearn_ls_pna solver with real data.
%          The data can be downloaded from
%                http://ima.umn.edu/~maxxa007/send_SICS/.
% NOTE:    
%   scopt_glearn_pga is an implementation of Proximal-Newton algorithm
%   for solving the following problem:
%           min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%           s.t. X > 0.
%
%clear all;
%clc;

% Input data.
%rho   = 0.5;        % Regularization parameter.
%pname = 'Lymph.mat'; %'hereditarybc.mat'; %'Arabidopsis.mat'; %'ER.mat';  % Data file.

plist  = {'Lymph.mat'; 'ER.mat'; 'Leukemia.mat'; 'Arabidopsis.mat'; 'hereditarybc.mat'};
rholst = {0.5; 0.1};

for p1 = 1:length(rholst)
    for p2 = 1:length(plist)
        rho = rholst{p1};
        pname = plist{p2};
        
% Load the data.
dataDir = which('scopt_glearn_pga_exam.m');
dataDir = [dataDir(1:end-23), 'glearn_data/'];
if exist([dataDir, pname])
    data  = load([dataDir, pname]);
    S_hat = data.S;
    p     = size(S_hat, 1);
else
    error('Cannot load the input data');
end

% Print problem information.
fprintf('+ Problem name: %s\n', pname);
fprintf('+ Problem size: %d\n', p);

% Default parameters.
param = scopt_optimset([]);

% User defined parameters.
param.AbsTol      = 1e-8;   % Absolute tolerance.
param.RelTol      = 1e-8;   % Relative tolerance.
param.MaxIters    = 200;    % Maximum number of iterations.
param.SubMaxIters = 0;      % Maximum number of iterations in subsolver.
param.SubRelTol   = 1e-9;   % Tolerance for terminating subsolver.
param.BtkMaxIters = 200;    % Maximum number of line-search iterations.
param.PrintLevel  = 2;      % Print 50 iterations for each line.
param.bcdsSubTol  = 1e-2;   % Tolerance for Block coordinate descent.
param.QuadCvRg    = 0;
param.numFxCheck  = 4;      % Terminate if the objective value does not 
                            % change significantly after 4 successive
                            % iterations.
                            
param.isFxEval    = true;   % Require to evaluate the objective value.
param.isTrace     = 1;      % Do not keep trace.
param.SubMethod   = 'dFISTA'; % Method for solving the subproblem.
param.ubSubIters  = 500;
SIGMA             = 0.219224;

% Generate an initial point.
X0 = diag(1./diag(S_hat));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                DUAL FISTA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
param.SubMethod = 'dFISTA';

%%%********************* CALL WITH D-FISTA and No-LS **********************
% Call the solver with dual-FISTA and with/without line-search.
param.TerminateType = 1;
param.isBtkLS       = false;
fprintf(' CALL THE SOLVER ...\n');
[fsol1_nls, output1_nls] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol1_nls = fsol1_nls.x;
fval1_nls = -sum(log(eig(Xsol1_nls))) + trace(S_hat*Xsol1_nls) ...
          + rho*norm(Xsol1_nls(:), 1);

%%%********************* CALL WITH D-FISTA and EBTK-LS ********************
%%% Call the solver with block-coordinate descent and Line-search.
param.TerminateType = 1;
param.lsMode        = 'EBTK';
param.isBtkLS       = true;
fprintf(' CALL THE SOLVER ...\n');
[fsol1_ebtk, output1_ebtk] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol1_ebtk = fsol1_ebtk.x;
fval1_ebtk = -sum(log(eig(Xsol1_ebtk))) + trace(S_hat*Xsol1_ebtk) ...
           + rho*norm(Xsol1_ebtk(:), 1);

%%%********************* CALL WITH D-FISTA and LS *************************
%%% Call the solver with block-coordinate descent and Line-search.
param.TerminateType = 1;
param.lsMode        = 'FWLK';
param.isBtkLS       = true;
fprintf(' CALL THE SOLVER ...\n');
[fsol1_fwlk, output1_fwlk] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol1_fwlk = fsol1_fwlk.x;
fval1_fwlk = -sum(log(eig(Xsol1_fwlk))) + trace(S_hat*Xsol1_fwlk) ...
           + rho*norm(Xsol1_fwlk(:), 1);

      
%%%********************* CALL WITH D-FISTA and LS *************************
% Call the solver with dual-FISTA and with/without line-search.
param.TerminateType = 1;
param.lsMode        = 'BTK';
param.isBtkLS       = true;
fprintf(' CALL THE SOLVER ...\n');
[fsol1_btk, output1_btk] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol1_btk = fsol1_btk.x;
fval1_btk = -sum(log(eig(Xsol1_btk))) + trace(S_hat*Xsol1_btk) ...
          + rho*norm(Xsol1_btk(:), 1);
      

%%%************************ COMPARE WITH QUIC *****************************
isquic_run = true;

% Compare with QUIC.
if isquic_run
    % Parameters for QUIC.
    param.BtkMaxIters = 100;
    param.RelTolFun   = 1e-8;%2.22e-16;
    param.MaxIters    = 1;

    % Call the QUIC solver 
    [Xsol_quic, Wsol, fopt, output_quic] = mat_QUIC(S_hat, rho, X0, param);
    fval_quic = -sum(log(eig(Xsol_quic))) + trace(S_hat*Xsol_quic) ...
              + rho*norm(Xsol_quic(:), 1);
end

% User defined parameters.
param.RelTol      = 1e-7;   % Relative tolerance.
param.MaxIters    = 10000;  % Maximum number of iterations.
param.RelTolFun   = 1e-7;   % Relative tolerance for objective values.
param.isFxEval    = true;   % Require to evaluate the objective value.
param.numFxCheck  = 4;      % Terminate if the objective value does not 
                            % change significantly after 4 successive
                            % iterations.
param.isTrace     = 0;      % Do not keep trace.
param.BtkMaxIters = 200;    % Maximum number of line-search iterations.
param.PrintLevel  = 1;      % Print 50 iterations for each line.
param.BBStepType  = 0;      % Do not perform the Barzilai-Borwein step size.

% Call the algorithm.
param.isBtkLS = false; % Turn of line-search.
fprintf('+ CALL THE SOLVER WITHOUT LINE-SEARCH ...\n');
[fsol_pg_nls, output_pg_nls] = scopt_glearn_pga(S_hat, rho, param);

% Recalculate the objective value.
Xsol_pg_nls = fsol_pg_nls.X;
fval_pg_nls = -sum(log(eig(Xsol_pg_nls))) + trace(S_hat*Xsol_pg_nls) ...
            + rho*norm(Xsol_pg_nls(:), 1);

% Call the algorithm.
param.isBtkLS = true; % Turn on line-search.
fprintf('+ CALL THE SOLVER WITHOUT LINE-SEARCH ...\n');
[fsol_pg_ls, output_pg_ls] = scopt_glearn_pga(S_hat, rho, param);

% Recalculate the objective value.
Xsol_pg_ls = fsol_pg_ls.X;
fval_pg_ls = -sum(log(eig(Xsol_pg_ls))) + trace(S_hat*Xsol_pg_ls) ...
           + rho*norm(Xsol_pg_ls(:), 1);

% Save the results.
result_dir = which('scopt_glearn_pn_exam.m');
result_dir = [result_dir(1:end-22), 'glearn_results/'];
ext = '_pg_pn';
save([result_dir, pname(1:end-4),'_',mat2str(rho), ext,'.mat'], ...
    'fsol1_nls',   'fsol1_btk',   'fsol1_ebtk',   'fsol1_fwlk', ...
    'Xsol1_nls',   'Xsol1_btk',   'Xsol1_ebtk',   'Xsol1_fwlk', ...
    'output1_nls', 'output1_btk', 'output1_ebtk', 'output1_fwlk', ...
    'fval1_nls',   'fval1_btk',   'fval1_ebtk',   'fval1_fwlk', ...
    'fsol_pg_nls', 'fsol_pg_ls',  'Xsol_pg_nls', 'Xsol_pg_ls', ...
    'fval_pg_nls', 'fval_pg_ls', 'output_pg_nls', 'output_pg_ls', ...
    'Xsol_quic', 'fval_quic', 'output_quic', 'S_hat', 'rho', ...
    'Wsol', 'fopt', 'param');

    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%