%
% PURPOSE: This function tests the scopt_glearn_pga solver with real data.
%          The data can be downloaded from
%                http://ima.umn.edu/~maxxa007/send_SICS/.
% NOTE:    
%   scopt_glearn_pga is an implementation of Proximal-Gradient algorithm
%   for solving the following problem:
%           min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%           s.t. X > 0.
%

% Input data.
rho   = 0.3;
pname = 'ER.mat';

% Load the data.
dataDir = which('scopt_glearn_pga_exam.m');
dataDir = [dataDir(1:end-23), 'glearn_data/'];
if exist([dataDir, pname])
    data  = load([dataDir, pname]);
    S_hat = data.S;
    p     = size(S_hat, 1);
else
    error('Cannot load the input data');
end

% Default algorithmic parameters.
param             = scopt_optimset([]);

% User defined parameters.
param.RelTol      = 1e-7;   % Relative tolerance.
param.MaxIters    = 10000;  % Maximum number of iterations.
param.RelTolFun   = 1e-7;   % Relative tolerance for objective values.
param.isFxEval    = true;   % Require to evaluate the objective value.
param.numFxCheck  = 4;      % Terminate if the objective value does not 
                            % change significantly after 4 successive
                            % iterations.
param.isTrace     = 1;      % Do not keep trace.
param.BtkMaxIters = 200;    % Maximum number of line-search iterations.
param.PrintLevel  = 1;      % Print 50 iterations for each line.
param.BBStepType  = 0;      % Do not perform the Barzilai-Borwein step size.

% Problem information.
fprintf('+ Problem name: %s\n', pname);
fprintf('+ Problem size: %d\n', p);

% Call the algorithm.
% param.isBtkLS = true; % Turn of line-search.
% fprintf('+ CALL THE SOLVER WITHOUT LINE-SEARCH ...\n');
% [final_sol_nols, output_nols] = scopt_glearn_pga(S_hat, rho, param);

% Call the algorithm.
param.isBtkLS = true; % Turn on line-search.
fprintf('+ CALL THE SOLVER WITHOUT LINE-SEARCH ...\n');
[final_sol_ls, output_ls] = scopt_glearn_pga(S_hat, rho, param);

% Call the algorithm.
param.isBtkLS = true; % Turn on line-search.
fprintf('+ CALL THE SOLVER WITHOUT LINE-SEARCH AND ENHANCED TECHNIQUE ...\n');
[final_sol_lse, output_lse] = scopt_glearn_fb_pga(S_hat, rho, param);

% Get the final solution.
% Xsol = final_sol_nols.X;
% fval_nols = -sum(log(eig(Xsol))) + trace(S_hat*Xsol) + rho*norm(Xsol(:), 1);
% fprintf(' + The objective value (No-backtracking line search): %6.6f\n', fval_nols);
Xsol = final_sol_ls.X;
fval_ls   = -sum(log(eig(Xsol))) + trace(S_hat*Xsol) + rho*norm(Xsol(:), 1);
fprintf(' + The objective value (Backtracking line search): %6.6f\n\n', fval_ls);
Xsol = final_sol_lse.X;
fval_lse  = -sum(log(eig(Xsol))) + trace(S_hat*Xsol) + rho*norm(Xsol(:), 1);
fprintf(' + The objective value (Backtracking line search and enhanced technique): %6.6f\n\n', fval_lse);

% Compare with ALM. See http://www1.se.cuhk.edu.hk/~sqma/ALM-SICS.html.
if 0%exist('SICS_ALM.m')
    fprintf('+ CALL THE ALM SOLVER ... \n');
    opts.mxitr     = param.MaxIters; % max iteration number.
    opts.mu0       = p; 
    opts.mu0       = 1e-1;       % initial mu.
    opts.muf       = 1e-3;       % final mu.
    opts.rmu       = 1/4;        % ratio of decreasing mu.
    opts.tol_gap   = 1e-1;       % tolerance for duality gap.
    opts.tol_frel  = param.RelTolFun; 
    opts.tol_Xrel  = param.RelTolX;   % tolerance for relative change of X.
    opts.tol_Yrel  = param.RelTolX;   % tolerance for relative change of Y.
    opts.numDG     = 1000;       % every numDG iterations, we compute
                                 % duality gap since it's expensive.
    opts.record    = 0;          % print stats.
    opts.sigma     = 1e-10;      % sigma is the smoothness parameter.

    timeaml        = tic;
    outALM         = SICS_ALM(S_hat, rho, opts);
    outALM.time    = toc(timeaml);
    Xsol_alm       = outALM.X;
    fval_alm       = -sum(log(eig(Xsol_alm))) + trace(S_hat*Xsol_alm) ...
                   + rho*norm(Xsol_alm(:), 1);
	fprintf('+ The objective value of ALM: %6.12f\n', fval_alm);
    fprintf('+ Computational time (s): %6.4f\n', outALM.time);
    fprintf('+ Finished!\n');
end

% Compare with APG.
% See http://people.math.sfu.ca/~zhaosong/.
if 0% exist('nest_solver.m')
    fprintf('+ CALL THE APG SOLVER ... \n');
    U         = zeros(p,p);
    init_beta = 1.0;
    
    outAPG.time = tic;
    [Xsol_apg, dump1, dump2, dump3, dump4] = nest_solver(S_hat, ...
                  rho*ones(size(S_hat)),U, init_beta, param.MaxIters, ...
                  0, 5e-2);
    outAPG.time = toc(outAPG.time);
    fval_apg    = -sum(log(eig(Xsol_apg))) + trace(S_hat*Xsol_apg) ...
                + rho*norm(Xsol_apg(:), 1);
	fprintf('+ The objective value of APG: %6.12f\n', fval_apg);
    fprintf('+ Computational time (s): %6.4f\n', outAPG.time);
    fprintf('+ Finished!\n');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%