% PURPOSE: This function tests the scopt_glearn_ls_pna solver with real data.
%          The data can be downloaded from
%                http://ima.umn.edu/~maxxa007/send_SICS/.
% NOTE:    
%   scopt_glearn_pga is an implementation of Proximal-Newton algorithm
%   for solving the following problem:
%           min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%           s.t. X > 0.
%
clear all;
clc;

% Input data.
rho   = 0.5;          % Regularization parameter.
pname = 'Lymph.mat';  % Data file.

% Load the data.
dataDir = which('scopt_glearn_pga_exam.m');
dataDir = [dataDir(1:end-23), 'glearn_data/'];
if exist([dataDir, pname])
    data  = load([dataDir, pname]);
    S_hat = data.S;
    p     = size(S_hat, 1);
else
    error('Cannot load the input data');
end

% Print problem information.
fprintf('+ Problem name: %s\n', pname);
fprintf('+ Problem size: %d\n', p);

% Default parameters.
param = scopt_optimset([]);

% User defined parameters.
param.AbsTol      = 1e-8;   % Absolute tolerance.
param.RelTol      = 1e-8;   % Relative tolerance.
param.MaxIters    = 200;    % Maximum number of iterations.
param.SubMaxIters = 0;      % Maximum number of iterations in subsolver.
param.SubRelTol   = 1e-9;   % Tolerance for terminating subsolver.
param.BtkMaxIters = 200;    % Maximum number of line-search iterations.
param.PrintLevel  = 2;      % Print 50 iterations for each line.
param.bcdsSubTol  = 1e-2;   % Tolerance for Block coordinate descent.
param.QuadCvRg    = 0;
param.numFxCheck  = 4;      % Terminate if the objective value does not 
                            % change significantly after 4 successive
                            % iterations.
                            
param.isFxEval    = true;   % Require to evaluate the objective value.
param.isTrace     = 1;      % Do not keep trace.
param.SubMethod   = 'dFISTA'; % Method for solving the subproblem.
param.ubSubIters  = 500;
SIGMA             = 0.219224;

% Generate an initial point.
X0 = diag(1./diag(S_hat));

%%%********************* CALL WITH D-FISTA and No-LS **********************
% Call the solver with dual-FISTA and with/without line-search.
param.TerminateType = 1;
param.isBtkLS       = false;
fprintf(' CALL THE SOLVER ...\n');
[fsol_nls, output_nls] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol_nls = fsol_nls.x;
fval_nls = -sum(log(eig(Xsol_nls))) + trace(S_hat*Xsol_nls) ...
         + rho*norm(Xsol_nls(:), 1);

%%%********************* CALL WITH D-FISTA and EBTK-LS ********************
%%% Call the solver with block-coordinate descent and Line-search.
param.TerminateType = 1;
param.lsMode        = 'EBTK';
param.isBtkLS       = true;
fprintf(' CALL THE SOLVER ...\n');
[fsol_ebtk, output_ebtk] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol_ebtk = fsol_ebtk.x;
fval_ebtk = -sum(log(eig(Xsol_ebtk))) + trace(S_hat*Xsol_ebtk) ...
          + rho*norm(Xsol_ebtk(:), 1);

%%%********************* CALL WITH D-FISTA and LS *************************
%%% Call the solver with block-coordinate descent and Line-search.
param.TerminateType = 1;
param.lsMode        = 'FWLK';
param.isBtkLS       = true;
fprintf(' CALL THE SOLVER ...\n');
[fsol_fwlk, output_fwlk] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol_fwlk = fsol_fwlk.x;
fval_fwlk = -sum(log(eig(Xsol_fwlk))) + trace(S_hat*Xsol_fwlk) ...
          + rho*norm(Xsol_fwlk(:), 1);

%%%********************* CALL WITH D-FISTA and LS *************************
% Call the solver with dual-FISTA and with/without line-search.
param.TerminateType = 1;
param.lsMode        = 'BTK';
param.isBtkLS       = true;
fprintf(' CALL THE SOLVER ...\n');
[fsol_btk, output_btk] = scopt_glearn_ls_pna(S_hat, rho, param, X0);

% Recalculate the objective value.
Xsol_btk = fsol_btk.x;
fval_btk = -sum(log(eig(Xsol_btk))) + trace(S_hat*Xsol_btk) ...
         + rho*norm(Xsol_btk(:), 1);

%%%************************ COMPARE WITH QUIC *****************************
isquic_run = true;
% Compare with QUIC.
if isquic_run
% Parameters for QUIC.
param.BtkMaxIters = 100;
param.RelTolFun   = 1e-8;%2.22e-16;
param.MaxIters    = 1;

% Call the QUIC solver 
[Xsol_quic, Wsol, fopt, output_quic] = mat_QUIC(S_hat, rho, X0, param);
fval_quic = -sum(log(eig(Xsol_quic))) + trace(S_hat*Xsol_quic) ...
          + rho*norm(Xsol_quic(:), 1);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%% PRINT THE OUTPUTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf(' ************** THE FINAL RESULTS ****************\n');
fprintf(' + The objective value (No-line-search):  %6.16f\n', fval_nls);
fprintf(' + The objective value (Forward-Looking): %6.16f\n', fval_fwlk);
fprintf(' + The objective value (E-Backtracking:   %6.16f\n', fval_ebtk);
fprintf(' + The objective value (Backtracking):    %6.16f\n', fval_btk);
if isquic_run
    fprintf(' + The objective value (QUIC):            %6.16f\n', fval_quic);
end

fprintf(' + CPU Time (No-line-search):     %6.7f.\n', output_nls.time);
fprintf(' + CPU Time (Forward-looking):    %6.7f.\n', output_fwlk.time);
fprintf(' + CPU Time (E-Backtracking):     %6.7f.\n', output_ebtk.time);
fprintf(' + CPU Time (Backtracking):       %6.7f.\n', output_btk.time);
if isquic_run
    fprintf(' + CPU Time (QUIC):               %6.7f.\n', output_quic.time);
end

fprintf(' + Number of iterations (No-line-search):  %5d.\n', output_nls.iter);
fprintf(' + Number of iterations (Forward-Looking): %5d.\n', output_fwlk.iter);
fprintf(' + Number of iterations (E-Backtracking):  %5d.\n', output_ebtk.iter);
fprintf(' + Number of iterations (Backtracking):    %5d.\n', output_btk.iter);
if isquic_run
    fprintf(' + Number of iterations (QUIC):            %5d.\n', output_quic.iter);
end

fprintf(' + Number of Cholesky decompositions (Forward-Looking): %5d.\n', ...
            output_fwlk.chol_cnt);
fprintf(' + Number of Cholesky decompositions (E-Backtracking):  %5d.\n', ...
            output_ebtk.chol_cnt);
fprintf(' + Number of Cholesky decompositions (Backtracking):    %5d.\n', ...
            output_btk.chol_cnt);
if isquic_run
    fprintf(' + Number of Cholesky decompositions (QUIC):            %5d.\n', ...
                output_quic.chol_cnt);
end

Xsol_nls = scopt_soft_threshold_oper(Xsol_nls, 1e-6);

% Save the results.
result_dir = which('scopt_glearn_pn_exam.m');
result_dir = [result_dir(1:end-22), 'glearn_results/'];
if param.isTrace, ext = '_full'; else ext = ''; end
save([result_dir, pname(1:end-4),'_',mat2str(rho), ext,'.mat'], 'fsol_nls', ...
    'fsol_btk', 'fsol_ebtk', 'fsol_fwlk', 'Xsol_nls', 'Xsol_btk', ...
    'Xsol_ebtk', 'Xsol_fwlk', 'Xsol_quic', 'output_nls', 'output_btk', ...
    'output_ebtk', 'output_fwlk', 'output_quic', 'S_hat', 'rho', ...
    'Wsol', 'fopt', 'param', 'fval_nls', 'fval_btk', 'fval_ebtk', ...
    'fval_fwlk', 'fval_quic');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%