% PURPOSE: This function tests the logistic regression solver.
%
% DATE: 08.04.2013.
%
% INPUTS:   file_name  : the name of the data file.
%           method     : proximal-gradient or proximal-Newton method.
%

% Input data.
rho_f         = 1/4;
file_name     = 'rcv1';

% If LIBSVM does not exist then use default mat data.
mat_file_name = 'w4a.mat';
fprintf('+ Read the data from file: %s ...\n', file_name);

if exist('svm_model_matlab.h')
    svm_path = which('svm_model_matlab.h');
    svm_path = svm_path(1:end-25);
    % Addpath to the libsvm toolbox.
    addpath([svm_path, 'matlab/']);
end
data_path = which('scopt_logistic_exam.m');
dirData   = [data_path(1:end-21), 'logistic_data/'];

% Read data from the data set.
if exist('libsvmread.m')
    [yhat, What] = libsvmread(fullfile(dirData, file_name));
    %save([dirData, mat_file_name], 'yhat', 'What');
else
    mdata = load([dirData, mat_file_name]);
    yhat  = mdata.yhat;
    What  = mdata.What;
end
if isempty(yhat) || isempty(What)
    error('Can not read the data from the file!');
end
bias         = 0;

Wold = What;

p = size(What,2);
N = size(What,1);
fprintf(' + We are solving problem: %s (p = %d, N = %d)\n', ...
         file_name, p, N);

% The penalty parameter.
rho = rho_f/sqrt(N);

% Default parameters.
param                   = scopt_optimset([]);

% User-defined parameters.
param.RelTolX           = 1e-7;
param.MaxIters          = 10000;   
param.SubMaxIters       = 100;
param.SubRelTol         = 1e-7;
param.RelTolFun         = 1e-12;
param.isFxEval          = true;
param.isDebug           = true;
param.numFxCheck        = 5;
param.PrintLevel        = 1;
param.UpdateH           = 'bfgs';
param.BBStepType        = 3;
param.isRestart         = 'none';
param.isBtkLS           = true;

% Temporary parameters.
Ahat  = repmat(yhat, 1, size(What,2)).*What;
bhat  = bias*yhat;

% Generate initial point.
x0 = 0*ones(p, 1);

%---------------------------------------------------------
% Solve the problem by the Normal proximal-Gradient method.
fprintf(' + CALL THE SOLVER (The normal proximal-gradient method) ...\n');
% Turn of the accelerating parameter.
param.isAccelerating = false;
param.isBktLS        = true;
[fsol_pg, output_pg] = scopt_logistic_pga(What, yhat, bias, rho, ...
                       param, x0);
% Evaluate the objective value.
xsol_pg = fsol_pg.x;
fval_pg = scopt_logistic_fx_eval(xsol_pg, Ahat, bhat, N) + rho*norm(xsol_pg, 1);    
nnz_pg  = nnz(round(xsol_pg*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE PROXIMAL-GRADIENT ALGORITHM ***\n');
fprintf(' + The objective value: %6.15f\n', fval_pg);
fprintf(' + CPU time: %6.4f\n', output_pg.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pg);
fprintf('********************************************************\n');

%--------------------------------------------------------
% Solve the problem by the Fast Proximal-Gradient method.
fprintf(' + CALL THE SOLVER (the fast proximal-gradient method) ...\n');
% Turn on the accelerating parameter.
param.isAccelerating   = true;
param.isBktLS          = true;
[fsol_fpg, output_fpg] = scopt_logistic_pga(What, yhat, bias, rho, ...
                         param, x0);
% Evaluate the objective value.
xsol_fpg = fsol_fpg.x;
fval_fpg = scopt_logistic_fx_eval(xsol_fpg, Ahat, bhat, N) ...
         + rho*norm(xsol_fpg, 1);    
nnz_fpg  = nnz(round(xsol_fpg*1e10)*1e-10);
fprintf('* FINAL OUTPUTS OF THE FAST PROXIMAL-GRADIENT ALGORITHM *\n');
fprintf(' + The objective value: %6.15f\n', fval_fpg);
fprintf(' + CPU time: %6.4f\n', output_fpg.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_fpg);
fprintf('********************************************************\n');

%---------------------------------------------------------------------
% Solve the problem by the Fast Proximal-Gradient method with restart.
fprintf(' + CALL THE SOLVER (the Fast proximal-gradient method (RS)) ...\n');
% Turn on the accelerating parameter.
param.isAccelerating   = true;
param.isRestart        = 'fixed';
param.isBktLS          = true;
[fsol_fpg1, output_fpg1] = scopt_logistic_pga(What, yhat, bias, rho, ...
                           param, x0);
% Evaluate the objective value.
xsol_fpg1 = fsol_fpg1.x;
fval_fpg1 = scopt_logistic_fx_eval(xsol_fpg1, Ahat, bhat, N) ...
            + rho*norm(xsol_fpg1, 1);    
nnz_fpg1  = nnz(round(xsol_fpg1*1e10)*1e-10);
fprintf('* FINAL OUTPUTS OF THE FAST PG ALGORITHM (RESTART) *\n');
fprintf(' + The objective value: %6.15f\n', fval_fpg1);
fprintf(' + CPU time: %6.4f\n', output_fpg1.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_fpg1);
fprintf('****************************************************\n');

%----------------------------------------------------------------------
% Solve the problem by the Fast PG method with restart and backtracking.
fprintf(' + CALL THE SOLVER (The Fast proximal-gradient method (RS-BT))...\n');
% Turn on the accelerating parameter.
param.isAccelerating   = true;
param.isRestart        = 'fx';
param.isBktLS          = true;
[fsol_fpg2, output_fpg2] = scopt_logistic_pga(What, yhat, bias, ...
                           rho, param, x0);
% Evaluate the objective value.
xsol_fpg2 = fsol_fpg2.x;
fval_fpg2 = scopt_logistic_fx_eval(xsol_fpg2, Ahat, bhat, N) ...
            + rho*norm(xsol_fpg2, 1);    
nnz_fpg2  = nnz(round(xsol_fpg2*1e10)*1e-10);
fprintf(['* FINAL OUTPUTS OF THE FAST PROXIMAL-GRADIENT ALGORITHM', ...
        ' (RESTART-BACKTRACKING) *\n']);
fprintf(' + The objective value: %6.15f\n', fval_fpg2);
fprintf(' + CPU time: %6.4f\n', output_fpg2.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_fpg2);
fprintf(['*****************************************************', ...
         '***************************\n']);

%---------------------------------------------------------------------
% Solve the problem by the Proximal-Newton method.
fprintf(' + CALL THE SOLVER (the proximal-Newton method) ...\n');
param.UpdateH          = 'newton';
[fsol_pn1, output_pn1] = scopt_logistic_pna(What, yhat, bias, rho, ...
                         x0, param);
% Evaluate the objective value.
xsol_pn1 = fsol_pn1.x;
fval_pn1 = scopt_logistic_fx_eval(xsol_pn1, Ahat, bhat, N) ...
           + rho*norm(xsol_pn1, 1);
nnz_pn1  = nnz(round(xsol_pn1*1e10)*1e-10);
fprintf('**** FINAL OUTPUTS OF THE PROXIMAL-NEWTON ALGORITHM ****\n');
fprintf(' + The objective value: %6.15f\n', fval_pn1);
fprintf(' + CPU time: %6.4f\n', output_pn1.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pn1);
fprintf('********************************************************\n');

%------------------------------------------------------
% Solve the problem by the BFGS proximal-Newton method.
fprintf(' + CALL THE SOLVER (the BFGS proximal-Newton method) ...\n');
param.UpdateH          = 'bfgs';
[fsol_pn2, output_pn2] = scopt_logistic_pna(What, yhat, bias, rho, ...
                         x0, param);
% Evaluate the objective value.
xsol_pn2 = fsol_pn2.x;
fval_pn2 = scopt_logistic_fx_eval(xsol_pn2, Ahat, bhat, N) ...
         + rho*norm(xsol_pn2, 1);
nnz_pn2  = nnz(round(xsol_pn2*1e10)*1e-10);
fprintf('**** FINAL OUTPUTS OF THE PROXIMAL-NEWTON ALGORITHM ****\n');
fprintf(' + The objective value: %6.15f\n', fval_pn2);
fprintf(' + CPU time: %6.4f\n', output_pn2.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pn2);
fprintf('********************************************************\n');

%---------------------------------------------------------
% Solve the problem by the LBFGS - proximal-Newton method.
fprintf(' + CALL THE SOLVER (the LBFGS proximal-Newton method) ...\n');
param.UpdateH          = 'lbfgs';
[fsol_pn3, output_pn3] = scopt_logistic_pna(What, yhat, bias, rho, ...
                         x0, param);
% Evaluate the objective value.
xsol_pn3 = fsol_pn3.x;
fval_pn3 = scopt_logistic_fx_eval(xsol_pn3, Ahat, bhat, N) ...
         + rho*norm(xsol_pn3, 1);
nnz_pn3  = nnz(round(xsol_pn3*1e10)*1e-10);
fprintf('**** FINAL OUTPUTS OF THE PROXIMAL-NEWTON ALGORITHM ****\n');
fprintf(' + The objective value: %6.15f\n', fval_pn3);
fprintf(' + CPU time: %6.4f\n', output_pn3.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pn3);
fprintf('********************************************************\n');

%------------------------------------------------
% Compare with TFOCS. See http://cvxr.com/tfocs/.  
%------------------------------------------------
if exist('tfocs.m')
    fprintf(' + CALL THE TFOCS SOLVER ...\n');
    time1 = tic;
    Ahat  = repmat(yhat, 1, size(What,2)).*What;
    bhat  = bias*yhat;
    N     = size(What, 1);
    fx    = @(x)(scopt_logistic_fx_eval(x, Ahat, bhat, N));
    % Set options.
    opts.maxIts       = param.MaxIters;
    opts.tol          = param.RelTolX;
    opts.algChoice    = 'N07';
    opts.restartState = 100;

    % Call the TFOCS solver.
    [xsol_tfocs, output_tfocs, info] = tfocs(fx, [], prox_l1(rho), x0, opts);
    output_tfocs.time                = toc(time1);
    % Evaluate the objective value.
    fval_tfocs = scopt_logistic_fx_eval(xsol_tfocs, Ahat, bhat, N) ...
               + rho*norm(xsol_tfocs, 1);
    nnz_tfocs  = nnz(round(xsol_tfocs*1e10)*1e-10);
    fprintf('**** FINAL OUTPUTS OF TFOCS ****\n');
    fprintf(' + The objective value: %6.15f\n', fval_tfocs);
    fprintf(' + CPU time: %6.4f\n', output_tfocs.time);
    fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_tfocs);
    fprintf('********************************************************\n');
end

%-----------------------------------------------------------
% Compare with PNOPT. 
% See http://www.stanford.edu/group/SOL/software/pnopt.html.
%-----------------------------------------------------------
if exist('pnopt.m')
    % Define the input functions.
    time1 = tic;
    Ahat  = repmat(yhat, 1, size(What,2)).*What;
    bhat  = bias*yhat;
    N     = size(What, 1);
    fx    = @(x)(scopt_logistic_fx_eval(x, Ahat, bhat, N));
    % Set options.
    method     = 'Lbfgs';
    pnopt_opts = pnopt_optimset( 'maxIter', param.MaxIters, ...
                                 'optim_tol', 0.1*param.RelTolX, ...
                                 'xtol', param.RelTolX, ...
                                 'ftol', param.RelTolFun, ...
                                 'method', method, ...
                                 'display', param.PrintLevel );
    % Call the PNOPT solver.
    [xsol_pnopt, fval_pnopt, output_pnopt] = pnopt(fx, prox_l1(rho), ...
                                             x0, pnopt_opts);
    output_pnopt.time = toc(time1);
    % Evaluate the objective value.
    fval_pnopt = scopt_logistic_fx_eval(xsol_pnopt, Ahat, bhat, N) ...
                 + rho*norm(xsol_pnopt, 1);   
    nnz_pnopt  = nnz(round(xsol_pnopt*1e10)*1e-10);
    fprintf('*************** FINAL OUTPUT OF PNOPT ******************\n');
    fprintf(' + The objective value: %6.15f\n', fval_pnopt);
    fprintf(' + CPU time: %6.4f\n', output_pnopt.time);
    fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pnopt);
    fprintf('********************************************************\n');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%