% PURPOSE:  Test the scopt_maxnorm_pfpna by using some synthetic data.
%
%

% Input data.
p    = 2;
Pe   = 0.1;
Cset = randi([10, 20], 1, p);

% Default parameters.
param             = scopt_optimset([]);

% User-defined parameters.
param.MaxItersP1  = 2000;
param.MaxItersP2  = 10000;
param.AbsTol      = 1e-8;
param.TolParam    = 1e-3;
param.SubMaxIters = 300;
param.SubRelTol   = 1e-7;
param.isFxEval    = false; 
param.isTrace     = false;
param.PrintLevel  = 1;

% Penalty parameter.
tau = 0.2;

% Generate matrix AK.
K = ones(Cset(1));
for i = 2:max(size(Cset))
    K = [K zeros(sum(Cset(1:i-1)),Cset(i)); zeros(Cset(i), ...
         sum(Cset(1:i-1))) ones(Cset(i))];
end
B  = double(rand(size(K)) > 1 - Pe);
AK = K;
for i=1:sum(Cset)
    for j=i+1:sum(Cset)
        if (B(i,j) == 1)
            AK(i,j) = 1 - K(i,j);
            AK(j,i) = 1 - K(j,i);
        end
    end
end
p          = size(AK, 1);
uLi        = 1;
uRi        = 1;

% Call the algorithm.
fprintf('+ Problem size: |C| = %d\n', sum(Cset));
fprintf('+ CALL THE SOLVER ...\n');
[final_sol, output] = scopt_maxnorm_pfpna(AK, uLi, uRi, [], [], [], ...
                                          tau, param);

% The final solution.
Ksol      = final_sol.K;
Lsol      = final_sol.L;
Rsol      = final_sol.R;
Ka        = Ksol - AK;
fval      = norm(Ka(:), 1);
fprintf(' + The objective value given by the solver: %6.12f\n\n', fval);

% Compute the nonzero elements.
tol   = 1e-8;
Ka(abs(Ka) < tol) = 0;
nnzKa = nnz(Ka);
Ax    = [Lsol, Ksol; Ksol', Rsol];

% Print the final results.                                                
fprintf(' + The real-objective value: %6.12f\n\n', fval);
fprintf(' + Number of nonzero elements of K-AK: %3d\n', nnzKa);
fprintf(' + Min(eig(Ax)) = %6.6f\n', min(eig(Ax)));
fprintf(' + Max(diag(L)): %3.6f and Max(diag(R)): %3.6f\n', ...
          max(diag(Lsol)), max(diag(Rsol)));

% Compare with Sedumi (CVX).
% See http://cvxr.com.
if exist('cvx_begin.m')
    time_cvx = tic;
    if param.PrintLevel == 0, cvx_quiet true; end
    cvx_precision(param.TolParam);
    cvx_begin sdp;
        variable K(p, p);
        variable L(p, p) symmetric;
        variable R(p, p) symmetric;

        minimize( norm(AK(:) - K(:), 1) )
        subject to
            [L  K; K' R] >= 0;
            diag(L) <= uLi*ones(p, 1);
            diag(R) <= uRi*ones(p, 1);
    cvx_end
    time_cvx  = toc(time_cvx);
    Ksol_cvx  = K;
    Lsol_cvx  = L;
    Rsol_cvx  = R;
    Ka_cvx    = Ksol_cvx - AK;
    Ka_cvx(abs(Ka_cvx) < tol) = 0;
    fval_cvx  = norm(Ka_cvx(:), 1);
    nnzKa_cvx = nnz(Ka_cvx);
    Ax_cvx    = [Lsol_cvx, Ksol_cvx; Ksol_cvx', Rsol_cvx];
    fprintf(' + The real-objective value of CVX: %6.12f\n', fval_cvx);
    fprintf(' + CPU time of CVX: %4.6f\n', time_cvx);
    fprintf(' + Min(eig(Ax)) = %6.6f\n', min(eig(Ax_cvx)));
    fprintf(' + Max(diag(L)): %3.6f and Max(diag(R)): %3.6f\n', ...
              max(diag(Lsol_cvx)), max(diag(Rsol_cvx)));
    nrm_x = norm(Ksol - Ksol_cvx, 1) / max(norm(Ksol_cvx, 1), 1);
    fprintf([' + Compare two solutions: norm(Ksol-Ksol_cvx, 1)', ...
             '/max(1, norm(Ksol_cvx, 1)): %6.7f\n'], nrm_x);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%