%   This function tests the multinomial logistic regression solver.
%   Date: 11.04.2013.
%
% INPUTS:   
%    file_name : the name of the data file.

% OUTPUTS: 
%    xsol      : the final approximate solution.
%    output    : the output information.
%
% NOTE: In order to run this script, one needs to download LIBSVM to
%       read the data. This package is available at 
%              http://www.csie.ntu.edu.tw/~cjlin/libsvm/. 

% Input data.
rhof = 1.0;  % Regularization parameter.

% Data file
% this data set can be downloaded at
% http://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/multiclass.html.
file_name     = 'glass.scale.txt';  

% If LIBSVM does not exist, then use the defalt mat data.
mat_file_name = 'glass.scale.mat';  
                                
fprintf('---> Read the data from file ...\n');
if exist('svm_model_matlab.h')
    svm_path = which('svm_model_matlab.h');
    svm_path = svm_path(1:end-18-7);
    % Addpath to the libsvm toolbox.
    addpath([svm_path, 'matlab/']);
end

data_path = which('scopt_mnlogistic_exam.m');
dirData   = [data_path(1:end-23), 'mnlogistic_data/'];

% Check if the file exists.
if ~exist([dirData, file_name])
    file_name = [file_name, '.txt'];
end
if ~exist([dirData, file_name])
    warning('File is not found!');
    Xsol = []; output = [];
    return;
end

% Read data from the data set.
if exist('libsvmread.m')
    [yhat, What] = libsvmread(fullfile(dirData, file_name));
else
    mdata = load([dirData, mat_file_name]);
    yhat  = mdata.yhat;
    What  = mdata.What;
end
if isempty(yhat) || isempty(What)
    error('Can not read the data file!');
end
yhat = double(uint8(yhat));
if any(yhat == 0)
    yhat = yhat + 1;
end
if any(yhat <= 0)
    error('The classification vector must be positive');
end
What = What';

% Problem parameters.
p    = size(What,1);
N    = size(What,2);
mx   = max(yhat);
Yhat = zeros(mx, N);
Yhat(sub2ind([mx, N], yhat(:)', 1:N)) = 1;
Yhat = full(Yhat);
What = full(What);

% Get default parameters.
param                   = scopt_optimset([]);

% User-defined parameters.
param.RelTolX           = 1e-7;
param.MaxIters          = 5000;   
param.SubMaxIters       = 50;
param.SubRelTol         = 1e-7;
param.RelTolFun         = 1e-10;
param.isFxEval          = true;
param.isDebug           = true;
param.numFxCheck        = 5;
param.PrintLevel        = 1;
param.UpdateH           = 'bfgs';
param.BBStepType        = 3;      % Can be 1, 2 or 3.
param.isRestart         = 'none';
param.isBtkLS           = true;

% Regularization parameter.
TN  = N;
rho = rhof/sqrt(N);

% Assign the output.
output.param = param;
output.size  = [mx, p, N];
output.What  = What;
output.Yhat  = Yhat;
output.rho   = rho;
Xsol         = [];

fprintf('+ The size of problem %s: (p = %d, mx = %d, N = %d)\n', ...
        file_name, p, mx, N);

% Generate initial point.
X0 = zeros(size(What,1), size(Yhat,1));

%**********************************************************
% Solve the problem by the Normal Proximal-Gradient method.
fprintf(' + CALL THE SOLVER (Normal Proximal-Gradient) ...\n');
param.isAccelerating = false;
[fsol_pg, output_pg] = scopt_mnlogistic_pga(What, Yhat, rho, X0, param, TN);
% Evaluate the objective value.
xsol_pg = fsol_pg.x;
fval_pg = scopt_mnlogistic_fx_eval(xsol_pg, What, Yhat, TN) ...
        + rho*sum(sum(abs(xsol_pg)));
nnz_pg = nnz(round(xsol_pg*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE PROXIMAL-GRADIENT ALGORITHM ***\n');
fprintf(' + The objective value: %6.15f\n', fval_pg);
fprintf(' + CPU time: %6.4f\n', output_pg.time);
fprintf(' + Number of nonzero elements of xsol: %d\n\n', nnz_pg);
fprintf('********************************************************\n');

%**********************************************************
% Solve the problem by the Fast Proximal-Gradient method.
fprintf(' + CALL THE SOLVER (Fast Proximal-Gradient) ...\n');
param.isAccelerating   = true;
[fsol_fpg, output_fpg] = scopt_mnlogistic_pga(What, Yhat, rho, X0, param, TN);
% Evaluate the objective value.
xsol_fpg = fsol_fpg.x;
fval_fpg = scopt_mnlogistic_fx_eval(xsol_fpg, What, Yhat, TN) ...
         + rho*sum(sum(abs(xsol_fpg)));
nnz_fpg  = nnz(round(xsol_fpg*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE FAST PROXIMAL-GRADIENT ALGORITHM ***\n');
fprintf(' + The objective value: %6.15f\n', fval_fpg);
fprintf(' + CPU time: %6.4f\n', output_fpg.time);
fprintf(' + Number of nonzero elements of xsol: %d\n',nnz_fpg);
fprintf('*************************************************************\n');

%*********************************************************************
% Solve the problem by the Fast Proximal-Gradient method with Restart.
fprintf(' + CALL THE SOLVER(Fast Proximal-Gradient + Restart) ...\n');
param.isAccelerating     = true;
param.isRestart          = 'fx';
[fsol_fpg1, output_fpg1] = scopt_mnlogistic_pga(What, Yhat, rho, X0, param, TN);
% Evaluate the objective value.
xsol_fpg1 = fsol_fpg1.x;
fval_fpg1 = scopt_mnlogistic_fx_eval(xsol_fpg1, What, Yhat, TN) ...
          + rho*sum(sum(abs(xsol_fpg1)));
nnz_fpg1  = nnz(round(xsol_fpg1*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE FAST PROXIMAL-GRADIENT ALGORITHM (RS) ***\n');
fprintf(' + The objective value: %6.15f\n', fval_fpg1);
fprintf(' + CPU time: %6.4f\n', output_fpg1.time);
fprintf(' + Number of nonzero elements of xsol: %d\n',nnz_fpg1);
fprintf('******************************************************************\n');

%*******************************************************************
% Solve the problem by the Fast Proximal-Gradient method with RT-BT.
fprintf('+ CALL THE SOLVER (Fast Proximal-Gradient + Restart + BTLS)) ...\n');
param.isAccelerating    = true;
param.isRestart        = 'fx';
param.isBackTracking   = true;  
[fsol_fpg2, output_fpg2] = scopt_mnlogistic_pga(What, Yhat, rho, X0, param, TN);
% Evaluate the objective value.
xsol_fpg2 = fsol_fpg2.x;
fval_fpg2 = scopt_mnlogistic_fx_eval(xsol_fpg2, What, Yhat, TN) ...
          + rho*sum(sum(abs(xsol_fpg2)));
nnz_fpg2  = nnz(round(xsol_fpg2*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE FAST PROXIMAL-GRADIENT ALGORITHM (RS-BT) ***\n');
fprintf(' + The objective value: %6.15f\n', fval_fpg2);
fprintf(' + CPU time: %6.4f\n', output_fpg2.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_fpg2);
fprintf('*********************************************************************\n');

%******************************************************
% Solve the problem by the BFGS proximal-Newton method.
fprintf(' + CALL THE SOLVER (BFGS Proximal-Newton) ...\n');
param.UpdateH          = 'bfgs';
[fsol_pn2, output_pn2] = scopt_mnlogistic_pna(What, Yhat, rho, X0, param, TN);
% Evaluate the objective value.
xsol_pn2 = fsol_pn2.x;
fval_pn2 = scopt_mnlogistic_fx_eval(xsol_pn2, What, Yhat, TN) ...
         + rho*sum(sum(abs(xsol_pn2)));
nnz_pn2  = nnz(round(xsol_pn2*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE BFGS PROXIMAL-NEWTON ALGORITHM ***\n');
fprintf(' + The objective value: %6.15f\n', fval_pn2);
fprintf(' + CPU time: %6.4f\n', output_pn2.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pn2);
fprintf('***********************************************************\n');

%*********************************************************
% Solve the problem by the LBFGS - proximal-Newton method.
fprintf(' + CALL THE SOLVER (Limitted memory BFGS Proximal-Newton) ...\n');
param.UpdateH          = 'lbfgs';
[fsol_pn3, output_pn3] = scopt_mnlogistic_pna(What, Yhat, rho, X0, param, TN);
% Evaluate the objective value.
xsol_pn3 = fsol_pn3.x;
fval_pn3 = scopt_mnlogistic_fx_eval(xsol_pn3, What, Yhat, TN) ...
         + rho*sum(sum(abs(xsol_pn3)));
nnz_pn3  = nnz(round(xsol_pn3*1e10)*1e-10);
fprintf('*** FINAL OUTPUTS OF THE L-BFGS PROXIMAL-NEWTON ALGORITHM ***\n');
fprintf(' + The objective value: %6.15f\n', fval_pn3);
fprintf(' + CPU time: %6.4f\n', output_pn3.time);
fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pn3);
fprintf('*************************************************************\n');

%*********************************************************
% Compare with TFOCS. See http://cvxr.com/tfocs/.  
if exist('tfocs.m')
    % Define the input functions.
    time_tfocs = tic;
    FX         = @(X)(scopt_mnlogistic_fx_eval3(X, What, Yhat, TN));

    % Set options for TFOCS.
    opts.maxIts  = param.MaxIters;
    opts.tol     = param.RelTolX;
    opts.alg     = 'N07';
    opts.restart = 100;

    % Call the TFOCS solver.
    [xsol_tfocs, output_tfocs, out] = tfocs( FX, [], prox_l1(rho), ...
                                      X0(:), opts );
    xsol_tfocs = reshape(xsol_tfocs, size(What, 1), size(Yhat, 1));
    time_tfocs = toc(time_tfocs);
    
    % Evaluate the objective value.
    fval_tfocs = scopt_mnlogistic_fx_eval(xsol_tfocs, What, Yhat, TN) ...
               + rho*sum(sum(abs(xsol_tfocs)));
    nnz_tfocs  = nnz(round(xsol_tfocs*1e10)*1e-10);
    fprintf('**************** FINAL OUTPUTS OF TFOCS ****************\n');
    fprintf(' + The objective value: %6.15f\n', fval_tfocs);
    fprintf(' + CPU time: %6.4f\n', time_tfocs);
    fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_tfocs);
    fprintf('********************************************************\n');
end

%*********************************************************
% Compare with PNOPT. 
% See http://www.stanford.edu/group/SOL/software/pnopt.html.
if exist('pnopt.m')
    % Define the input functions.
    time_pnopt = tic;
    FX         = @(X)(scopt_mnlogistic_fx_eval3(X, What, Yhat, TN));
    % Set options.
    method     = 'Lbfgs';
    pnopt_opts = pnopt_optimset( 'maxIter', param.MaxIters, ...
                                 'optim_tol', 0.1*param.RelTolX, ...
                                 'xtol', param.RelTolX, ...
                                 'ftol', param.RelTolFun, ...
                                 'method', method, ...
                                 'display', param.PrintLevel );
    % Call the PNOPT solver.
    [xsol_pnopt, fval_pnopt, output_pnopt] = pnopt(FX, prox_l1(rho), ...
                                             X0(:), pnopt_opts);
    output_pnopt.time = toc(time_pnopt);
    
    % Evaluate the objective value.
    xsol_pnopt = reshape(xsol_pnopt, size(What, 1), size(Yhat, 1));
    fval_pnopt = scopt_mnlogistic_fx_eval(xsol_pnopt, What, Yhat, TN) ...
                 + rho*sum(sum(abs(xsol_pnopt)));   
    nnz_pnopt  = nnz(round(xsol_pnopt*1e10)*1e-10);
    fprintf('*************** FINAL OUTPUT OF PNOPT ******************\n');
    fprintf(' + The objective value: %6.15f\n', fval_pnopt);
    fprintf(' + CPU time: %6.4f\n', output_pnopt.time);
    fprintf(' + Number of nonzero elements of xsol: %d\n', nnz_pnopt);
    fprintf('********************************************************\n');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE EXAMPLE.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%