% EXAMPLE 1: TEST THE POISSON IMAGE RECONSTRUCTION SOLVER.
% DATE: 07.06.2013.

% Data file.
%pname = 'cameraman256';
%pname = 'house';
pname = 'barbara256';
%pname = 'lena';

% Load the data.
path0 = which('scopt_poisson_exam.m');
path  = [path0(1:end-20), 'poisson_data/'];
path2 = [path0(1:end-20), 'poisson_results/'];
data  = load([path, pname, '.mat']);

% Problem parameters.
prob.tau     = 2.5e-5;
prob.tauf    = 1.0e-5;
prob.nsize   = size(data.f);
prob.Yf      = data.y;
prob.beta    = 1e-10; 

% Print the information.
fprintf('We are solving the problem: %s.\n', pname);
fprintf('The size of the problem: (%4d, %4d).\n', prob.nsize(1), prob.nsize(2));

% Blurring operator. Comment if blurring is not considered.
prob.Aopr    = @(x)(conv2(x, data.blur, 'same'));
prob.ATopr   = @(x)(conv2(x, data.blur, 'same'));

% Uncomment the following for the denoising problem case (without
% blurring).
% scale        = 1.0;
% prob.Aopr    = @(x)(scale*x);
% prob.ATopr   = @(x)(scale*x);

% Other data components.
prob.Yf      = data.y;
prob.logeps  = 0; % Ground value.

% Gradient descent - Algorithmic parameters.
param.RelTolX        = 1e-8;
param.MaxIters       = 500;
param.SubMaxIters    = 0;
param.ubSubIters     = 30;
param.SubRelTol      = 1e-9; % For the denoising problems, this should be 1e-9.
param.RelTolFun      = 1e-6;
param.isFxEval       = true;
param.isTrace        = true;
param.TraceMode      = 2;
param.PrintLevel     = 1;
param.numFxCheck     = 3;
param.nSubFxCheck    = 1;
param.proxOperType   = 'l1';
param.isBtkLS        = false;

% Generate a starting point.
Af      = prob.Aopr(data.f);
ATy     = prob.ATopr(data.y);
xtmp    = sum(sum((prob.ATopr(ones(size(data.y))))));
prob.X0 = (sum(sum(data.y)).*numel(ATy))./(sum(sum(ATy)) .* xtmp).*ATy;
Fval0   = scopt_poisson_fx_eval(prob.X0, prob.Aopr, double(prob.Yf), ...
          prob.tau, prob.beta);

% Options for SPIRAL-TAP algorithm.
param.MaxIters = 100;
%maxiter = 100;
miniter = param.MaxIters;

% Call the SPIRAL-TV solver. See http://people.ee.duke.edu/~zth/software/.
is_existing_spiral_tap = 0;%exist('SPIRALTAP.m');
if is_existing_spiral_tap
    fprintf('+ CALL THE SPIRAL-TAP SOLVER ...\n');
    t1 = tic;
    [spiral_xsol, spiral_iter, spiral_fval, spiral_error, ...
        spiral_cputime] = SPIRALTAP(double(prob.Yf), prob.Aopr, prob.tau,...
        'penalty', 'TV', ...
        'AT', prob.ATopr,...
        'maxiter', param.MaxIters,...
        'Initialization', prob.X0,...
        'miniter', miniter,...
        'stopcriterion', 3,...
        'monotone', 1,...
        'saveobjective', 1,...
        'savereconerror', 1,...
        'savecputime', 1,...
        'tolerance', 1.0e-5, ...
        'savesolutionpath', 0,...
        'truth', data.f,...
        'verbose', 1);
    spiral_time = toc(t1);         

    % Evaluate the objective value.
    Fval2 = scopt_poisson_fx_eval(spiral_xsol, prob.Aopr, double(prob.Yf), ...
                                  prob.tau, prob.beta);

    % Compute PSNR.
    MaxF2 = max(data.f(:))^2;
    MSE2  = norm( spiral_xsol - data.f, 'fro')^2/prod(prob.nsize);
    PSNR2 = 10*log10( MaxF2 / MSE2); 
    nrmf2 = norm(data.f(:) - spiral_xsol(:), 2) / max( 1, norm(data.f(:), 2));
else
    Fval2 = -inf;
end

% Proximal Gradient/Newton method.
param.MaxIters  = 1000;
fprintf('+ CALL THE PROXIMAL-GRADIENT SOLVER ...\n');
[final_sol, final_output] = scopt_poisson_main(prob, param, 'PG', Fval2);

% Get the final solution.
Xsol = final_sol.x;

% Evaluate the objective value.
Fval1 = scopt_poisson_fx_eval(Xsol, prob.Aopr, double(data.y), prob.tau, ...
                              prob.beta);

% Compute the PSNR value.
MaxF2 = max(data.f(:))^2;
MSE1  = norm( Xsol - data.f, 'fro')^2/prod(prob.nsize);
PSNR1 = 10*log10( MaxF2 / MSE1); 
nrmf1 = norm(data.f(:) - Xsol(:), 2) / max( 1, norm(data.f(:), 2));

% Print the output information.
fprintf(' + The objective value of NGrad:  %6.8f\n', Fval1);
if is_existing_spiral_tap
    fprintf(' + The objective value of SPIRAL: %6.8f\n', Fval2);
end
fprintf(' + Computational time of NGrad:   %6.5f\n', final_output.time);
if is_existing_spiral_tap
    fprintf(' + Computational time of SPIRAL:  %6.5f\n', spiral_time);
end
fprintf(' + PSNR of the NGrad algorithm:   %6.6f dB\n', PSNR1);
if is_existing_spiral_tap
    fprintf(' + PSNR of the SPIRAL algorithm:  %6.6f dB\n', PSNR2);
end
fprintf(' + Norm(f-fsol)/norm(f) of the NGrad algorithm:   %6.6f\n', nrmf1);
if is_existing_spiral_tap
    fprintf(' + Norm(f-fsol)/norm(f) of the SPIRAL algorithm:  %6.6f\n', nrmf2);
end

% Plot the images.
figure('Name', 'NGrad');       colormap gray; imagesc(Xsol);
if is_existing_spiral_tap
    figure('Name', 'SPIRAL');  colormap gray; imagesc(spiral_xsol);
end
figure('Name', 'Noise image'); colormap gray; imagesc(prob.Yf);

% Save the data or not.
is_save_results  = true;  % Save the result in a mat file.
if is_save_results
    if is_existing_spiral_tap
        save( [path2, 'pg_x1_', num2str(prob.tau*1e5), '_', pname, '.mat'], ...
              'Fval2', 'MSE2', 'PSNR2', 'spiral_cputime', 'spiral_error', ...
              'spiral_time', 'spiral_xsol',  'spiral_iter' ,...
              'spiral_fval', ...
              'Fval1', 'MSE1', 'PSNR1', 'final_output', 'Xsol', 'prob', 'param');
    else
         save( [path2, 'pg_x1_', num2str(prob.tau*1e5), '_', pname, '.mat'], ...
               'Fval1', 'MSE1', 'PSNR1', 'final_output', 'Xsol', 'prob', 'param');
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE TEST.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%