% PURPOSE: Test the scopt_scheds_pga solver via some synthetic data.
% Date: 01.07.2013.
% By Quoc Tran Dinh, LIONS, EPFL, Lausanne, Switzerland.
%
%

% Input data.
rho = 0.0002;       % The regularization parameter (can be changed).
n   = 10000;        % Size of vector X(i), i = 1, ..., m.
m   = round(0.3*n); % Size of observation vectors.

% Problem dimension.
fprintf('+ The size of problem: m = %d, n = %d\n', m, n);
fprintf('+ Penalty parameter: %6.5f\n', rho);
fprintf('+ Generating the data ...\n');

% Synthetic data.
sigma = 0.01;
X_hat = [ones(m, 1), randn(m, n-1)]/sqrt(m);
beta  = [1; round(2*full(sprand(n-1, 1, 0.05)))];
beta  = full(beta) / norm(beta);

% Generate noise.
idd   = randn(m, 1);
idd   = sigma*idd;  
y_hat = X_hat*beta;
nrmy  = norm(y_hat);
nrmn  = norm(idd);
y_hat = y_hat + 1e-2*(nrmy/nrmn)*idd;
W_hat = []; 

% Get default parameters.
param = scopt_optimset([]);

% User-defined parameters (can be changed).
param.RelTolX     = 1e-6;
param.MaxIters    = 20000;
param.RelTolFun   = 1e-10;
param.isFxEval    = true;
param.isDebug     = true;
param.numFxCheck  = 4;
param.PrintLevel  = 1;
param.isBtkLS     = 1;

% Call the solver.
fprintf('+ CALL THE SOLVER ...\n');
[fsol, output] = scopt_scheds_pga(X_hat, y_hat, rho, [], [], param, W_hat);

% Get the final solution.
xsol    = fsol.x;
sigsol  = fsol.sig;
nrm_sol = norm(beta - xsol, 2)/max(norm(beta, 2), 1);

% Call TFOCS
param.Method  = 'N07';
param.Restart = -inf;
[xsol_tfocs, sigsol_tfocs, output_tfocs] = tfocs_exam(X_hat, y_hat, rho, param);
fval_tfocs    = output_tfocs.tfocs.f(end);
nrm_sol_tfocs = norm(beta - xsol_tfocs, 2)/max(norm(beta, 2), 1);

% Call TFOCS
param.Method   = 'GRA';
param.Restart  = -inf;
[xsol_tfocs1, sigsol_tfocs1, output_tfocs1] = tfocs_exam(X_hat, y_hat, rho, param);
fval_tfocs1    = output_tfocs1.tfocs.f(end);
nrm_sol_tfocs1 = norm(beta - xsol_tfocs1, 2)/max(norm(beta, 2), 1);

% Get outputs.
EPS = 1e-7;
fprintf(' + ProxGrad:  CPU time: %3.6f.\n', output.time);
fprintf(' + TFOCS-N07: CPU time: %3.6f.\n', output_tfocs.time);
fprintf(' + TFOCS-AT:  CPU time: %3.6f.\n', output_tfocs1.time);
fprintf(' + ProxGrad:  The objective value: %6.18f\n', output.fval);
fprintf(' + TFOCS-N07: The objective value: %6.18f\n', fval_tfocs);
fprintf(' + TFOCS-AT:  The objective value: %6.18f\n', fval_tfocs1);
fprintf(' + ProxGrad:  The value of sigma: %6.6f\n', sigsol);
fprintf(' + TFOCS-N07: The value of sigma: %6.6f\n', sigsol_tfocs);
fprintf(' + TFOCS-AT:  The value of sigma: %6.6f\n', sigsol_tfocs1);
fprintf(' + ProxGrad:  Norm(beta - betasol): %6.6f\n', nrm_sol);
fprintf(' + TFOCS-N07: Norm(beta - betasol): %6.6f\n', nrm_sol_tfocs);
fprintf(' + TFOCS-AT:  Norm(beta - betasol): %6.6f\n', nrm_sol_tfocs1);

fprintf([' + ProxGrad:  Number of nonzero elements of Xsol (rounding up to ', ...
         '1e-7) and the true solution: %d/%d\n'], ...
        nnz(round(xsol/EPS)*EPS), nnz(beta));
fprintf([' + TFOCS-N07: Number of nonzero elements of Xsol (rounding up to ', ...
         '1e-7) and the true solution: %d/%d\n'], ...
        nnz(round(xsol_tfocs/EPS)*EPS), nnz(beta));
fprintf([' + TFOCS-AT:  Number of nonzero elements of Xsol (rounding up to ', ...
         '1e-7) and the true solution: %d/%d\n\n'], ...
        nnz(round(xsol_tfocs1/EPS)*EPS), nnz(beta));
    

% save scheds.mat ...
%      param rho m n X_hat y_hat beta sigma xsol sigsol nrm_sol fsol ...
%      output xsol_tfocs sigsol_tfocs output_tfocs nrm_sol_tfocs fval_tfocs ...
%      xsol_tfocs1 sigsol_tfocs1 output_tfocs1 nrm_sol_tfocs1 fval_tfocs1;
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE EXAMPLE.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%