%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%        -------------------------------------------------------- 
%%%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%%%                          Version 1.0.0 (2013)  
%%%        ---------------------------0*0*0------------------------
%%%
%%%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%%%       algorithms for solving Sparse Inverse Covariance Estimation 
%%%                 Problems in Graphical Model Selection.
%%%
%%% INFORMATION:
%%%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%%%       EPFL, Lausanne, Switzerland.
%%%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%%%    Date: 26.11.2012.
%%%    Last modified: 27.05.2013.
%%%    Contact: quoc.trandinh@epfl.ch, volkan.cevher@epfl.ch
%%% 
%%% REFERENCES:
%%%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%%%    framework for composite minimization: Graph learning without 
%%%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%%%    USA.
%%%
%%% FUNCTION: [DX, lambda, output, Xsol] = scopt_glearn_blcp(X_cur, Shat, ...
%%%                                        rho, X0, oiter, old_lbd, ...
%%%                                        param, varargin) 
%%%
%%% PURPOSE: 
%%%     Solve the primal subproblem to compute the proximal-Newton
%%%     direction by using an active set method.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX, lambda, output, Xsol] = scopt_glearn_blcp(X_cur, Shat, ...
                                      rho, X0, oiter, old_lbd, ...
                                      param, varargin) 

output.msg      = 'start';
output.iter     = 0;
output.prox_cnt = 0;
output.time     = tic;
p               = size(Shat, 1);
QuadRgRadius    = 0.75*0.219224;

% Set the maximum number of iterations.
MaxIters = param.SubMaxIters;
if param.SubMaxIters == 0 || param.SubMaxIters == inf
    MaxIters = min(5*oiter + 1, param.ubSubIters);
end

% Compute the Cholesky decomposition of matrix X_cur and W_cur =
% inv(X_cur).
[L_cur, ret] = chol(X_cur);
if ret ~= 0
    %error( 'Cholesky decomposition of Xk is failed!' );
    [V_cur, D_cur] = eig(X_cur);
    Dp_cur         = max( real(diag(D_cur)), 1 - 0.99);
    X_cur          = V_cur*diag(Dp_cur)*V_cur';
    L_cur          = V_cur*diag(sqrt(Dp_cur));
end
Ip                 = eye(size(Shat));
invL_cur           = L_cur\Ip;
W_cur              = invL_cur*invL_cur';
output.chol_cnt    = 1;

% Compute b = (R'*x_k - invR*grad).
grad = Shat - W_cur;
b    = invL_cur'*X_cur*invL_cur - L_cur*grad*L_cur';
b    = b(:);

% Using BLPC algorithm, we want to solve: 
%               min 0.5||Rx - b||_2^2 + lambda||x||_1.
%
% Step 1 - Initialization.
F_plus  = [];
F_minus = [];
H       = 1:p^2;

% The primal variable is initialized at zero.
beta_cur = zeros(p^2, 1);

% Initialize the dual variable (with warm-start).
Xy       = invL_cur*reshape(b, size(Shat))*invL_cur';
Xy       = Xy(:);

if( oiter == 1 )
    d_cur = -Xy;
else
    d_cur = X0(:);
end

k    = MaxIters;
t    = p^2 + 1;
init = 1;
iter = 0;

% Step 3: The main loop.
while (init || ...
    scopt_glearn_blcp_feas_check(d_cur, beta_cur, rho, H, F_plus, F_minus))
    
    % Number of iterations.
    iter            = iter + 1;
    init            = 0;
    output.prox_cnt = output.prox_cnt + 1; 
    
    % Step 4: Compute the infeasible set.
    J1 = H(d_cur(H) > rho);
    J2 = H(d_cur(H) < -rho);
    J3 = F_plus(beta_cur(F_plus) < 0);
    J4 = F_minus(beta_cur(F_minus) > 0);
    
    G = union(union(J1, J2), union(J3, J4));
    
    % Steps 5: Take the subset of G to update the active set.
    if (oiter == 1)
        if (length(G) < t)
            t     = length(G);
            k     = MaxIters;
            G_hat = G;   
        elseif (length(G) >= t && k >= 1)
            k     = k - 1;
            G_hat = G;
        elseif (length(G) >= t && k == 0)
            G_hat = max(G);
        end;
    else
        G_hat = G;
    end
    
    % Step 6: Compute the infeasible subsets for updating the active set.
    J1_hat = intersect(G_hat, J1);
    J2_hat = intersect(G_hat, J2);
    J3_hat = intersect(G_hat, J3);
    J4_hat = intersect(G_hat, J4);
    
    % Update the active set (and inactive set)
    H       = union( setdiff(H, union(J1_hat, J2_hat)), union(J3_hat, J4_hat) );
    F_plus  = union( setdiff(F_plus, J3_hat), J1_hat );
    F_minus = union( setdiff(F_minus, J4_hat), J2_hat );
    
    % Step 7a: Set the known variables to their values.
    beta_cur(H)     = 0;
    d_cur(F_plus)   = rho;
    d_cur(F_minus)  = -rho;
    
    % Step 7b: Recompute the rest of d and beta.
    F           = union(F_plus, F_minus);
    beta_cur(F) = cgsolve(@(x)op_A(x,F,F,W_cur,size(Shat,1)), ...
                  (Xy(F) - d_cur(F)) ,1e-6,100,0);
    %beta_cur2(F) = cgsolve(@(x)op_A3(x,F,F,W_cur,size(Shat,1)), ...
    %               (Xy(F) - d_cur(F)) ,1e-6,100,0);
    %beta_cur(F) = RFRFT\(Xy(F) - d_cur(F));
    d_cur(H)    = Xy(H) - op_A(beta_cur(F),F,H,W_cur,size(Shat,1));           
    %d_cur2(H)    = Xy(H) - op_A(beta_cur2(F),F,H,W_cur,size(Shat,1));    
    
    if iter > MaxIters, break; end
end
% End of the main loop.

% Compute the proximal-Newton search direction.
Xsol = reshape(beta_cur, p, p); 
DX   = Xsol - X_cur;

% Compute lambda = sqrt(trace((wzw - W_cur)*DX)).
lambda = nan;
if ~strcmpi(param.lsMode, 'BTK') || ~param.isBtkLS
    if ~param.TerminateType == 1 || old_lbd > QuadRgRadius
        WDX    = W_cur*DX;
        lambda = sqrt( sum(sum(WDX'.*WDX)) );
    end
end

output.error = 0;
output.time  = toc(output.time);
output.iter  = iter;
output.Lips  = NaN;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: b = op_A(x,I,H,iTheta,d)
%%% PURPOSE: Applying A operator.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function b = op_A(x,I,H,iTheta,d)

X    = zeros(d,d);
X(I) = x; 
M    = iTheta*X*iTheta;
b    = reshape(M(H),[],1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: b = op_A2(x,F,H,iTheta,d)
%%% PURPOSE: Applying A2 operator.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function b = op_A2(x,F,H,iTheta,d)

R = zeros(d^2, length(F));
for i = 1:length(F)
    col     = rem(F(i), d) + 1;
    kcol    = rem(F(i), d) + 1;
    R(:, i) = kron(iTheta(:, kcol), iTheta(:, col));    
end;

if (issame(F, H))
    Rt = R;
else
    Rt = zeros(d^2, length(H)); 
    for i = 1:length(H)
        col      = rem(H(i), d) + 1;
        kcol     = rem(H(i), d) + 1;
        Rt(:, i) = kron(iTheta(:, kcol), iTheta(:, col));    
    end;
end

b = Rt'*R*x;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: b = op_A3(x,F,H,iTheta,d)
%%% PURPOSE: Applying A3 operator.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function b = op_A3(x,F,H,iTheta,d)

A      = zeros(d, d);
X      = zeros(d, d);
X(F)   = x;
[I, J] = ind2sub(size(iTheta), F);

for ii = 1:length(I)
    i = I(ii); j = J(ii);
    for jj = 1:length(I)
        l       = I(jj); k = J(jj);
        A(i, j) = iTheta(i, l)*X(l, k)*iTheta(k, j);
    end;
end;

A = A(:);
b = A(H);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
