% FUNCTION: [DX, Usol, lambda, output] = scopt_glearn_coord_descent(...
%                                        X_cur, Shat, rho, U0, ...
%                                        oiter, old_lbd, param, varargin)
% PURPOSE: Perform the block-coordinate descent algorithm for solving the
%          subproblem in computing the proximal Newton direction.
% NOTE:    This method bases on the QUIC solver.
% 
% INFOMATION: 
%   Create by Quoc Tran Dinh, LIONS, EPFL, Lausanne, Switzerland.
%   Date 05.07.2013.
%

function [DX, Usol, lambda, output] = scopt_glearn_coord_descent(X_cur, ...
                                      Shat, rho, U0, oiter, old_lbd, ...
                                      param, varargin)
                                  
% Set the maximum number of iterations.
MaxIters = param.SubMaxIters;
if param.SubMaxIters == 0 || param.SubMaxIters == inf
    MaxIters = min(5*oiter + 1, param.ubSubIters);
end

% Initialization.
p            = size(Shat, 1);
normD        = 0.0;
diffD        = 0.0;
U_cur        = U0;
output.time  = tic;
output.msg   = 'start';
QuadRgRadius = 0.75*0.219224;

% Compute the Cholesky decomposition of matrix X_cur and W_cur =
% inv(X_cur).
[L_cur, ret] = chol(X_cur);
if ret ~= 0
    %error( 'Cholesky decomposition of Xk is failed!' );
    [V_cur, D_cur] = eig(X_cur);
    Dp_cur         = max( real(diag(D_cur)), 1 - 0.99);
    X_cur          = V_cur*diag(Dp_cur)*V_cur';
    L_cur          = V_cur*diag(sqrt(Dp_cur));
end
Ip                 = eye(size(Shat));
invL_cur           = L_cur\Ip;
W_cur              = invL_cur*invL_cur';
output.chol_cnt    = 1;

% The first iteration.
if oiter == 1 && ~(nnz(X_cur) - nnz(diag(X_cur))) % check if diagonal.

    if param.PrintLevel >= 3
        fprintf('Newton iteration 1.\n');
        fprintf('X is a diagonal matrix.\n');
    end
    D_cur              = zeros(p, p);
    [output.fX, D_cur] = diagonalNewton(Shat, rho, X_cur, W_cur, D_cur);

    % Fill in the full matrix.
    D_cur       = tril(D_cur) + tril(D_cur,-1)';
    cdSweep     = 1;
    output.msg  = 'Convergence achieved';

% Other iterations.
else

    % Initialize 
    numActive = 0;
    U_cur     = zeros(p, p);
    D_cur     = zeros(p,p);
    subgrad   = 0.0;

    % Compute the active set.
    for i = 1:p
        for j = 1:i

            % Compute the gradient vector.
            g = Shat(i,j) - W_cur(i,j);

            % Generate the active set.
            if X_cur(i,j) ~= 0.0 || abs(g) > rho

                activeSet(numActive+1, 1).i = i;
                activeSet(numActive+1, 1).j = j;
                numActive                   = numActive + 1;

                if X_cur(i,j) > 0.0
                    g = g + rho;
                elseif X_cur(i,j) < 0.0
                    g = g - rho;
                else
                    g = abs(g) - rho;
                end

                subgrad = subgrad + abs(g);
            end
        end
    end

    % Print the messages.
    if param.PrintLevel >= 3
        fprintf('Newton iteration %d.\n', NewtonIter);
        fprintf('Active set size %d.\n', numActive);
        fprintf('Sub-gradident = %6.10f, l1-norm of X = %6.10f.\n', ...
                 subgrad, l1normX);
    end

    % Solve the subproblem by using coordinate descent method.
    for cdSweep = 1 : MaxIters

        % Initialize diffD.
        diffD = 0.0;

        % Change the active set by some random exchanges.
        for i = 1:numActive

            % Generate a random coordinate.
            j                = i + randi([0, max(numActive - i-1, 0)], 1);

            % Exchange the coordinate.
            k1               = activeSet(i,1).i;
            k2               = activeSet(i,1).j;
            activeSet(i,1).i = activeSet(j,1).i;
            activeSet(i,1).j = activeSet(j,1).j;
            activeSet(j,1).i = k1;
            activeSet(j,1).j = k2;
        end

        % For each coordinate, perform the coordinate descent step.
        for l = 1:numActive
            i = activeSet(l,1).i;
            j = activeSet(l,1).j;
            [U_cur, D_cur, normD, diffD] = coordinateDescentUpdate(Shat, ...
                                           rho, X_cur, W_cur, U_cur, ...
                                           D_cur, i, j, normD, diffD);
        end

        % Print the message in the coordinate descent algorithm.
        if param.PrintLevel >= 3
            fprintf(['Coordinate descent sweep %d. Norm of ', ...
                     'D = %6.10f, change in D = %6.10f'], cdSweep, ...
                      normD, diffD);
        end

        % Terminate condition.
        if diffD <= normD*param.bcdsSubTol
            output.msg = 'Convergence achieved';
            break;
        end

    end
    % End of the coordinate descent method.

end
% End of the if statement for computing the search direction.   

% Fill in the full metrix.
Usol   = tril(U_cur) + tril(U_cur,-1)';
DX     = tril(D_cur) + tril(D_cur,-1)';
    
% Compute lambda = sqrt(trace((wzw - W_cur)*DX)).
lambda = nan;
if ~strcmpi(param.lsMode, 'BTK') || ~param.isBtkLS
    if ~param.TerminateType == 1 || old_lbd > QuadRgRadius
        WDX    = W_cur*DX;
        lambda = sqrt( sum(sum(WDX'.*WDX)) );
    end
end

% Assign the output.
output.time     = toc(output.time);
output.error    = diffD/max(1, normD);
output.iter     = cdSweep;
output.Lips     = nan;
output.L_cur    = L_cur;
output.prox_cnt = nan;

% If required, compute the dual norm of the gradient of f.
if param.isGradNorm
    SX                 = Shat*X_cur;
    output.dualGradNrm = sqrt( sum(sum(SX'.*SX)) - 2*sum(diag(SX)) + p );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [U_out, D_out, normD_out, diffD_out] = coordinateDescentUpdate(
%%%                                                Shat, rho, X_cur, ...
%%%                                                W_cur, U_cur, D_cur, ...
%%%                                                i, j, normD, diffD)
%%% PURPOSE: Update the coordinate of the subproblem.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [U_out, D_out, normD_out, diffD_out] = coordinateDescentUpdate(...
                                                Shat, rho, X_cur, ...
                                                W_cur, U_cur, D_cur, ...
                                                i, j, normD, diffD)
% Compute the quantity a.
a = W_cur(i, j)*W_cur(i,j);
if(i ~= j)
   a = a + W_cur(i,i)*W_cur(j,j); 
end

% Compute the quantity b.
inva  = 1.0/a;
b     = Shat(i,j) - W_cur(i,j);
b     = b + sum( W_cur(i, :)*U_cur(:, j) );
l     = rho*inva;
c     = X_cur(i,j) + D_cur(i,j);
f     = b*inva;
normD = normD - abs(D_cur(i,j));

if( c > f )
    mu = -f - l;
    if (c + mu < 0.0)
        mu = -c;
        D_cur(i,j) = -X_cur(i,j);
    else
        D_cur(i,j) = D_cur(i,j) + mu;
    end
else
    mu = -f + l;
    if (c + mu > 0.0)
        mu = -c;
        D_cur(i,j) = -X_cur(i,j);
    else
        D_cur(i,j) = D_cur(i,j) + mu;
    end
end

diffD = diffD + abs(mu);
normD = normD + abs(D_cur(i,j));

if (mu ~= 0.0)
    U_cur(i,:) = U_cur(i,:)  + mu*W_cur(j,:);
    
    if (i ~= j)
        U_cur(j,:) = U_cur(j,:)  + mu*W_cur(i,:);
    end
end

% Assign the output.
diffD_out = diffD;
normD_out = normD;
U_out     = U_cur;
D_out     = D_cur;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [fX, D_out] = diagonalNewton(Shat, rho, X_cur, W_cur, ...
%%%                         D_cur)
%%% PURPOSE: Diagonal Newton direction.
%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [fX, D_out] = diagonalNewton(Shat, rho, X_cur, W_cur, D_cur)

p = size(Shat, 1);

for i = 1:p
    for j = 1:i-1
        a    = W_cur(i,i)*W_cur(j,j);
        inva = 1.0/a;
        b    = Shat(i,j);

        l    = rho*inva;
        f    = b*inva;
        x    = -b*inva;
        
        if (f < 0 )
            mu = -f - l;
            x  = x - l;
            
            if( mu < 0.0 )
                mu         = 0.0;
                D_cur(i,j) = -X_cur(i,j);
            else
                D_cur(i,j) = D_cur(i,j) + mu;
            end
        else
            mu = -f + l;
        
            if (mu > 0.0)
                mu         = 0.0;
                D_cur(i,j) = -X_cur(i,j);
            else
                D_cur(i,j) = D_cur(i,j) + mu;
            end
            
        end
    end
end

logDet  = 0.0;
l1normX = 0.0;
trSX    = 0.0;

for i = 1:p
    logDet  = logDet  + log(X_cur(i,i));
    l1normX = l1normX + rho*abs(X_cur(i,i));
    trSX    = trSX    + Shat(i,i)*X_cur(i,i);
    
    a       = W_cur(i,i)*W_cur(i,i);
    inva    = 1.0/a;
    b       = Shat(i,i) - W_cur(i,i);
    l       = rho*inva;
    c       = X_cur(i,i);
    f       = b*inva;
    
    if(c > f)
        mu = -f - l;
        if( c + mu < 0.0 )
            D_cur(i,i) = -X_cur(i,i);
            continue;
        end
    else
        mu = -f + l;
        if( c + mu > 0.0 )
            D_cur(i,i) = -X_cur(i,i);
            continue;
        end
    end
    D_cur(i,i) = D_cur(i,i) + mu;
end

% Evaluate the objective value.
fX    = -logDet + trSX + l1normX;
D_out = tril(D_cur) + tril(D_cur, -1)'; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
