%
%        -------------------------------------------------------- 
%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%                          Version 1.0.0 (2013)  
%        ---------------------------0*0*0------------------------
%
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%       algorithms for solving Sparse Inverse Covariance Estimation 
%                 Problems in Graphical Model Selection.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 26.11.2012.
%    Last modified: 27.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%    framework for composite minimization: Graph learning without 
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%
% FUNCTION: [DX, Ysol, output] = scopt_glearn_dFISTA(X_cur, Shat, rho, ...
%                                Y0, oiter, lbd_old, param, varargin)
% PURPOSE: 
%     Solving the primal subproblem to compute the proximal-Newton
%     direction by using the dual approach.
%     We transform the primal subproblem into the dual problem and
%     applying the fast proximal-gradient method (Nesterov2007, Beck2009)
%     to solve the dual problem. Then we recover the primal solution.
%

function [DX, Ysol, lambda, output] = scopt_glearn_dFISTA(X_cur, Shat, ...
                                      rho, Y0, oiter, old_lbd, ...
                                      param, varargin)

% Initialize the output.
output.msg      = 'start';
output.iter     = 0;
output.prox_cnt = 0;
output.time     = tic;
fx_cur          = inf;
QuadRgRadius    = 0.75*0.219224;

% Set the maximum number of iterations.
MaxIters = param.SubMaxIters;
if param.SubMaxIters == 0 || param.SubMaxIters == inf
    MaxIters = min(5*oiter + 1, param.ubSubIters);
end

% Evaluate the Lipschitz constant of the Hessian matrix at X_cur.
l2nrmx = scopt_l2normx(X_cur, param.PwMaxIters);
Lipsx  = l2nrmx*l2nrmx;

% Evaluate the strong convexity parameter if required.
if param.isEvalStrCvx
    mux = min(eig(X_cur));
    mux = mux*mux;
end

% Initialization.
y_cur = Y0;   % Starting point (for warm-start).
u_cur = y_cur;
t_cur = 1.0;
p     = size(Shat, 1);

% Compute the fixed term of the gradient.
SX    = Shat*X_cur;
Qt    = X_cur*SX - 2.0*X_cur;
rhoQt = Qt/rho;

% The main loop.
for iter = 1:MaxIters
    
    % Compute the gradient vector.
    grad_f = scopt_glearn_dqp_eval(y_cur, X_cur, rhoQt, true);
    
    % Evalute the objective value if required.
    if strcmpi(param.isSubRestart, 'fx')
        fx_old = fx_cur;
        fx_cur = scopt_glearn_dqp_eval(u_cur, X_cur, rhoQt);
    end
    
    % Perform the projected gradient step.
    u_next = scopt_clipping_oper( y_cur - (1/Lipsx)*grad_f, 1 );
    
    % Count the number of prox operations.
    output.prox_cnt = output.prox_cnt + 1;
    
    % Compute the relative error.
    du     = u_next - u_cur;
    error  = sqrt(sum(sum(du.*du)))/max(1.0, sqrt(sum(sum(u_cur.*u_cur))));
    
    % Check the stopping criterion.
    if error <= param.SubRelTol
        output.msg = 'Convergence achieved!';
        
        % Keep the last iteration.
        u_cur = u_next;
        break;        
    end;
    
    % Compute new step size.
    t_next = 0.5*(1 + sqrt(1 + 4*t_cur*t_cur));
    if param.isEvalStrCvx
        fact   = (1 - sqrt(mux/Lipsx))/(1 + sqrt(mux/Lipsx));
    else
    	fact   = (t_cur - 1) / t_next;
    end
    
    % Update y_next
    y_next = u_next + fact*du;
    
    % Reassign for the next iteration.
    t_cur = t_next;
    u_cur = u_next;
    y_cur = y_next;
    
    % Perform restart strategy.
    if strcmpi(param.isSubRestart, 'fixed')
        if mod(iter, param.nSubRsIters) == 0
            t_cur = 1;
            y_cur = u_cur;
        end
    elseif strcmpi(param.isSubRestart, 'fx')
        if iter > 1 && fx_cur > fx_old
            t_cur = 1;
            y_cur = u_cur;
        end
    end
end 
% End of the main loop.

if iter >= param.SubMaxIters
    output.msg = 'Exceed the maximum number of iterations';
end

% Form the dual solution.
Ysol   = u_cur;

% Form the primal proximal-Newton direction.
YX     = Ysol*X_cur;
XYX    = X_cur*YX;
DX     = -rho*XYX - Qt - X_cur;

% Compute lambda - the norm of the proximal-Newton decrement.
lambda = nan;
if ~strcmpi(param.lsMode, 'BTK') || ~param.isBtkLS
    if ~param.TerminateType == 1 || old_lbd > QuadRgRadius
        Stmp   = SX +  rho*YX;
        lambda = sqrt(abs(p - 2*sum(diag(Stmp)) + sum(sum(Stmp'.*Stmp))));
    end
end

% Assign the output 
output.time  = toc(output.time);
output.error = error;
output.iter  = iter;
output.Lips  = Lipsx;

% Compute the dual norm of gradf.
if param.isGradNorm
    output.dualGradNrm = sqrt( sum(sum(SX'.*SX)) - 2*sum(diag(SX)) + p );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%