% FUNCTION: [X_new, fval, alpha_new, output] = scopt_glearn_ls(X_cur, ...
%                                              DX, alphaj, fval, Shat, ...
%                                              rho, param)
% PURPOSE:
%  This procedure performs a line-search procedure: Looking-forward or 
%  back-tracking mode.
%
% INFORMATION:
%   By Quoc Tran Dinh, LIONS, EPFL, Switzerland.
%   Date: 09.07.2013.
%

function [X_new, fval, alpha_new, output] = scopt_glearn_ls(X_cur, DX, ...
                                            alphaj, fval, Shat, rho, param)

% Count the number of Cholesky decomposition.
output.chol_cnt = 0;

% Update the new point with the safe step size.
if alphaj < 1.0 
    X_new = X_cur + alphaj*DX;
else
    X_new = X_cur + DX;
end

% Perform a looking-forward-procedure if required.
if param.isBtkLS && strcmpi(param.lsMode, 'FWLK')

    lsiter    = 1; 
    alpha_new = alphaj;
    if alphaj < 1.0
        while( alpha_new <= 1.0 )

            % Update the new step size by increasing it with a
            % factor 2.
            alpha_new    = min(lsiter*alphaj, 1.0);
            lsiter       = param.btkSlop*lsiter;
            X_old        = X_new;

            % Update the trial point
            X_new        = X_cur + alpha_new*DX;

            % Evaluate the objective value at the new point.
            [L_new, ret] = chol(X_new);
            if ret == 0
                fval_new = scopt_glearn_fx_eval(X_new, Shat, rho, L_new);
            else
                fval_new = inf;
            end
            output.chol_cnt = output.chol_cnt + 1;

            % If the objective value is improved then move on.
            if (fval_new < fval) && (fval_new ~= inf) 
                if param.PrintLevel >= 2
                    fprintf('Accepted step-size: %6.6f\n', alpha_new);
                end
                fval = fval_new;

            % Otherwise, get back to the safe value.
            else
                X_new = X_old;
                break;
            end
        end % End of while.
    end
    
% Perform the normal back-tracking line-search procedure.    
elseif param.isBtkLS && strcmpi(param.lsMode, 'BTK')
    
    % Compute the trial point X_new.
    alpha_new = 1.0;
    X_new     = X_cur + alpha_new*DX;
    
    % Evaluate the objective value at the new point.
    [L_new, ret] = chol(X_new);
    if ret == 0
        fval_new = scopt_glearn_fx_eval(X_new, Shat, rho, L_new);
    else
        fval_new = inf;
    end
    output.chol_cnt = output.chol_cnt + 1;
    
    % For loop to check if the objective function is decreasing.
    for lsiter = 1:param.BtkMaxIters
        
        % If the step-size is at the lower bound, then terminate.
        if alpha_new <= param.minStepSize
            alpha_new = param.minStepSize;
            if param.PrintLevel >= 2
                fprintf('Lower bound step size: %6.6f.\n', alpha_new);
            end
            break;
        end
        
        % If X_new is not in the SDP cone or the objective value is not
        % improved then decrease alpha_new.
        if fval_new == inf || fval_new > fval
            
            % The trial step size.
            if param.PrintLevel >= 2
                fprintf('Trial step size: %6.6f.\n', alpha_new);
            end
            
            % Decrease alpha_new.
            alpha_new = alpha_new/param.btkSlop;
            
            % The trial point.
            X_new     = X_cur + alpha_new*DX;
    
            % Evaluate the objective value at the new point.
            [L_new, ret] = chol(X_new);
            if ret == 0
                fval_new = scopt_glearn_fx_eval(X_new, Shat, rho, L_new);
            else
                fval_new = inf;
            end
            output.chol_cnt = output.chol_cnt + 1;
            
        else
            % Keep the objective value and terminate.
            if param.PrintLevel >= 2
                fprintf('Accepted step size: %6.6f.\n', alpha_new);
            end
            fval = fval_new;
            break;
        end
    end
    
% Perform the enhancement back-tracking line-search procedure
elseif param.isBtkLS && strcmpi(param.lsMode, 'EBTK')
    
    % Compute the trial point X_new.
    alpha_new = 1.0;
    X_new     = X_cur + alpha_new*DX;
    
    % Evaluate the objective value at the new point.
    [L_new, ret] = chol(X_new);
    if ret == 0
        fval_new = scopt_glearn_fx_eval(X_new, Shat, rho, L_new);
    else
        fval_new = inf;
    end
    output.chol_cnt = output.chol_cnt + 1;
    
    % For loop to check if the objective function is decreasing.
    for lsiter = 1:param.BtkMaxIters
        
        % If the step-size is at alphaj, then terminate.
        if alpha_new <= alphaj 
            alpha_new = alphaj;
            if param.PrintLevel >= 2
                fprintf('Safe step size: %6.6f.\n', alpha_new);
            end
            break;
        end
        
        % If X_new is not in the SDP cone or the objective value is not
        % improved then decrease alpha_new.
        if fval_new == inf || fval_new > fval
            
            % The trial step size.
            if param.PrintLevel >= 2
                fprintf('Trial step size: %6.6f.\n', alpha_new);
            end
            
            % Decrease alpha_new until meeting the safe value alphaj.
            alpha_new = alpha_new/param.btkSlop;
            if alpha_new < alphaj, alpha_new = alphaj; end
            
            % The trial point.
            X_new     = X_cur + alpha_new*DX;
    
            % Evaluate the objective value at the new point.
            [L_new, ret] = chol(X_new);
            if ret == 0
                fval_new = scopt_glearn_fx_eval(X_new, Shat, rho, L_new);
            else
                fval_new = inf;
            end
            output.chol_cnt = output.chol_cnt + 1;
            
        else
            % Keep the objective value and terminate.
            if param.PrintLevel >= 2
                fprintf('Accepted step size: %6.6f.\n', alpha_new);
            end
            fval = fval_new;
            break;
        end
    end
    
% No line-search procedure is needed.    
else
    alpha_new = alphaj;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%