%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%        -------------------------------------------------------- 
%%%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%%%                          Version 1.0.0 (2013)  
%%%        ---------------------------0*0*0------------------------
%%%
%%%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%%%       algorithms for solving Sparse Inverse Covariance Estimation 
%%%                 Problems in Graphical Model Selection.
%%%
%%% INFORMATION:
%%%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%%%       EPFL, Lausanne, Switzerland.
%%%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%%%    Date: 26.11.2012.
%%%    Last modified: 27.05.2013.
%%%    Contact: quoc.trandinh@epfl.ch, volkan.cevher@epfl.ch
%%% 
%%% REFERENCES:
%%%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%%%    framework for composite minimization: Graph learning without 
%%%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%%%    USA.
%%%
%%% FUNCTION: [DX, X0, lambda, fval, output] = compute_glearn_ntdir(...
%%%                                            X_cur, rho, Shat, X0, Y0, ...
%%%                                            param, lambda, outiter, ...
%%%                                            varargin) 
%%%
%%% PURPOSE: Compute the proximal Newton direction and proximal-Newton
%%%          decrement.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX, lambda, fval, X0, Y0, output] = scopt_glearn_ntdir(X_cur, ...
                                              Shat, rho, X0, Y0, param, ...
                                              lambda, outiter, varargin)
 
 % Check the param.QuadCvRg parameter.
 SIGMA = 0.219224;
 if ~isfield(param, 'QuadCvRg')
    param.QuadCvRg = 0.75*SIGMA;
 end
 
 % Increase the accuracy near the solution.
 if lambda <= param.QuadCvRg
     param.SubMaxIters  = param.ubSubIters;
     param.SubRelTol    = param.Epsilon;
 end

 % Solve the Primal subproblem directly by FISTA.
 if strcmpi(param.SubMethod, 'pFISTA') 
  
    % Call the FISTA solver to solve the primal subproblem.
    [DX, X0, lambda, output] = scopt_glearn_pFISTA(X_cur, Shat, ...
                               rho, X0, outiter, lambda, ...
                               param, varargin{:}); 

 % Solve the primal problem with primal-dual subproblems.
 elseif strcmpi(param.SubMethod, 'dpFISTA') 
     
    % If lambda > SIGMA then perform the dual method.
    if lambda > 0.75*SIGMA
    
        % Call the FISTA solver to solve the dual subproblem.
        [DX, Y0, lambda, output] = scopt_glearn_dFISTA(X_cur, Shat, ...
                                   rho, Y0, outiter, lambda, ...
                                   param, varargin{:});
        
    % When it is in the quadratic region, switch to primal method.
    else
        % Call the FISTA solver to solve the primal subproblem.
        [DX, X0, lambda, output] = scopt_glearn_pFISTA(X_cur, Shat, ...
                                   rho, X0, outiter, lambda, ...
                                   param, varargin{:}); 
    end

 % Solve the dual subproblem by FISTA.
 elseif strcmpi(param.SubMethod, 'dFISTA')
     
    % Call the FISTA solver to solve the dual subproblem.
    [DX, Y0, lambda, output] = scopt_glearn_dFISTA(X_cur, Shat, ...
                               rho, Y0, outiter, lambda, ...
                               param, varargin{:});

% Block-active set method.
 elseif (strcmpi(param.SubMethod, 'BLCP'))            
    [DX, lambda, output, X0] = scopt_glearn_blcp(X_cur, Shat, rho, ...
                               X0, outiter, lambda, param, varargin{:});
 
 % Primal-dual active set method.    
 elseif (strcmpi(param.SubMethod, 'pdAS'))            
    [DX, lambda, output, X0] = scopt_glearn_pdAS(X_cur, Shat, rho, ...
                               X0, outiter, lambda, param, varargin{:});
 
% Block-coordinate descent.
 elseif (strcmpi(param.SubMethod, 'BCDS'))
    [DX, X0, lambda, output] = scopt_glearn_coord_descent(X_cur, Shat, ...
                               rho, X0, outiter, lambda, param, varargin{:});
     
 % User-defined solver.
 elseif (strcmpi(param.SubMethod, 'user-defined'))
     
     [DX, X0, lambda, output] = scopt_glearn_subsolver(X_cur, Shat, ...
                                rho, X0, outiter, lambda, ...
                                param, varargin{:}); 
                            
 % No subsolver is provided.
 else 
     error('Select a submethod');
 end
 
 % Evaluate the objective value if required.
 fval = nan;
 if param.isFxEval
     if isfield(output, 'L_cur'), L_cur = output.L_cur; else L_cur = []; end
     fval = scopt_glearn_fx_eval(X_cur, Shat, rho, L_cur);
 end
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
