%
%        -------------------------------------------------------- 
%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%                          Version 1.0.0 (2013)  
%        ---------------------------0*0*0------------------------
%
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%       algorithms for solving Sparse Inverse Covariance Estimation 
%                 Problems in Graphical Model Selection.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 26.11.2012.
%    Last modified: 27.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%    framework for composite minimization: Graph learning without 
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%
% FUNCTION: [DX, Wsol, output] = scopt_glearn_pFISTA(X_cur, Shat, ...
%                                W0, oiter, old_lbd, param, varargin)
%
% PURPOSE: 
%    Solving the primal subproblem to compute the proximal-Newton
%    direction by using the proximal-gradient method.
%    We applying the fast proximal-gradient method (Nesterov2007, Beck2009)
%    to solve directly the primal problem. 
%
% INPUTS:  
%          X_cur  : the current iteration vector.
%          Shat   : The inverse of the covariance matrix.
%          rho    : the penalty parameter.
%          W0     : The initial guess for the dual problem (hot start).
%          oiter  : The outer iteration counter.
%          old_lbd: The previous proximal-Newton decrement.
%          param  : optional parameter (structure).
%
% OUTPUTS: 
%          DX     : the approximate proximal Newton direction.
%          Wsol   : the approxinate dual solution.
%          lambda : the proximal-Newton decrement.
%          output : the output information (structure).
%

function [DX, Wsol, lambda, output] = scopt_glearn_pFISTA(X_cur, Shat, ...
                                      rho, W0, oiter, old_lbd, ...
                                      param, varargin)

% Initialize the output.
output.msg      = 'start';
output.time     = tic;
output.prox_cnt = 0;
fx_cur          = inf;
QuadRgRadius    = 0.219224;

% Set the maximum number of iterations.
MaxIters = param.SubMaxIters;
if param.SubMaxIters == 0 || param.SubMaxIters == inf
    MaxIters = min(30*oiter + 1, param.ubSubIters);
    if old_lbd <= QuadRgRadius
        MaxIters = 2000; %min(30*oiter + 1, param.ubSubIters);
    end
end

% Compute the Cholesky decomposition of matrix X_cur and W_cur = inv(X_cur).
[L_cur, ret] = chol(X_cur);
if ret ~= 0
    warning( 'Cholesky decomposition of Xk is failed! Regularize it ...' );
    [V_cur, D_cur] = eig(X_cur);
    Dp_cur         = max( real(diag(D_cur)), 1 - 0.99);
    X_cur          = V_cur*diag(Dp_cur)*V_cur';
    L_cur          = V_cur*diag(sqrt(Dp_cur));
end
Ip                 = eye(size(Shat));
invL_cur           = L_cur\Ip;
W_cur              = invL_cur*invL_cur';
W_cur              = 0.5*(W_cur + W_cur');
output.chol_cnt    = 1;
    
% Evaluate a Lipschitz constant if necessary.
l2normw = scopt_l2normx(W_cur, param.PwMaxIters);
Lipsw   = l2normw*l2normw;

% Evaluate the strong convexity parameter if required.
muw = 0;
if param.isEvalStrCvx
    muw = scopt_l2normx(X_cur, param.PwMaxIters);
    muw = 1/muw;
    muw = muw*muw;
    if muw >= Lipsw, muw = 0; end
end

% Initialization.
z_cur = W0; 
w_cur = z_cur;
t_cur = 1.0;
p     = size(Shat, 1);

% The main loop.
for iter = 1 : MaxIters
    
    % Compute the gradient vector.
    grad_f = scopt_glearn_qp_eval(z_cur, X_cur, W_cur, Shat, true);
    
    % Evaluate the objective value at w_cur for the restart.
    if strcmpi(param.isSubRestart, 'fx')
        fx_old = fx_cur;
        fx_cur = scopt_glearn_qp_eval(w_cur, X_cur, W_cur, Shat);
        fx_cur = fx_cur + rho*sum(sum(abs(w_cur)));
    end
    
    % Perform the gradient step.
    w_next = z_cur - (1/Lipsw)*grad_f;
    
    % Soft-thresholding step.
    w_next = scopt_soft_threshold_oper(w_next, rho/Lipsw);
    
    % Count the number of prox operators.
    output.prox_cnt = output.prox_cnt + 1;
    
    % Compute the relative error. 
    dw     = w_next - w_cur;
    err    = sqrt(sum(sum(dw.*dw)))/max(1.0, sqrt(sum(sum(w_cur.*w_cur))));
    
    % Check the stopping criterion.
    if err <= param.SubRelTol
        output.msg = 'Convergence achieved!';
        
        % Keep the last iteration before leaving.
        w_cur = w_next;
        break;        
    end;
    
    % Compute new step size.
    t_next = 0.5*(1 + sqrt(1 + 4*t_cur*t_cur));
    if muw > 0
        fact   = (1 - sqrt(muw/Lipsw))/(1 + sqrt(muw/Lipsw));
    else
        fact   = (t_cur - 1) / t_next;
    end
    
    % Update z_next.
    z_next = w_next + fact*dw;
    
    % Reassign for the next iteration.
    t_cur = t_next;
    w_cur = w_next;
    z_cur = z_next;

    % Perform restart technique.
    if strcmpi(param.isSubRestart, 'fixed')
        if mod(iter, param.nSubRsIters) == 0
            t_cur = 1;
            z_cur = w_cur;
        end
    elseif strcmpi(param.isSubRestart, 'fx')
        if iter > 1 && fx_cur > fx_old
            t_cur = 1;
            z_cur = w_cur;
        end
    end
end 
% End of the main loop.

% Exceed the maximum number of iterations.
if iter >= param.SubMaxIters
    output.msg = 'Exceed the maximum number of iterations';
end

% Form the primal solution.
Wsol = w_cur;

% Form the primal proximal-Newton direction.
DX   = Wsol - X_cur;

% Compute lambda = sqrt(trace((wzw - W_cur)*DX)).
lambda = nan;
if ~strcmpi(param.lsMode, 'BTK') || ~param.isBtkLS
    if ~param.TerminateType == 1 || old_lbd > QuadRgRadius
        WDX    = W_cur*DX;
        lambda = sqrt( sum(sum(WDX.*WDX)) );
    end
end

% Assign the output.
output.time  = toc(output.time);
output.error = err;
output.iter  = iter;
output.Lips  = Lipsw;
output.L_cur = L_cur;

% If required, compute the dual norm of the gradient of f.
if param.isGradNorm
    SX                 = Shat*X_cur;
    output.dualGradNrm = sqrt( sum(sum(SX.*SX)) - 2*sum(diag(SX)) + p );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%