%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%        -------------------------------------------------------- 
%%%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%%%                          Version 1.0.0 (2013)  
%%%        ---------------------------0*0*0------------------------
%%%
%%%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%%%       algorithms for solving Sparse Inverse Covariance Estimation 
%%%                 Problems in Graphical Model Selection.
%%%
%%% INFORMATION:
%%%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%%%       EPFL, Lausanne, Switzerland.
%%%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%%%    Date: 26.11.2012.
%%%    Last modified: 27.05.2013.
%%%    Contact: quoc.trandinh@epfl.ch, volkan.cevher@epfl.ch
%%% 
%%% REFERENCES:
%%%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%%%    framework for composite minimization: Graph learning without 
%%%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%%%    USA.
%%%
%%% FUNCTION: [DX, lambda, output, Xsol] = scopt_glearn_pdAS(X_cur, ...
%%%                                        Shat, rho, X0, oiter, ...
%%%                                        old_lbd, param varargin) 
%%%
%%% PURPOSE: 
%%%     Solve the primal subproblem to compute the proximal-Newton
%%%     direction by using an active set method.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX, lambda, output, Xsol] = scopt_glearn_pdAS(X_cur, Shat, ...
                                      rho, X0, oiter, old_lbd, ...
                                      param, varargin) 

output.msg      = 'start';
output.iter     = 0;
output.prox_cnt = 0;
output.time     = tic;
p               = size(Shat, 1);
QuadRgRadius    = 0.75*0.219224;

% Set the maximum number of iterations.
MaxIters = param.SubMaxIters;
if param.SubMaxIters == 0 || param.SubMaxIters == inf
    MaxIters = min(5*oiter + 1, param.ubSubIters);
end

% Compute the Cholesky decomposition of matrix X_cur and W_cur =
% inv(X_cur).
[L_cur, ret] = chol(X_cur);
if ret ~= 0
    %error( 'Cholesky decomposition of Xk is failed!' );
    [V_cur, D_cur] = eig(X_cur);
    Dp_cur         = max( real(diag(D_cur)), 1 - 0.99);
    X_cur          = V_cur*diag(Dp_cur)*V_cur';
    L_cur          = V_cur*diag(sqrt(Dp_cur));
end
Ip                 = eye(size(Shat));
invL_cur           = L_cur\Ip;
W_cur              = invL_cur*invL_cur';
output.chol_cnt    = 1;

% Compute the vector gradient h_cur.
h_cur = Shat - 2*W_cur;
h_cur = h_cur(:);

% Using BLPC algorithm, we want to solve: 
%               min 0.5||Rx - b||_2^2 + lambda||x||_1.
%
% Step 1 - Initialize the inactive and active sets.
F_plus  = [];
F_minus = [];
H       = 1:p^2;

% The primal variable is initialized at zero.
beta_cur = zeros(p^2, 1);

% Initialize the dual variable (with warm-start).
if( oiter == 1 )
    d_cur = Shat - W_cur;
    d_cur = d_cur(:);
else
    d_cur = X0(:);
end

% Preparation for the main loop.
k    = MaxIters;
t    = p^2 + 1;
init = 1;
iter = 0;

% Step 3: The main loop.
while (init || ...
    scopt_glearn_blcp_feas_check(d_cur, beta_cur, rho, H, F_plus, F_minus))
    
    % Number of iterations, prox operations.
    iter            = iter + 1;
    init            = 0;
    output.prox_cnt = output.prox_cnt + 1; 
    
    % Terminate if the number of iterations exceeds the maximum.
    if iter > MaxIters, break; end
        
    % Step 4: Compute the infeasible set.
    J1 = H(d_cur(H) > rho);
    J2 = H(d_cur(H) < -rho);
    J3 = F_plus(beta_cur(F_plus) < 0);
    J4 = F_minus(beta_cur(F_minus) > 0);
    G  = union(union(J1, J2), union(J3, J4));
    
    % Steps 5: Take the subset of G to update the active set.
    if (oiter == 1)
        if (length(G) < t)
            t     = length(G);
            k     = MaxIters;
            G_hat = G;   
        elseif (length(G) >= t && k >= 1)
            k     = k - 1;
            G_hat = G;
        elseif (length(G) >= t && k == 0)
            G_hat = max(G);
        end;
    else
        G_hat = G;
    end
    
    % Step 6: Compute the infeasible subsets for updating the active set.
    J1_hat = intersect(G_hat, J1);
    J2_hat = intersect(G_hat, J2);
    J3_hat = intersect(G_hat, J3);
    J4_hat = intersect(G_hat, J4);
    
    % Update the active set (and inactive set).
    H       = union( setdiff(H, union(J1_hat, J2_hat)), union(J3_hat, J4_hat) );
    F_plus  = union( setdiff(F_plus, J3_hat), J1_hat );
    F_minus = union( setdiff(F_minus, J4_hat), J2_hat );
    
    % Step 7a: Set the known variables to their values.
    beta_cur(H)     = 0;
    d_cur(F_plus)   = rho;
    d_cur(F_minus)  = -rho;    
    
    % Step 7b: Recompute the rest of d and beta.
    F        = union(F_plus, F_minus);
    [IF, JF] = ind2sub(size(Shat), F);
    W_II     = W_cur(IF, IF);
    W_JJ     = W_cur(JF, JF);
    H_curFF  = W_II.*W_JJ;
    
    % Compute d_F.
    rhsF        = h_cur + d_cur;
    rhsF        = rhsF(F);
    beta_cur(F) = -H_curFF\rhsF;
    d_cur(H)    = -compute_iaDX(beta_cur, W_cur, H) - h_cur(H);
end
% End of the main loop.

% Compute the proximal-Newton search direction.
Xsol = reshape(beta_cur, p, p);
DX   = Xsol - X_cur;

% Compute lambda = sqrt(trace((wzw - W_cur)*DX)).
lambda = nan;
if ~strcmpi(param.lsMode, 'BTK') || ~param.isBtkLS
    if ~param.TerminateType == 1 || old_lbd > QuadRgRadius
        WDX    = W_cur*DX;
        lambda = sqrt( sum(sum(WDX'.*WDX)) );
    end
end

output.error = 0;
output.time  = toc(output.time);
output.iter  = iter;
output.Lips  = NaN;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: iaDX = compute_iaDX(beta, W_cur, H)
%%% PURPOSE: Compute the inactive part of DX.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function iaDX = compute_iaDX(beta, W_cur, H)

p    = size(W_cur, 1);
BETA = reshape(beta, p, []);
Hd   = W_cur*BETA*W_cur;
Hd   = Hd(:);
iaDX = Hd(H);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
