%
%        -------------------------------------------------------- 
%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%                          Version 1.0.0 (2013)  
%        ---------------------------0*0*0------------------------
%
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%       algorithms for solving Sparse Inverse Covariance Estimation 
%                 Problems in Graphical Model Selection.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 26.11.2012.
%    Last modified: 27.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%    framework for composite minimization: Graph learning without 
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%
% FUNCTION: [DX, X0, lambda, output] = scopt_glearn_subsolver(X_cur, ...
%                                      Shat, rho, X0, oiter, old_lbd, ...
%                                      param, varargin) 
%
% PURPOSE: 
%     Solving the primal subproblem to compute the proximal-Newton
%     direction by using user-defined solver.
%     The user defined solver is used to solve the following unconstrained
%     LASSO problem:
%                       min_x { 0.5*||A(x) - b||^2 + rho*||x||_1 }.
%     
%     The inputs of this procedure consist of:
%      - Aopr   : A linear operator to evaluate A(x); 
%      - ATopr  : Adjoint operator of A;
%      - b      : Vector b;
%      - rho    : Regularization parameters;
%      - x0     : The initial point if required;
%      - oiter  : The outer iteration counter.
%      - old_lbd: The previous proximal-Newton decrement.
%      - param  : optional parameters.
%     This solver works on the vector form instead of matrix form.
%     The outputs of this solver consist of:     
%      - x      : Final solution vector;
%      - fval   : Final objective value;
%      - output : Other information.
%

function [DX, Xsol, lambda, output] = scopt_glearn_subsolver(X_cur, ...
                                      Shat, rho, X0, oiter, old_lbd, ...
                                      param, varargin) 

% Output parameters.
output.msg      = 'start';
output.iter     = 0;
output.prox_cnt = 0;
output.time     = tic;
p               = size(Shat, 1);
QuadRgRadius    = 0.75*0.219224;

% Compute the Cholesky decomposition of matrix X_cur and W_cur =
% inv(X_cur).
[L_cur, ret] = chol(X_cur);
if ret ~= 0
    warning( 'Cholesky decomposition of Xk is failed!' );
    [V_cur, D_cur] = eig(X_cur);
    Dp_cur         = max( real(diag(D_cur)), 1 - 0.99);
    X_cur          = V_cur*diag(Dp_cur)*V_cur';
    L_cur          = V_cur*diag(sqrt(Dp_cur));
end
Ip                 = eye(size(Shat));
invL_cur           = L_cur\Ip;
invL_cur           = invL_cur';
W_cur              = invL_cur'*invL_cur;
output.chol_cnt    = 1;

% Compute vector b.
h_cur = Shat - W_cur; 
b     = invL_cur*X_cur*invL_cur - L_cur'*h_cur*L_cur';
b     = b(:);

% Generate a linear operator Aopr = A(x).
Aopr  = @(x)(Ax_opr(x, invL_cur));

% Define the adjoint operator ATopr = A'(x).
ATopr = @(x)(ATx_opr(x, invL_cur));

% Initial point (vector).
x0    = X0(:);

% Check the user-defined solver.
if isa(varargin{3}, 'function_handle')
    user_defined_solver = varargin{3};
else
    error('The user-defined solver is not provided properly!');
end

% Call the user-defined solver.
[xsol, fval, outputf] = user_defined_solver(Aopr, b, rho, ATopr, x0, ...
                                            param);

% Get the final solution.
Xsol = reshape(xsol, p, p);

% The objective value.
output.fval = fval;

% Other information.
output.L_cur = L_cur;
if isfield(outputf, 'msg'),
    output.msg = outputf.msg;
else
    output.msg = '';
end
if isfield(outputf, 'iter'),
    output.iter = outputf.iter;
else
    output.iter = nan;
end
if isfield(outputf, 'error'),
    output.error = outputf.error;
else
    output.error = nan;
end
if isfield(outputf, 'Lips'),
    output.Lips = outputf.Lips;
else
    output.Lips = nan;
end
if isfield(outputf, 'prox_cnt'),
    output.prox_cnt = outputf.prox_cnt;
else
    output.prox_cnt = nan;
end

% Form the primal proximal-Newton direction.
DX     = Xsol - X_cur;

% Compute lambda = sqrt(trace((wzw - W_cur)*DX)).
lambda = nan;
if ~strcmpi(param.lsMode, 'BTK') || ~param.isBtkLS
    if ~param.TerminateType == 1 || old_lbd > QuadRgRadius
        WDX    = W_cur*DX;
        lambda = sqrt( sum(sum(WDX'.*WDX)) );
    end
end

% If require to compute the dual norm of the gradient of f.
if param.isGradNorm
    SX                 = Shat*X_cur;
    output.dualGradNrm = sqrt( sum(sum(SX'.*SX)) - 2*sum(diag(SX)) + p );
end

% Computational time.
output.time  = toc(output.time);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: Ax = Ax_opr(x, iL )
%%% PURPOSE:  Performing the linear operator A(x).
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Ax = Ax_opr(x, iL)

p  = size(iL, 1);
X  = reshape(x, p, p);
Ax = iL*X*iL;
Ax = Ax(:);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: ATx = ATx_opr(x, iL )
%%% PURPOSE:  Performing the adjoint operator A'(x).
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function ATx = ATx_opr(x, iL)

p   = size(iL, 1);
X   = reshape(x, p, p);
ATx = iL'*X*iL';
ATx = ATx(:);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%