%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DX, lambda, H, Dfx, lbfgs, output] = scopt_mnlogistic_ntdir(
%%%                    solve_subprob(x, px, qx, Ahat, NAThat, N, rho, ...
%%%                    Dfx_old, x_old, H_old, lbfgs, iter, param)
%%%
%%% PURPOSE:  Compute the proximal-Newton direction.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [dx_cur, lambda, H_cur, dfx_cur, fx_cur, lbfgs, output] = ...
          scopt_logistic_ntdir(x_cur, Ahat, bhat, NAThat, N, rho, ...
                               dfx_old, x_old, H_old, lbfgs, param)
                             
% Evaluate the objective value and its gradient.
[fx_cur, dfx_cur, qx_cur] = scopt_logistic_fx_eval(x_cur, Ahat, ...
                            bhat, N, NAThat, param.isFxEval);

% Add the nonsmooth term.
fx_cur  = fx_cur + rho*sum(sum(abs(x_cur)));
H_cur   = [];

% Using the BFGS method.
if strcmpi(param.UpdateH, 'bfgs')
        
    % Compute vector h and define the H-operator.
    if ~isempty(dfx_old)
        y_cur = dfx_cur - dfx_old;
        s_cur = x_cur   - x_old;
        
        % Update the Hessian by the BFGS formula.
        H_cur = scopt_bfgs_update(H_old, s_cur, y_cur);
        
        % Calculate the gradient vector of Q(x).
        Dqx   = @(z, mode)( smooth_fx2(z, H_cur, dfx_cur, x_cur, mode) );
    else
        H_cur = eye(length(x_cur));
        Dqx   = @(z, mode)( smooth_fx2(z, H_cur, dfx_cur, x_cur, mode) );
    end
    
% Using the limited memory BFGS-type methods.
elseif strcmpi(param.UpdateH, 'lbfgs')
    
    % Compute vector h and define the H-operator.
    if ~isempty(dfx_old)
        y_cur = dfx_cur - dfx_old;
        s_cur = x_cur   - x_old;
        
        % Using the Limited memory BFGS formula.
        if y_cur'*s_cur > 1e-9
            if size(lbfgs.s_prev, 2) > param.LBfgsLength
              lbfgs.s_prev = [lbfgs.s_prev(:, 2:param.LBfgsLength), s_cur];
              lbfgs.y_prev = [lbfgs.y_prev(:, 2:param.LBfgsLength), y_cur];
              lbfgs.H_diag = (y_cur'*y_cur)/( y_cur'*s_cur);
            else
              lbfgs.s_prev = [lbfgs.s_prev, s_cur];
              lbfgs.y_prev = [lbfgs.y_prev, y_cur];
              lbfgs.H_diag = (y_cur'*y_cur)/(y_cur'*s_cur);
            end
        end
        Dqx   = @(z, mode)(scopt_lbfgs_update(z, x_cur, lbfgs.H_diag, ...
                           lbfgs.s_prev, lbfgs.y_prev, dfx_cur, mode) );
    else
        % The first iteration. 
        lbfgs.s_prev = zeros( length(x_cur), 0 );
        lbfgs.y_prev = zeros( length(x_cur), 0 );
        lbfgs.H_diag = 1;
        H_cur        = eye(length(x_cur));
        Dqx          = @(z, mode)(smooth_fx2(z, H_cur, dfx_cur, x_cur, mode));
    end

% Using the pure Newton method.
else
    Dqx = @(z, m)( smooth_fx(z, x_cur, qx_cur, NAThat, Ahat, dfx_cur, N, m));
end

% Solve the subproblem by the SPARSA method.
[x_trial, output] = scopt_mnlogistic_sparsa(x_cur, Dqx, rho, param);

% Define the search direction.
dx_cur = x_trial - x_cur;

% Compute lambda.
Adx    = Ahat*dx_cur;
nrmHd2 = sum(qx_cur.*Adx.*Adx); 
lambda = sqrt( nrmHd2/N );

% Assign the output.
output.x_trial = x_trial;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [gradf, fx] = smooth_fx(z, x, qx, NAThat, Ahat, Dfx, N, mode)
%%%
%%% PURPOSE:  Compute the smooth term function.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gradf, fx] = smooth_fx(z, x, qx, NAThat, Ahat, Dfx, N, mode)

% Compute the gradient of fx: gradf = NAThat*(qx.*(Ahat*x)) + Dfx.
dz    = z - x;
Adz   = Ahat*dz;
qAdz  = qx.*Adz;
gradf = [];
if mode == 1
    gradf = NAThat*qAdz + Dfx;
end

% Compute the function value if required.
if nargout > 1
    fx = 0.5*sum(qAdz.*Adz)/N + sum(Dfx.*dz);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [gradf, fx] = smooth_fx2(z, H, h, x, mode)
%%% PURPOSE:  Compute the smooth term function.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gradf, fx] = smooth_fx2(z, H, h, x, mode)

% Compute the gradient of fx.
gradf = [];
dz    = z - x;
Hdz   = H*dz;
if mode == 1
    gradf = Hdz + h; 
end

% Compute the function value if required.
if nargout > 1
    fx = sum((0.5*Hdz + h).*dz);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%