%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%                   ------------------------------------ 
%%%                   MAX-NORM IN CLUSTERING v.1.0. (2013)
%%%                   ------------------------------------ 
%%%
%%%    An implementation of the Proximal Gradient/Newton algorithms for 
%%%        solving Max-Norm Clustering Problems.
%%%
%%% INFORMATION:
%%%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%%%       EPFL, Lausanne, Switzerland.
%%%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%%%    Date: 16.04.2013.
%%%    Last modified: 16.04.2013.
%%%    Contact: quoc.trandinh@epfl.ch
%%% 
%%% FUNCTION: l2norm = scopt_maxnorm_l2norm(iAx, maxiter)
%%% PURPOSE:  Evaluate the L2-norm of the Hessian matrix of f at x.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function l2norm = scopt_maxnorm_l2norm(iAx, maxiter)

% Check the inputs.
if nargin < 2, maxiter = 20; end

% Initialize the direction.
p  = 0.5*size(iAx, 1);
dK = eye(p);
dL = eye(p);
dR = eye(p);

% The main iteration loop.
for iter = 1:maxiter
    
    % First compute u = f''(x)*A(d).
    Ad    = [dL, dK; dK', dR];
    u     = iAx*Ad*iAx;
    
    % Then compute vec_u = A'*u.
    uK    = 2*u(1:p, p+1:2*p);
    uL    = u(1:p,1:p);
    uR    = u(p+1:2*p,p+1:2*p);
    
    % Compute the l2-norm of u.
    nrm_u = sqrt( sum(sum(uK.*uK)) + sum(sum(uL.*uL)) + sum(sum(uR.*uR)) );
    
    % Normalize u to get d.
    dK    = uK/nrm_u;
    dL    = uL/nrm_u;
    dR    = uR/nrm_u;
end

% Compute u = H*d.
Ad     = [dL, dK; dK', dR];
u      = iAx*Ad*iAx;
uK     = 2*u(1:p, p+1:2*p);
uL     = u(1:p,1:p);
uR     = u(p+1:2*p,p+1:2*p);

% Compute norm(d)^2 and d'*v.
nrm_d2 = sum(sum(dK.*dK)) + sum(sum(dL.*dL)) + sum(sum(dR.*dR));
prd_du = sum(sum(uK.*dK)) + sum(sum(uL.*dL)) + sum(sum(uR.*dR));

% Compute l2-norm.
l2norm  = prd_du/nrm_d2;
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
