%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%                   ------------------------------------ 
%%%                   MAX-NORM IN CLUSTERING v.1.0. (2013)
%%%                   ------------------------------------ 
%%%
%%%    An implementation of the Proximal Gradient/Newton algorithms for 
%%%        solving Max-Norm Clustering Problems.
%%%
%%% INFORMATION:
%%%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%%%       EPFL, Lausanne, Switzerland.
%%%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%%%    Date: 16.04.2013.
%%%    Last modified: 16.04.2013.
%%%    Contact: quoc.trandinh@epfl.ch, volkan.cevher@epfl.ch
%%% 
%%% FUNCTION: [Ksol, Lsol, Rsol, output] = scopt_maxnorm_pFISTA(iAx, ...
%%%                                        K0, L0, R0, tau, Lips, mu, param)
%%%
%%% PUPROSE: Solve the primal subproblem to compute the proximal-Newton
%%%          direction by using the fast proximal-gradient method 
%%%          (Nesterov2007, Beck2009) to solve this problem. 
%%%
%%% INPUTS: 
%%%   iAx        : the inverse of matrix A(x).
%%%   xK, xL, xR : the current iteration point.
%%%   K0, L0, R0 : the initial point.
%%%   tau        : The barrier parameter.
%%%   Lips, mu   : The Lipschitz constant and strong convexity params.
%%%   AK         : data input.
%%%   param      : optional parameter (structure)
%%%
%%% OUTPUTS: 
%%%   Ksol, Lsol, Rsol : The approximate solution.
%%%   output           : The output information (structure)
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DKsol, DLsol, DRsol, output] = scopt_maxnorm_pFISTA(Ax, iAx, ...
                                         xK, xL, xR, K0, L0, R0, tau, ...
                                         AK, uLi, uRi, param)
% Output parameter.
output.msg      = 'start';
output.iter     = 0;
output.time     = tic;
output.prox_cnt = 0;

% Compute an approximation for the Lipschitz constant of f''(x). 
Lips = scopt_maxnorm_l2norm(iAx, param.PwMaxIters);

% Evaluate the strong convexity parameter.
mu  = 0;
if param.isEvalStrCvx
    mu = 0;
end

% Initialization.
p     = size(xK, 1);
uK    = K0;
uL    = L0;
uR    = R0;
vK    = uK;
vL    = uL;
vR    = uR;
theta = 1.0;

% The main loop.
for iter = 1:param.SubMaxIters
    
    % Form matrix Au.
    Au      = [uL, uK; uK', uR];
    
    % Compute the step size.
    alpha   = 1.0/Lips;
    
    % Compute the term phi'(A(x))*A(u) of the gradient vector.
    Atgradf = iAx*Au*iAx - 2*iAx;
    
    % Gradient of the function by applying the adjoint operator A'.
    gradfL = Atgradf(1:p, 1:p);   
    gradfK = Atgradf(1:p, p+1:2*p) + Atgradf(p+1:2*p, 1:p)'; 
    gradfR = Atgradf(p+1:2*p, p+1:2*p);
    
    % Compute putative solutions for K, L and R.
    vKnext = scopt_soft_threshold_oper(uK - alpha*gradfK - AK, alpha/tau) + AK;
    vLnext = scopt_clipping_oper(uL - alpha*gradfL, uLi);
    vRnext = scopt_clipping_oper(uR - alpha*gradfR, uRi);
    output.prox_cnt = output.prox_cnt + 1;
    
    % Compute the relative error.
    dK     = vKnext - vK;
    dL     = vLnext - vL;
    dR     = vRnext - vR;
    nrm_d  = sqrt( sum(sum(dK.*dK)) + sum(sum(dL.*dL)) + sum(sum(dR.*dR)));
    nrm_v  = sqrt( sum(sum(vK.*vK)) + sum(sum(vL.*vL)) + sum(sum(vR.*vR)));
    errork = nrm_d / max( nrm_v, 1 );
    
    % Check the stopping criterion.
    if errork <= param.SubRelTol
        output.msg = 'Convergence achieved!';
        break;        
    end;
    
    % Compute new step size.
    theta_next = 0.5*(1 + sqrt(1 + 4*theta^2));
    if mu <= 0
        fact = (theta - 1) / theta_next;
    else
        fact = (sqrt(Lips) - sqrt(mu))/(sqrt(Lips) + sqrt(mu));
    end
    
    % Update the vector u.
    uK = vKnext + fact*dK;
    uL = vLnext + fact*dL;
    uR = vRnext + fact*dR;
    
    % Reassign for the next iteration.
    theta  = theta_next;
    vK     = vKnext;
    vL     = vLnext;
    vR     = vRnext;
    
    % Perform Fixed restart if necessary.
    if strcmpi(param.isSubRestart, 'fixed')
        if mod(iter, param.nSubRsIters) == 0
            theta = 1;
            uK    = vK;
            uL    = vL;
            uR    = vR;
        end
    end
end 
% End of the main loop.

% Exceed the maximum number of iterations.
if iter >= param.SubMaxIters
    output.msg = 'Exceed the maximum number of iterations';
end

% Form the primal solution.
DKsol  = vK - xK;
DLsol  = vL - xL;
DRsol  = vR - xR;

% Get the final output. 
output.iter  = iter;
output.time  = toc(output.time);
output.Lips  = Lips;
output.mu    = mu;
output.error = errork;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%