%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DX, lambda, H, Dfx, lbfgs, output] = scopt_mnlogistic_ntdir(
%%%                solve_subprob(x, px, qx, Ahat, NAThat, N, rho, ...
%%%                Dfx_old, x_old, H_old, lbfgs, iter, param)
%%%
%%% PURPOSE:  Compute the proximal-Newton direction.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX_cur, lambda, H_cur, DFX_cur, FX_cur, lbfgs, output] = ...
          scopt_mnlogistic_ntdir(X_cur, What, Yhat, mWhat, TN, rho, ...
                                 DFX_old, X_old, H_old, lbfgs, param)
                             
% Evaluate the objective value and its gradient.
[FX_cur, DFX_cur, Phat] = scopt_mnlogistic_fx_eval(X_cur, What, ...
                          Yhat, TN, mWhat);

% Add the nonsmooth term.
FX_cur  = FX_cur + rho*sum(sum(abs(X_cur)));

% Make sure that DFX_cur is a vector.
if ~isvector(DFX_cur), DFX_cur = DFX_cur(:); end
H_cur = [];

% Using the BFGS method.
if strcmpi(param.UpdateH, 'bfgs')
        
    % Compute vector h and define the H-operator.
    if ~isempty(DFX_old)
        y_cur = DFX_cur - DFX_old;
        s_cur = X_cur   - X_old;
        
        % Update the Hessian by the BFGS formula.
        H_cur = scopt_bfgs_update(H_old, s_cur, y_cur);
        
        % Calculate the gradient vector of Q(x).
        Dqx   = @(Z, mode)( smooth_fx2(Z, H_cur, DFX_cur, X_cur, mode) );
    else
        H_cur = eye(length(X_cur));
        Dqx   = @(Z, mode)( smooth_fx2(Z, H_cur, DFX_cur, X_cur, mode) );
        
    end
    
% Using the limited memory BFGS-type methods.
elseif strcmpi(param.UpdateH, 'lbfgs')
    
    % Compute vector h and define the H-operator.
    if ~isempty(DFX_old)
        y_cur = DFX_cur - DFX_old;
        s_cur = X_cur   - X_old;
        
        % Using the Limited memory BFGS formula.
        if y_cur'*s_cur > 1e-9
            if size(lbfgs.s_prev, 2) > param.LBfgsLength
              lbfgs.s_prev = [lbfgs.s_prev(:, 2:param.LBfgsLength), s_cur];
              lbfgs.y_prev = [lbfgs.y_prev(:, 2:param.LBfgsLength), y_cur];
              lbfgs.H_diag = (y_cur'*y_cur)/( y_cur'*s_cur);
            else
              lbfgs.s_prev = [lbfgs.s_prev, s_cur];
              lbfgs.y_prev = [lbfgs.y_prev, y_cur];
              lbfgs.H_diag = (y_cur'*y_cur)/(y_cur'*s_cur);
            end
        end
        Dqx   = @(Z, mode)(scopt_lbfgs_update(Z, X_cur, lbfgs.H_diag, ...
                           lbfgs.s_prev, lbfgs.y_prev, DFX_cur, mode) );
    else
        % The first iteration. 
        lbfgs.s_prev = zeros( length(X_cur), 0 );
        lbfgs.y_prev = zeros( length(X_cur), 0 );
        lbfgs.H_diag = 1;
        H_cur        = eye(length(X_cur));
        Dqx          = @(Z, mode)(smooth_fx2(Z, H_cur, DFX_cur, X_cur, mode));
    end

% Using the pure Newton method.
else
    Dqx   = @(Z, mode)(smooth_fx(Z, X_cur, What, Yhat, Phat, ...
                                 mWhat, TN, mode) );
end

% Solve the subproblem by the SPARSA method.
[X_trial, output] = scopt_mnlogistic_sparsa(X_cur, Dqx, rho, param);

% Define the search direction.
DX_cur = X_trial - X_cur;

% Compute lambda.
DX = DX_cur;
if isvector(DX_cur)
    p  = size(What, 1);
    mx = size(Yhat, 1);
    DX = reshape(DX_cur, p, mx); 
end

QDX    = DX'*What;
QPhat  = sum(QDX.*Phat, 1);
nrmHd2 = sum(sum(Phat.*QDX.^2)) - sum( QPhat.^2);
lambda = sqrt(nrmHd2/TN);

% Assign the output.
output.X_trial = X_trial;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [gradFX, FX] = smooth_fx(Z, X, What, Yhat, Phat, mWhat, ...
%%%                                    TN, mode)
%%% PURPOSE:  Compute the smooth term function.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gradFX, FX] = smooth_fx(Z, X, What, Yhat, Phat, ...
                                  mWhat, TN, mode)

% NOTE: This function has not been implemented yet!
error('This option has not implemented yet in this version!');

% Compute the gradient of fx.
DZ     = Z - X;
gradFX = [];
if mode == 1
    gradFX = DZ;
end

% Compute the function value if required.
if nargout > 1
    FX = 0;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DFX, FX] = smooth_fx2(Z, H_cur, h_cur, X_cur, mode)
%%% PURPOSE:  Compute the smooth term function.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DFX, FX] = smooth_fx2(Z, H_cur, h_cur, X_cur, mode)

% Compute the gradient of fx.
DFX = [];
DZ  = Z - X_cur;
HDZ = H_cur*DZ;
if mode == 1
    DFX = HDZ + h_cur; 
end

% Compute the function value if required.
if nargout > 1
    FX = sum((0.5*HDZ + h_cur).*DZ);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%