%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [xsol, output] = scopt_mnlogistic_sparsa(x, smooth_fx, ...
%%%                            rho, param)
%%% PURPOSE: Solve the basic LASSO subproblem by the SPARSA algorithm.
%%%
%%%               min_y 0.5*(y-x)'*H*(y-x) + h'*(y-x) + rho*|y|_1.
%%% INFORMATION:
%%%   Date: 24.06.2013.
%%%   By Quoc Tran Dinh.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [xsol, output] = scopt_mnlogistic_sparsa(x, smooth_fx, rho, param)

% Start the clock.
output.time = tic;

% Evaluate fx and its gradient at x.
[Dfx, fx] = smooth_fx(x, 1);
    
% Evaluate the proximal point.
x_prox = scopt_soft_threshold_oper(x - Dfx, rho);

% Count the number of prox-operators and back-tracking line-search.
output.prox_cnt    = 1;
output.btkit_count = 0;

% Check if x is the optimal point.
optim = norm(x_prox - x, 'inf');
if optim <= param.SubTolOpt
    output.msg   = 'The optimality residual is below TolFun';
    xsol         = x;
    output.iter  = 0;
    output.time  = toc(output.time);
    output.optim = optim;
    output.error = 0;
    output.Lips  = inf;
    return;
end

% Evaluate the objective value.
Fx = fx + rho*norm(x, 1);

% The main loop.
for iter = 1:param.SubMaxIters
    
    % Compute the Barzilai-Borwein step.
    if iter > 1
        s       = x   - x_old;
        y       = Dfx - Dfx_old;
        sumy2   = sum(y.*y);
        
        if sumy2 > 0
            BB_step = sum(y.*s) / sumy2;
        else
            BB_step = 1e-10;
        end
        
        % Prevent an infinity step.
        if BB_step <= 1e-9 || 1e9 <= BB_step
        
            norm1Dfx = norm(Dfx, 1);
            if isnan(norm1Dfx) || isinf(norm1Dfx) || norm1Dfx < 1e-9
                BB_step = 1;
            else
                BB_step = min(1, 1/norm1Dfx);
            end
        end
    else
        norm1Dfx = norm(Dfx, 1);
        if isnan(norm1Dfx) || isinf(norm1Dfx) || norm1Dfx < 1e-9
            BB_step = 1;
        else
            BB_step = min(1, 1/norm1Dfx);
        end
    end
    
    % Memorize the iterative data.
    x_old   = x;
    if iter + 1 > param.BtkMemIters
        Fx_old = [Fx_old(2:end), Fx];
    else
        Fx_old(iter) = Fx;
    end
    Dfx_old  = Dfx;
    
    % Perform the backtracking line-search procedure.
    [x, Fx, Dfx, alpha, outputk] = btklinesearch(x, -Dfx, ...
                                   BB_step, Fx_old, -norm(Dfx)^2, ...
                                   smooth_fx, rho, param);
	                           
    % Compute the optimality condition.
    x_prox = scopt_soft_threshold_oper(x - Dfx, rho);
    optim  = norm(x_prox - x, 'inf');
    errork = norm(x - x_old, 'inf')/max(1, norm(x_old, 'inf'));
    output.prox_cnt    = output.prox_cnt    + outputk.prox_cnt + 1;
    output.btkit_count = output.btkit_count + outputk.iter;
    
    % Check the stopping criterion.
    if optim <= param.SubTolOpt
        output.msg = 'The optimility residual is below TolFun';
        break;
    elseif errork <= param.SubRelTol
        output.msg = 'The relative error is below TolX';
        break;
    end
    
end
% End of the main loop.

% Exceed the maximum number of iterations.
if iter >= param.SubMaxIters
    output.msg = 'Exceed the maximum number of iterations';
end

% The final output.
output.fval  = Fx;
output.Dfx   = Dfx;
output.alpha = alpha;
output.optim = optim;
output.error = errork;
output.iter  = iter;
output.time  = toc(output.time);
output.Lips  = BB_step;
xsol         = x;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [xsol, Fsol, Dgsol, alpha, output] = btklinesearch( ...
%%%                                                x, dx, alpha, Fx_old, ...
%%%                                                Dfx, smoothFx, param )
%%% PURPOSE: Perform the backtracking line search procedure.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [xsol, fxval, dfx, alpha, output] = btklinesearch( ...
                                             x, dx, alpha, Fx_old, ...
                                             Dfx, smoothFx, rho, param)
% The output.
output.prox_cnt = 0;

% The tracking factor.
beta = 0.5;

% The main loop.
for btkiter = 1:param.BtkMaxIters
    
    % Generate the new point.
    x_plus = x + alpha*dx;
    
    % Compute the prox-point of x_plus.
    x_new  = scopt_soft_threshold_oper(x_plus, rho*alpha);
    
    % Count the number of prox-operator.
    output.prox_cnt = output.prox_cnt + 1;
    
    % Evaluate the function value f and its gradient at a new point.
    [Df_new, fx_new] = smoothFx(x_new, 0);
    
    % The objective value.
    Fx_new = fx_new + rho*norm(x_new, 1);
    
    % Check the stopping criterion.
    Dx_descent = 0.5*norm(x_new - x)^2;
    if Fx_new < max(Fx_old) + param.DescentSlop*alpha*Dx_descent
        output.msg = 'Line-search is satisfied!';
        break;
    elseif alpha <= param.TolStep
        output.msg = 'The step size is below TolStep';
        break;
    end
    
    % Backtrack if objective value not well-defined of function seems
    % linear.
    if isnan(Fx_new) || isinf(Fx_new) || ...
       abs(Fx_new - Fx_old(end) - alpha*Dfx) <= 1e-9
        
        alpha = beta*alpha;
        
    % Safeguard quadratic interpolation
    else
        deno         = (2*(Fx_new - Fx_old(end) - alpha*Dfx));
        if abs(deno) > 1e-9
            alpha_interp = -(Dfx*alpha^2)/deno;
            if 0.1 <= alpha_interp || alpha_interp <= 0.9*alpha 
                alpha = alpha_interp;
            else
                alpha = beta*alpha;
            end
        else
            alpha = beta*alpha;
        end
    end
end
% End of the backtracking loop.

% Compute the gradient at the final iteration.
Df_new = smoothFx(x_new, 1);

% Exceed the maximum number of iterations.
if btkiter >= param.BtkMaxIters
    output.msg = 'Exceed the maximum number of iterations!';
end

% The final output and solution.
output.iter = btkiter;
dfx         = Df_new;
fxval       = Fx_new;
xsol        = x_new;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
