% 
%                ----------------------------------------- 
%                POISON IMAGE RECONSTRUCTION v.1.0. (2013)
%                ----------------------------------------- 
% 
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%    algorithms for solving Poisson image reconstruction problems.
% 
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 22.05.2013.
%    Last modified: 22.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% 
% FUNCTION: [Xsol, Wsol, output] = scopt_poisson_dbFISTA(Xk, Axk, ...
%                                  gradfk, tau, lbx, Aopr, ATopr, yf, ...
%                                  beta, Mf, Lips, param)
% 
% PURPOSE: Solve the deblurring problem by FISTA:
%            min_{0.5*x'*H*x + h'*x + tau*|x|_TV : x >= lbx }
%          where tau > 0 is a given parameter.
%
% This implementation is based on the procedure deblur_tv_fista in the
% SPIRAL-TAP package from http://www.ee.duke.edu/~zth/spiral/.
%

function [Xsol, Wsol, output] = scopt_poisson_dbFISTA(Xk, Axk, gradfk, ...
                                W0, tau, lbx, Aopr, ATopr, yf, beta, ...
                                Mf, Lips, param)

% Initialization.
output.msg     = 'start';
output.time    = tic;
check_stop     = 0;
fx_val         = inf;
Xj             = Xk;
Yj             = Xj;
tj             = 1;

% Generate the initial point for poisson_fista_tv procedure (warm-start).
Wj{1}          = W0{1};
Wj{2}          = W0{2};

% The main iteration loop.
for iter1 = 1:param.SubMaxIters
    
   % Evaluate the objective value.
   if strcmpi(param.isSubRestart, 'fx')
       fx_old = fx_val;
       fx_val = eval_quad_tv_fx(Xj, Xk, Axk, Aopr, yf, beta, Mf);
   end
    
   % Evaluate the gradient of F.
   DFj    = gradf_eval(Yj, Xk, Axk, gradfk, Aopr, ATopr, yf, beta, Mf);
   
   % Compute the putative solution Yj_new.
   Yj     = Yj - (1.0/Lips)*DFj; 
   
   % Compute the threadholding level.
   gamma  = tau/Lips; 
   
   % Solve the denoising subproblem.
   [Xj_new, Wj, ~] = scopt_poisson_dnFISTA(Yj, gamma, lbx, ...
                                                 Xj, Wj, param);
                                                
   % Compute the search direction.
   DXj    = Xj_new - Xj;
   
   % Compute the parameter tj.
   tj_new = 0.5*(1.0 + sqrt(1.0 + 4*tj^2));
   factor = (tj - 1)/tj_new;
   
   % Compute Yj
   Yj     = Xj_new + factor*DXj;
   
   % Compute the error.
   nrmk   = norm(DXj(:), 2);
   errork = nrmk / max(1, norm(Xj(:)));
   
   % Check the stopping criterion.
   if errork <= param.SubRelTol
       check_stop = check_stop + 1;
   else
       check_stop = 0;
   end
   if check_stop >= param.nSubFxCheck
       output.msg  = 'Convergence acheived!';
       break;
   end
   
   % Assignment for the next iteration.
   tj = tj_new;
   Xj = Xj_new;
   
   % Restart if necessary.
    if strcmpi(param.isSubRestart, 'fixed')
        if mod(iter, param.nSubRsIters) == 0
            tj = 1;
            Yj = Xj;
        end
    elseif strcmpi(param.isSubRestart, 'fx')
        if iter > 1 && fx_val > fx_old
            tj = 1;
            Yj = Xj;
        end
    end
end
% End of the main loop.

% Exceed the maximum number of iterations.
if iter1 > param.SubMaxIters
    output.msg  = 'Exceed the maximum number of iterations';
end

% Get the final solution.
Xsol         = Xj;
Wsol{1}      = Wj{1};
Wsol{2}      = Wj{2};
output.iter  = iter1;
output.time  = toc(output.time);
output.error = errork;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: gradf = gradf_eval(X, Xk, Axk, gradfk, Aopr, ATopr, ...
%%%                              yf, beta, Mf)
%%% PURPOSE: Evaluate the gradient of f(x) = 0.5*x'*H*x + h'*x.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function gradf = gradf_eval(X, Xk, Axk, gradfk, Aopr, ATopr, yf, beta, Mf)

Adk   = Aopr(X - Xk);
Dfk   = yf.*Adk./(Axk + beta).^2;
gradf = Mf*ATopr(Dfk) + gradfk;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: fval = eval_quad_tv_fx(X, Xk, Axk, Aopr, yf, beta, Mf)
%%% PURPOSE: Evaluate the objective value.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function fval = eval_quad_tv_fx(X, Xk, Axk, Aopr, yf, beta, Mf)

Adk   = Aopr(X - Xk);
Ad2k  = Adk.^2;
dHd   = yf.*Ad2k./(Axk + beta).^2;
Dfk   = 1 - yf./(Axk + beta);
fval  = 0.5*Mf*dHd + Mf*sum(sum(Dfk.*Adk));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


                      