%
%                ----------------------------------------- 
%                POISON IMAGE RECONSTRUCTION v.1.0. (2013)
%                ----------------------------------------- 
% 
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%    algorithms for solving Poisson image reconstruction problems.
% 
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 22.05.2013.
%    Last modified: 22.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% FUNCTION: [Xout, Xsol, Wsol, output] = scopt_poisson_dnFISTA(...
%                                        Xcenter, tau, lbx, X0, W0, ...
%                                        param, oiter)
% PURPOSE: This implementation is based on the procedure denoise_bound 
%          in the SPIRAL-TAP package from
%                   http://www.ee.duke.edu/~zth/spiral/.
%
function [Xout, Xsol, Wsol, output] = scopt_poisson_dnFISTA(Xcenter, ...
                                      tau, lbx, X0, W0, param, oiter)

% The size of variables.
[m, n] = size(Xcenter);

% Define the maximum number of iterations.
if param.SubMaxIters == 0
    MaxIters = min(2*oiter + 1, param.ubSubIters);
else
    MaxIters = param.SubMaxIters;
end

% Initialization.
output.time     = tic;
output.iter     = inf;
output.msg      = 'start';
output.prox_cnt = 0;
X_cur           = X0;
W_cur{1}        = W0{1};
W_cur{2}        = W0{2};
Y_cur{1}        = W_cur{1};
Y_cur{2}        = W_cur{2};
t_cur           = 1;
fx_val          = inf;
check_stopping  = 0;

% The main loop.
for iter = 1:MaxIters
    
    % Compute the prox-operation step.
    X_next = max(Xcenter - tau*L_forward(Y_cur), lbx);
    
    % Apply the adjoint operator.
    Z_cur   = L_trans(X_next);
    
    % Gradient descent step.
    factor1   = 1/(8*tau);
    W_next{1} = Y_cur{1} + factor1*Z_cur{1};
    W_next{2} = Y_cur{2} + factor1*Z_cur{2};
    
    % Peforming the projection step.
    if strcmpi(param.proxOperType, 'iso')
        A_cur     = [W_next{1}; zeros(1, n)].^2 + [W_next{2}, zeros(m, 1)].^2;
        A_cur     = sqrt(max(A_cur, 1));
        W_next{1} = W_next{1}./A_cur(1:m-1, :);
        W_next{2} = W_next{2}./A_cur(:, 1:n-1);
    else
        W_next{1} = W_next{1}./(max(abs(W_next{1}), 1));
        W_next{2} = W_next{2}./(max(abs(W_next{2}), 1));
    end
    
    % Count the number of prox-operations.
    output.prox_cnt = output.prox_cnt + 1;
    
    % Updating the parameter tk.
    t_next  = 0.5*(1 + sqrt(1 + 4*t_cur^2));
    factor2 = (t_cur - 1)/t_next;
    
    % Update the iterative point Yk.
    Y_next{1} = W_next{1} + factor2*(W_next{1} - W_cur{1});
    Y_next{2} = W_next{2} + factor2*(W_next{2} - W_cur{2});
    
    % Compute the error.
    DX_cur  = X_next - X_cur;
    nrm_dx  = sqrt(sum(sum(DX_cur.*DX_cur)));
    err_cur = nrm_dx / max( sqrt(sum(sum(X_cur.*X_cur))), 1);
    if err_cur <= param.SubRelTol
        check_stopping = check_stopping + 1;
    else
        check_stopping = 0;
    end
    
    % Check stopping criterion.
    if check_stopping >= param.nSubFxCheck
        output.msg  = 'Convergence achieved!';
        break;
    end
    
    if strcmpi(param.isSubRestart, 'fx')
        U_cur  = Xcenter - tau*L_forward(W_next);
        V_cur  = max(U_cur, lbx);
        UV_cur = U_cur - V_cur;
        fx_old = fx_val;
        fx_val = -sum(sum(UV_cur.*UV_cur)) + sum(sum(UV_cur.*UV_cur));
    end
    
    % Assignment for the next iteration.
    Y_cur{1} = Y_next{1};
    Y_cur{2} = Y_next{2};
    W_cur{1} = W_next{1};
    W_cur{2} = W_next{2};
    X_cur    = X_next;
    t_cur    = t_next;
    
    % Restart if necessary.
    if strcmpi(param.isSubRestart, 'fixed')
        if mod(iter, param.nSubRsIters) == 0
            t_cur    = 1;
            Y_cur{1} = W_cur{1};
            Y_cur{2} = W_cur{2};
        end
    elseif strcmpi(param.isSubRestart, 'fx')
        if iter > 1 && fx_val > fx_old
            t_cur    = 1;
            Y_cur{1} = W_cur{1};
            Y_cur{2} = W_cur{2};
        end
    end
end
% End of the main loop.

% Exceed the maximum number of iterations.
if iter >= param.SubMaxIters
    output.msg  = 'Exceed the maximum number of iterations';
end

% Get the final solution.
Xout        = X_cur;
Xsol        = X_cur;
Wsol        = W_cur;
output.time = toc(output.time);
output.iter = iter;
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: X = L_forward(P)
%%% PURPOSE:  Compute the forward linear operator.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function X = L_forward(P)

[m1, n1] = size(P{1});
[m2, n2] = size(P{2});

if(n1 ~= n2 + 1), error('Dimensions are not consistent'); end
if(m2 ~= m1 + 1), error('Dimensions are not consistent'); end

m = m1 + 1;
n = n1;

X           = zeros(m, n);
X(1:m-1, :) = P{1};
X(:, 1:n-1) = X(:, 1:n-1) + P{2};
X(2:m,:)    = X(2:m,:)    - P{1};
X(:,2:n)    = X(:,2:n)    - P{2};

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: P = L_trans(X)
%%% PURPOSE:  Compute the linear transform operator.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function P = L_trans(X)

[m, n] = size(X);
P{1}   = X(1:m-1, :) - X(2:m, :);
P{2}   = X(:, 1:n-1) - X(:, 2:n);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
