
% Load the data.
pname   = 'Lymph.mat';
dataDir = '';
data    = load([dataDir, pname]);
Shat    = data.S;
p       = size(Shat, 1);
rho     = 0.5;

% Optional parameters.
param = struct( ...
        'MaxIters', 100, ...
        'RelTol',   1e-8, ...
        'SubRelTol', 1e-9, ...
        'PrintLevel', 1 ...
        );

% Transform to gpuArrays.
%Shat    = gpuArray(Shat);
%rho     = gpuArray(rho);

% Initialize the parameters.
output.time  = tic;
QuadRgRadius = 0.75*0.219224;

% Generate the initial point.
X0 = diag(1./diag(Shat));
Y0 = diag(diag(Shat));

% Initialization.
X_cur = X0;

% The main iteration loop.
for iter = 1:param.MaxIters
    
    % Evaluate the Lipschit constant.
    DZ = ones(p, 1);
    for pwiter = 1:20
        DZ = (X_cur*DZ)/norm(DZ, 2);
    end
    l2nrmx = DZ'*(X_cur*DZ)/(DZ'*DZ);
    Lipsx  = l2nrmx*l2nrmx;

    % Compute the Newton-direction by solving the dual problem.
    y_cur = Y0;   
    u_cur = y_cur;
    t_cur = 1.0;
    
    % Compute the fixed term of the gradient.
    SX    = Shat*X_cur;
    Qt    = X_cur*SX - 2.0*X_cur;
    rhoQt = Qt/rho;

    % The iteration loop for solving the subproblem.
    for subiter = 1:min(5*iter + 1, 500)

        % Compute the gradient vector.
        XY      = X_cur*y_cur;
        XYX     = XY*X_cur;
        grad_QY = XYX + rhoQt;
   
        % Perform the projected gradient step.
        u_trial = y_cur - (1/Lipsx)*grad_QY;
        u_next  = sign(u_trial).*min(abs(u_trial), 1.0);

        % Compute the relative error.
        du     = u_next - u_cur;
        errorj = sqrt(sum(sum(du.*du)))/max(1.0, sqrt(sum(sum(u_cur.*u_cur))));

        % Check the stopping criterion.
        if errorj <= param.SubRelTol
            % Keep the last iteration.
            u_cur = u_next;
            break;        
        end;

        % Compute new step size.
        t_next = 0.5*(1 + sqrt(1 + 4*t_cur^2));
        fact   = (t_cur - 1) / t_next;
        
        % Update y_next
        y_next = u_next + fact*du;

        % Reassign for the next iteration.
        t_cur = t_next;
        u_cur = u_next;
        y_cur = y_next;
    end 
    % End of the subiteration loop.

    % Form the dual solution.
    Y0      = u_cur;

    % Form the primal proximal-Newton direction.
    YX      = Y0*X_cur;
    XYX     = X_cur*YX;
    DX      = -rho*XYX - Qt - X_cur;

    % Compute lambda - the norm of the proximal-Newton decrement.
    Stmp    = SX +  rho*YX;
    lambdak = sqrt(abs(p - 2*sum(diag(Stmp)) + sum(sum(Stmp'.*Stmp))));

    % Compute the step size.
    alphak = 1.0/(1.0 + lambdak);
    
    % Update the next iteration.
    if lambdak <= QuadRgRadius
        X_new = X_cur + DX;
    else
        X_new = X_cur + alphak*DX;
    end
            
    % Symmetrizing to avoid the nonsymetric due to numerical issues.
    X_new = (X_new + X_new')/2;
    
    % Print the iteration.
    if param.PrintLevel > 0
        fprintf(['Iter = %4d, lambda = %6.3e, alpha = %6.3e,', ...
                 ' serror = %6.3e, siter = %4d\n'], iter, lambdak, ...
                 alphak, errorj, subiter);
    end
    
    % Check the stoping criterion.
    if lambdak <= param.RelTol
        output.msg = 'Convergence acheived!';
        % Update the final iteration before breaking.
        X_cur      = X_new;
        break;
    end
    
    % Update the next iteration.
    X_cur = X_new;
end
% End of the main loop

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = 'Exceed the maxinum number of iterations';
end

% Get the final information.
output.time = toc(output.time);
output.iter = iter;

% Get the final solution.
Xsol        = X_cur;

% Print the output.
if param.PrintLevel > 0
    fprintf('+ CPU time (s): %6.5f\n', output.time);
    fprintf('+ Number of iterations: %4d\n', output.iter);
end

%%% END OF THE IMPLEMENTATION.
