%
%        -------------------------------------------------------- 
%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%                          Version 1.0.0 (2013)  
%        ---------------------------0*0*0------------------------
%
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%       algorithms for solving Sparse Inverse Covariance Estimation 
%                 Problems in Graphical Model Selection.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 26.11.2012.
%    Last modified: 27.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%    framework for composite minimization: Graph learning without 
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%
% FUNCTION: [final_sol, output] = scopt_glearn_fb_pga(Shat, rho, param, ... 
%                                 varargin)
%
% PURPOSE: An implementation of the Function based Proximal-Gradient 
%          algorithm for solving Sparse Inverse Covariance Selection 
%          Problem of the form:
%
%               min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%               s.t. X > 0.   
%
% INPUTS:
%    Shat     : a pxp symmetric positive definite matrix, inverse of the
%               covariance matrix.
%    rho      : the penalty parameter (or regularization parameter)
%               rho should be a positive number.
%    param    : optional structure to control the algorithm,
%               (see scopt_optimset() for more details).
%    varargin : user-defined inputs,
%               (vararagin{1} is an initial guess).
%
% OUTPUTS:
%    final_sol: A structure containing the final solution.
%    output   : A structure containing the output information.
%
% NOTE:
%    The output information at each iteration will be printed by setting
%    param.PrintLevel = 1 or 2. If param.PrintLevel = 0 then no output
%    information is printed.
%    To evaluate the objective value, set param.isFxEval = true.
%    To turn of back-tracking line-search, set param.isBtkLS = false.
%    Type "help scopt_prints" for further output information.    
%

function [final_sol, output] = scopt_glearn_fb_pga(Shat, rho, param, varargin)

% Check the input parameters.
if nargin < 4
    X0 = diag(1./diag(Shat));
else
    X0 = varargin{1}; 
end
if isempty(X0),    X0    = diag(1./diag(Shat)); end
if nargin < 3,     param = scopt_optimset([]);  end
if isempty(param), param = scopt_optimset([]);  end

% Check the regularization parameters.
if rho < 0 || isnan(rho) || isinf(rho) || ~isnumeric(rho)
    error('The penalty parameters should be positive'); 
end

% Check consistency issues.
if any(size(X0) ~= size(Shat))
    error('The initial point is inconsistent.'); 
end 
if size(Shat, 1) ~= size(Shat, 2), 
    error('The input matrix is not square.'); 
end
if norm(Shat - Shat', 'fro') > 1e-5 
    error('The input matrix is nonsymmetric'); 
end

% Check the algorithmic parameters.
param = scopt_optimset(param);
param.Algorithm = 'PG';
max_svd_value   = 4;

if param.PrintLevel == 1, nsp = 42; else nsp = 97; end
scopt_copyright(param, nsp, 'glearn');

p               = size(Shat, 1);
output.nx       = p;
output.msg      = 'start';
output.prox_cnt = 0;
output.info     = scopt_trace(param, []);
output.time     = tic;

% Initalization phase.
X_old     = [];
gradF_cur = [];
X_cur     = X0;
fx_list   = [];

% Compute the inverse of X_cur by Cholesky decomposition method.
[L_cur, X_cur, W_cur] = compute_Cholesky_decomp(X_cur, [], param);

% Evaluate the objective value if required.
fx_cur = scopt_glearn_fx_eval(X_cur, Shat, rho, L_cur);

% The main loop.
for iter = 1:param.MaxIters
    
    % Compute the gradient of Fx at X_cur.
    gradF_old = gradF_cur;
    gradF_cur = Shat - W_cur;
    
    % Evaluate the Lipschitz constant by the Barzilai-Borwein rule.
    if param.BBStepType ~= 0
        Lips_cur = scopt_compute_BBstep(X_old, X_cur, gradF_old, gradF_cur, ...
                  param, max_svd_value);
        % Damp the BB step if necessary.
	else
        Lips_cur = scopt_l2normx(X_cur, param.PwMaxIters);
        Lips_cur = Lips_cur*Lips_cur;
    end
	
    % Compute the trial step.
    [DX_cur, S_cur, lambda, nrm_Dx, beta] = compute_trial_step(X_cur, ...
                                            gradF_cur, rho, W_cur, Lips_cur);
                                 
    % Count the number of prox-operations.
    output.prox_cnt = output.prox_cnt + 1;

    % Back-tracking if required.
    if param.isBtkLS
        for btkit = 1:param.BtkMaxIters
            % If back-tracking is no longer required, then break.
            if Lips_cur <= (lambda*lambda)/(nrm_Dx*nrm_Dx)
                break; 
            end
            
            % Otherwise, half the Lipschitz constant.
            Lips_cur = Lips_cur/param.btkSlop;
            
            % And recompute the trial step.
            [DX_cur, S_cur, lambda, nrm_Dx, beta] = compute_trial_step(...
                                                    X_cur, gradF_cur, ...
                                                    rho, W_cur, Lips_cur);
                                                
            % Count the number of prox-operations.
            output.prox_cnt = output.prox_cnt + 1;
        end
    end
    
    % Compute the step size and make sure that alpha <= 1.
    alpha   = min( beta/(lambda*(lambda + beta)), 1.0);
    
    % Update the next step.
    X_next  = X_cur + alpha*DX_cur;
    
    % Evaluate the objective value at X_next
    [fx_next, Lx_next] = scopt_glearn_fx_eval(X_next, Shat, rho);
    
    % Evaluate the objective value at S_cur.
    [fs_cur, Ls_cur]   = scopt_glearn_fx_eval(S_cur, Shat, rho);
    
    % Check the objective values fx_next and fs_cur.
    if fs_cur < fx_next
        fx_next = fs_cur;
        X_next  = S_cur;
        alpha   = 1;
        L_cur   = Ls_cur;
    else
        L_cur   = Lx_next;
    end
    
    % Symmetrization step.
    X_next  = 0.5*(X_next + X_next');
    
    % Compute the theoretical decrease of the objective function.
    ratio   = beta/lambda;
    dec_fx  = abs(ratio - log(1 + ratio));
    
    % Compute the relative error.
    err_cur = nrm_Dx / max(1.0, sqrt(sum(sum(X_cur'.*X_cur))));
    
    % Check stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, fx_list] = scopt_fx_terminate(fx_cur, fx_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop && err_cur <= param.acptRelTolX
            output.msg = 'The objective values do not improve significantly';
            % Keep the last iteration.
            X_cur  = X_next;
            fx_cur = fx_next;
            break;
        end
    end
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX
        output.msg = 'Convergence achieved! Error is below the tolerance!';
        % Keep the last iteration.
        X_cur  = X_next;
        fx_cur = fx_next;
        break;
    end
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, fx_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, beta, 0, ...
                              dec_fx, Lips_cur);

    % Print the iteration.
    scopt_prints(param, iter, fx_cur, err_cur, lambda, beta, alpha, ...
                 dec_fx, output.prox_cnt, Lips_cur);
    
    % Jump to the next iteration.
    X_old  = X_cur;
    X_cur  = X_next;
    fx_cur = fx_next;
    
    % Compute the inverse of X_cur by Cholesky decomposition method.
    [L_cur, X_cur, W_cur] = compute_Cholesky_decomp(X_cur, L_cur, param);
end

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', ...
                  char(13), 'Increase the number of iterations to ', ...
                  'improve or change the tolerance param.abs_tol.'];
end

% Get the final information.
output.time  = toc(output.time);
output.iter  = iter;
output.fval  = fx_cur;
output.error = err_cur;
output.X0    = X0;

% Get the final solution.
final_sol.X     = X_cur;
final_sol.f     = fx_cur;
final_sol.error = err_cur;

% Print the final outputs.
scopt_finaloutput(param, output, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [L, X_out, W_out, iL_out] = compute_Cholesky_decomp(X, param)
%%% PURPOSE:  Compute the Cholesky decomposition.
%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [L, X_out, W_out, iL_out] = compute_Cholesky_decomp(X, L, param)

ret = 0;
if isempty(L)
    [L, ret] = chol(X);
end
if ret ~= 0
    if strcmpi(param.isWarning, 'on')
        fprintf('Warning: Cholesky factorization failed!\n');
    end
    [V, D] = eig(X);
    Dp     = max( real(diag(D)), 1 - 0.99);
    X      = V*diag(Dp)*V';
    L      = V*diag(sqrt(Dp));
end
if nargout > 1, X_out = X; end
if nargout > 2
    Ip     = eye(size(X, 1));
    iL     = L\Ip;
    W_out  = iL*iL';
    if nargout > 3, iL_out = iL; end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DX, X_trial, lambda, nrm_Dx, beta] = compute_trial_step(...
%%%                                                 X, gradFx, rho, W, L)
%%% PURPOSE:  Compute required quantities for step size generation.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX, X_trial, lambda, nrm_Dx, beta] = compute_trial_step(X, ...
                                               gradFx, rho, W, L)

% The trial step size.
gamma   = 1/L;
    
% Update the next iteration.
X_trial = scopt_soft_threshold_oper(X - gamma*gradFx, gamma*rho);

% Compute the search direction DX.
DX      = X_trial - X;

% Compute lambdak and betak. 
WDX    = W*DX;
lambda = sqrt(sum(sum(WDX'.*WDX))); 
nrm_Dx = sqrt(sum(sum(DX'.*DX)));
beta   = L*nrm_Dx*nrm_Dx;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
