%
%        --------------------------------------------------------
%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%                          Version 1.0.0 (2013)
%        ---------------------------0*0*0------------------------
%
%    An implementation of the (Path-Following) Proximal Gradient/Newton
%       algorithms for solving Sparse Inverse Covariance Estimation
%                 Problems in Graphical Model Selection.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 26.11.2012.
%    Last modified: 27.05.2013.
%    Contact: quoc.trandinh@epfl.ch, volkan.cevher@epfl.ch
%
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton
%    framework for composite minimization: Graph learning without
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%
% FUNCTION: [fsol, output] = scopt_glearn_ls_pna(Shat, rho, param, ...
%                                                varargin)
%
% PURPOSE: An implementation of the Proximal-Newton algorithm for solving
%          the following Sparse Inverse Covariance Selection Problem:
%
%
%               min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%               s.t. X > 0   
%
% INPUTS:
%    Shat     : a pxp symmetric positive definite matrix, inverse of the
%               covariance matrix.
%    rho      : the penalty parameter (or regularization parameter)
%               rho should be a positive number.
%    param    : optional structure to control the algorithm
%               (see scopt_optimset() for more details).
%    varargin : user-defined inputs:
%               - varargin{1} is an initial guess for X;
%               - varargin{2} is an initial guess for X^{-1};
%               - varargin{3} is the user-defined function handle for
%                 solving the unconstrained LASSO problem.
%
% OUTPUTS:
%    fsol     : A structure containing the final solution.
%    output   : A structure containing the output information.
%
% NOTE:
%    The output information at each iteration will be printed by setting
%    param.PrintLevel = 1 or 2. If param.PrintLevel = 0 then no output
%    information is printed.
%    To evaluate the objective value, set param.isFxEval = true.
%    To turn of back-tracking line-search, set param.isBtkLS = false.
%    Type "help scopt_prints" for further output information.    
%
%

function [fsol, output] = scopt_glearn_ls_pna(Shat, rho, param, varargin)

% Check the inputs.
if nargin < 5,  Y0 = diag(diag(Shat));    else Y0 = varargin{2}; end
if isempty(Y0), Y0 = diag(diag(Shat));    end;
if nargin < 4,  X0 = diag(1./diag(Shat)); else X0 = varargin{1}; end
if isempty(X0), X0 = diag(1./diag(Shat)); end;
if nargin < 3,     param = scopt_optimset([]); end;
if isempty(param), param = scopt_optimset([]); end;

% Check the regularization parameter.
if rho < 0
    error('The penalty parameter should be positive');
end

% Check consistency issues.
if any(size(X0) ~= size(Shat))
    error('The initial point is inconsistent.');
end
if size(Shat, 1) ~= size(Shat, 2),
    error('The input matrix is not square.');
end
if norm(Shat - Shat', 'fro') > 1e-5
    error('The input matrix is nonsymmetric');
end

param           = scopt_optimset(param);
param.Algorithm = 'PN';
output.msg      = 'start';
output.info     = scopt_trace(param, []);
output.prox_cnt = 0;
output.chol_cnt = 0;
output.time     = tic;
QuadRgRadius    = 0.75*0.219224;
if param.isBtkLS, param.isFxEval = true; end

% Initalization.
if param.PrintLevel == 1, nsp = 65; else nsp = 91; end
scopt_copyright(param, nsp, 'glearn');
X_cur   = X0;
fval    = inf;
lambdaj = inf;

% The main loop.
for iter = 1:param.MaxIters
    
    % Compute the Newton-direction, Newton-decrement and evaluate the
    % objective value.
    [DX, lambdaj, fval, X0, Y0, outputj] = scopt_glearn_ntdir(X_cur, ...
        Shat, rho, X0, Y0, param, lambdaj, iter, varargin{:});
    
    % The theoretical decrement of the objective value.
    decfj   = abs(lambdaj - log(1 + lambdaj));
    
    % Compute the step size.
    alphaj = 1.0/(1.0 + lambdaj);
    
    % Count the number of prox-operations.
    if isfield(outputj, 'prox_cnt')
        n_proxj = outputj.prox_cnt;
        output.prox_cnt = output.prox_cnt + n_proxj;
    else
        n_proxj = 0;
    end
    % Count the number of Cholesky decomposition.
    if isfield(outputj, 'chol_cnt')
        output.chol_cnt = output.chol_cnt + outputj.chol_cnt;
    end
    
    % If we are in the quadratic convergence region, just do full-step.
    if param.isBtkLS && strcmpi(param.lsMode, 'BTK')
        [X_new, fval, alphaj, lsout] = scopt_glearn_ls(X_cur, DX, ...
                                       alphaj, fval, Shat, rho, param);
        % Count the number of Cholesky decomposition.
        output.chol_cnt = output.chol_cnt + lsout.chol_cnt;
    else
        % Check if the iteration is also in the quadratic convergence region.
        if lambdaj <= QuadRgRadius
            if param.PrintLevel >= 2
                fprintf('Quadratic convergence region!\n');
            end
            X_new = X_cur + DX;

        % Otherwise, perform a forward-looking procedure.
        else
            [X_new, fval, alphaj, lsout] = scopt_glearn_ls(X_cur, DX, ...
                                           alphaj, fval, Shat, rho, param);
            % Count the number of Cholesky decomposition.
            output.chol_cnt = output.chol_cnt + lsout.chol_cnt;
        end
    end
    
    % Symmetrizing to avoid the nonsymetric due to numerical issues.
    X_new = 0.5*(X_new + X_new');
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, fval, lambdaj, ...
                  n_proxj, alphaj, lambdaj, 0, 0, decfj, 0, ...
                  outputj.iter, outputj.time, outputj.error, ...
                  outputj.Lips, X_cur);
    
    % Print the iteration.
    scopt_prints(param, iter, fval, lambdaj, alphaj, decfj, ...
               output.prox_cnt, outputj.error, outputj.iter, outputj.time);
    
	% Check the stopping criterion.
    if param.TerminateType == 1 || strcmpi(param.lsMode, 'BTK')
        normDx    = sqrt(sum(sum(DX.*DX)));
        relNormDx = normDx / max(1.0, sqrt(sum(sum(X_cur.*X_cur)))); 
        if relNormDx <= param.RelTolX
            output.msg = 'Convergence acheived!';

            % Update the final iteration before breaking.
            X_cur      = X_new;
            fsol.error = relNormDx;
            break;
        end
    else
        if lambdaj <= param.AbsTol
            output.msg = 'Convergence acheived!';

            % Update the final iteration before breaking.
            X_cur      = X_new;
            fsol.error = lambdaj;
            break;
        end
    end
    
    % Terminate if the time exceed the required time
    total_time = toc(output.time);
    if total_time >= 60*60*5
        output.msg = 'Exceed the required time';
        break;
    end
    
    % Update the next iteration.
    X_cur = X_new;
    
end
% End of the main loop

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = 'Exceed the maxinum number of iterations';
end

% Get the final information.
output.time = toc(output.time);
output.iter = iter;

% Get the final solution.
fsol.x   = X_cur;
fsol.f   = fval;
fsol.lbd = lambdaj; 

% Print the final outputs.
outf = output; outf.fval = fval;
scopt_finaloutput(param, outf, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
