%
%        -------------------------------------------------------- 
%        SPARSE INVERSE COVARIANCE ESTIMATION IN GRAPHICAL MODELS
%                          Version 1.0.0 (2013)  
%        ---------------------------0*0*0------------------------
%
%    An implementation of the (Path-Following) Proximal Gradient/Newton 
%       algorithms for solving Sparse Inverse Covariance Estimation 
%                 Problems in Graphical Model Selection.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 26.11.2012.
%    Last modified: 27.05.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton
%    framework for composite minimization: Graph learning without
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%
% FUNCTION: [final_sol, output] = scopt_glearn_pfpna(Shat, rho, rhof, ... 
%                                 param, varargin)
%
% PURPOSE: An implementation of the Path-Following Proximal-Newton
%          algorithm for solving the following Sparse Inverse Covariance 
%          Selection Problem:
%
%               min{ -log(X) + trace(Shat*X) + rho*||vec(X)||_1 }
%                X > 0   
%
%          This solver computes approximately the solution of this problem
%          for rho = rho0 to rho = rhof.
%
% INPUTS:
%    Shat     : a pxp symmetric positive definite matrix, inverse of the
%               covariance matrix.
%    rho0     : The initial penalty parameter (regularization parameter).
%    rhof     : The final penalty parameter (rhof > rho0).
%               rho should be a positive number.
%    param    : optional structure to control the algorithm
%               (see scopt_optimset() for more details).
%    varargin : user-defined inputs:
%               - varargin{1} is an initial guess for X;
%               - varargin{2} is an initial guess for X^{-1};
%               - varargin{3} is percentage of the sparsity (in (0, 1)).
%               If this parameter is provided, them set 
%                   param.isSparseCheck = true.
%
% OUTPUTS:
%    final_sol: A structure containing the final solution.
%    output   : A structure containing the output information.
%
% NOTE:
%    The output information at each iteration will be printed by setting
%    param.PrintLevel = 1 or 2. If param.PrintLevel = 0 then no output
%    information is printed.
%    To evaluate the objective value, set param.isFxEval = true.
%    To turn of back-tracking line-search, set param.isBtkLS = false.
%    Type "help scopt_prints" for further output information.    
%

function [final_sol, output] = scopt_glearn_pfpna(Shat, rho0, rhof, ...
                               param, varargin)

% Check the inputs.
if nargin < 6,     Y0    = diag(diag(Shat));    else Y0 = varargin{2}; end;
if isempty(Y0),    Y0    = diag(diag(Shat));                           end;
if nargin < 5,     X0    = diag(1./diag(Shat)); else X0 = varargin{1}; end;
if isempty(X0),    X0    = diag(1./diag(Shat)); end;
if nargin < 4,     param = scopt_optimset([]);  end;
if isempty(param), param = scopt_optimset([]);  end;

% Check the regularization parameters.
if rho0 < 0 || rhof < 0
    error('The penalty parameters should be positive'); 
end

% Check consistency issues.
if any(size(X0) ~= size(Shat))
    error('The initial point is inconsistent.');
end
if size(Shat, 1) ~= size(Shat, 2),
    error('The input matrix is not square.');
end
if norm(Shat - Shat', 'fro') > 1e-5
    error('The input matrix is nonsymmetric');
end

param              = scopt_optimset(param);
param.isGradNorm   = true;
param.Algorithm    = 'PFPN';
output1.msg        = 'Phase 1 is starting';
output1.prox_cnt   = 0;
output1.info       = scopt_trace(param, []);
output1.time       = tic;
p                  = size(Shat, 1);

if param.PrintLevel == 1, nsp = 75; else nsp = 102; end
scopt_copyright(param, nsp, 'PFPN');

% Define of the value of beta in (0, 0.2).
beta       = 0.045551;
Delta_beta = (sqrt(beta) - sqrt(5)*beta)/(sqrt(5) + sqrt(beta)); 
sigma      = Delta_beta/((1 + Delta_beta)*sqrt(p));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PHASE I: Perform the Phase I to find and appropriate 
% initial point. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('%s\n', repmat('*', 1, nsp));
fprintf('*** PERFORM PHASE I ... \n');
fprintf('%s\n', repmat('*', 1, nsp));

% Initialize the main loop.
X_cur   = X0;
fval    = inf;
lambdaj = inf;

% The main loop
for iter1 = 1:param.MaxItersP1
    
    % Compute the Newton-direction, Newton-decrement and evaluate the
    % objective value.
    [DX, lambdaj, fval, X0, Y0, outputj] = scopt_glearn_ntdir(X_cur, ...
                                           Shat, rho0, X0, Y0, ...
                                           param, lambdaj, iter1);

	decfj   = lambdaj - log(1 + lambdaj);
    
    % Compute step size.
    alphaj = 1.0/(1.0 + lambdaj);
    
    % Update the next iteration.
    X_new = X_cur + alphaj*DX;
    
    % Symmetrization step to make sure that X_new is symmetric.
    X_new = 0.5*(X_new + X_new');
    
    % Count the number of prox-operations.
    if isfield(outputj, 'prox_cnt')
        n_proxj = outputj.prox_cnt;
        output1.prox_cnt = output1.prox_cnt + n_proxj;
    else
        n_proxj = 0;
    end
    
    % Record the iteration information.
    output1.info = scopt_trace(param, output1.info, fval, lambdaj, ...
                   n_proxj, alphaj, lambdaj, 0, 0, decfj, rho0, ...
                   outputj.iter, outputj.time, outputj.error, ...
                   outputj.Lips, X_cur);
               
    % Print the iteration.
    scopt_prints(param, iter1, fval, lambdaj, alphaj, decfj, ...
                 outputj.prox_cnt, outputj.error, outputj.iter, ...
                 outputj.time, rho0);
             
    % Check the stopping criterion.
    if lambdaj <= 0.75*beta
        output1.msg = 'Phase I: convergence acheived!';
        % Keep the last iteration.
        X_cur       = X_new;
        break;
    end
    
    % Update the next iteration.
    X_cur = X_new;
    
end
% End of the main loop.

% Exceed the maximum number of iterations.
if iter1 >= param.MaxItersP1
    output1.msg = 'Phase I: Exceed the maxinum number of iterations';
end

% Get the final output of Phase I.
output1.iter     = iter1;
output1.time     = toc(output1.time);
output1.fval     = fval;
final_sol.error1 = lambdaj;
final_sol.x0     = X_cur;
final_sol.f1     = fval;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PHASE II: Perform the Phase II to find the final 
% solution.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('%s\n', repmat('*', 1, nsp));
fprintf('*** PERFORM PHASE II ... \n');
fprintf('%s\n', repmat('*', 1, nsp));

output2.msg      = 'start phase 2';
output2.info     = scopt_trace(param, []);
output2.prox_cnt = 0;
output2.time     = tic;

% Initialization.
X0      = X_cur;
tau     = 1/rho0; % Penalty parameter.
lambdak = inf;

% Main loop of Phase II.
for iter2 = 1:param.MaxItersP2
    
    % Update the penalty parameter tau and then compute rho.
    if iter2 > 1 && isfield(outputk, 'dualGradNrm')
        nu12  = outputk.dualGradNrm;
        if nu12 > sqrt(p), nu12 = sqrt(p); end
        sigma = Delta_beta/((1 + Delta_beta)*nu12);
    end
    tau   = (1 + sigma)*tau;
    rho   = 1/tau;
    
    % Compute the Newton-direction, Newton-decrement and evaluate the
    % objective value.
    [DX, lambdak, fval, X0, Y0, outputk] = scopt_glearn_ntdir(X_cur, ...
                                           Shat, rho, X0, Y0, ...
                                           param, lambdak, iter2);
                                       
    % Compute the theoretical decrease of the objective function.
	decfk   = lambdak - log(1 + lambdak);
    
    % Compute step size.
    alphak  = 1;
    
    % Update the trial point.
    X_new   = X_cur + DX;
    
    % Symmetrization step to make sure that X_new is symmetric.
    X_new = 0.5*(X_new + X_new');
    
    % Count the number of prox-operations.
    if isfield(outputk, 'prox_cnt')
        n_proxk = outputk.prox_cnt;
        output2.prox_cnt = output2.prox_cnt + n_proxk;
    else
        n_proxk = 0;
    end
    
    % Record the iteration information.
    output2.info = scopt_trace(param, output2.info, fval, lambdak, ...
                               n_proxk, alphak, lambdak, 0, 0, decfk, ...
                               rho, outputk.iter, outputk.time, ...
                               outputk.error, outputk.Lips, X_cur);
                           
    % Print the iteration.
    scopt_prints(param, iter2, fval, lambdak, alphak, decfk, ...
                 outputk.prox_cnt, outputk.error, outputk.iter, ...
                 outputk.time, rho);

    % Terminate whenever you reach the desired solution.
    if param.isSparseCheck
        if nargin < 7, percent = 0.05; else percent = varargin{3}; end;
        if percent < 0 || percent > 1
            error('The sparsity should be in (0, 1]');
        end
        lbnnz = 0.99*percent*p^2;
        ubnnz = 1.01*percent*p^2;
        if lbnnz <= nnzj && nnzj <= ubnnz
            output2.msg = 'Phase II: Reach the desired solution!';
            break;
        end
    end
    
    % Check the stopping criterion.
    if rho <= rhof
        output2.msg = 'Phase II: convergence acheived!';
        % Keep the last iteration.
        X_cur       = X_new;
        break;
    end
    
    % Update the next iteration.
    X_cur = X_new;
    
end % End of the main loop.

% Exceed the maximum number of iterations.
if iter2 >= param.MaxItersP2
    output2.msg = 'Phase II: Exceed the maxinum number of iterations';
end

% Get the final information.
output2.fval    = fval;
output2.iter    = iter2;
output2.lambda  = lambdak;
output2.time    = toc(output2.time);
output2.rho     = rho;

output.p1       = output1;
output.p2       = output2;

% Get the final solution.
final_sol.x     = X_cur;
final_sol.f     = fval;
final_sol.error = lambdak;

% Print the final outputs.
outf.iter1    = output1.iter; outf.iter2    = output2.iter; 
outf.prox_cnt = output1.iter; outf.prox_cnt = output2.iter; 
outf.time1    = output1.time; outf.time2    = output2.time;
outf.msg      = output2.msg;  outf.fval     = output2.fval;
scopt_finaloutput(param, outf, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%