%
%                     --------------------------------- 
%                     LOGISTIC REGRESSION v.1.0. (2013)
%                     --------------------------------- 
%
%    An implementation of the Proximal Gradient/Newton algorithms for 
%        solving Sparse (Multinomial) Logistic Regression Problems.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher.
%    Date: 16.04.2013.
%    Last modified: 16.04.2013.
%    Contact: quoc.trandinh@epfl.ch
%
% REFERENCES:
%    [1]. Quoc Tran Dinh, Volkan Cevher and Yen-Huan Li: Minimization of 
%         self-concordant-like functions, Tech. Report., LIONS, EPFL, 2013.
%    [2]. Krishnapuram, B., Figueiredo, M., Carin, L., & Hartemink, A.
%         (2005) ?Sparse Multinomial Logistic Regression: Fast Algorithms 
%         and Generalization Bounds.? IEEE Transactions on Pattern Analysis 
%         and Machine Intelligence (PAMI), 27, June 2005. pp. 957?968.
%
% FUNCTION: [xsol, output] = scopt_logistic_pga(What, yhat, bias, rho, ...
%                                               x0, param, varargin)
%
% PURPOSE: An implementation of the proximal-gradient algorithm 
%          for solving the logistic regression problem of the form:
%
%                min{ L(x, z) + tau*|x|_1 }
%
%          where 
%                L(x, z) = (1/N)*sum_{i=1}^Nlog(1 + exp(-(ai'*x + bi))).
%          Here, ai = yi*wi and bi = yi*bias.
%          
% INPUTS:
%     What      : Training matrix (N x p) (N is the number of samples).
%     yhat      : Classification vector (N x 1).
%     bias      : bias term (scalar).
%     rho       : The penalty parameter in the regularization term.
%     x0        : An nx1 initial point for vector x.
%     param     : optional structure to control the algorithm,
%                 (see scopt_optimset() for more details).
%     varargin  : user defined inputs (if required).
%     
% OUTPUTS:
%     final_sol : a structure containing the final solution.
%     output    : a structure containing the output information.
%
% NOTE: See scopt_logistic_exam.m for more details of using this solver.
%

function [final_sol, output] = scopt_logistic_pga( What, yhat, bias, ...
                                                   rho, param, varargin)

% Check the input parameters.
if nargin < 6
    x0 = [];
else
    x0 = varargin{1}; 
end
if nargin < 5,     param    = scopt_optimset([]);  end
if isempty(param), param    = scopt_optimset([]);  end
if rho < 0 || isnan(rho) || isinf(rho) || ~isnumeric(rho)
    error('The penalty parameter should be a finite positive number'); 
end

yhat = yhat(:);       % Make sure that yhat is a column vector.
N    = size(What, 1); % Number of training samples/examples.
p    = size(What, 2); % Number of variables.
if length(yhat) ~= N, error('Size is inconsistent'); end;

fx_list         = [];
param           = scopt_optimset(param);
param.Algorithm = 'PG';
if param.isAccelerating, param.Algorithm = 'FPG'; end
output.nsize    = [p, N];
output.msg      = 'start';
output.info     = scopt_trace(param, []);
output.time     = tic;
output.prox_cnt = 0;

if param.PrintLevel == 1, nsp = 65; else nsp = 91; end
scopt_copyright(param, nsp, 'logistic');

% Generate the input matrix Ahat and vector bhat.
Ahat   = repmat(yhat, 1, p).*What;
NAThat = Ahat'/N; 
bhat   = bias*yhat;

% Compute the self-concordent-like parameter Mf.
Mf     = full(max(sqrt(sum(Ahat.*Ahat, 2))));

% Generate an initial point.
if isempty(x0) || ~any(isnumeric(x0)) || any(isnan(x0)) || any(isinf(x0))
    x0 = zeros(size(Ahat, 2), 1); 
else
    x0 = x0(:);
end

% Intialization step.
L0          = 0.05;
x_old       = []; 
grad_fx_old = [];
theta       = 1;
x_cur       = x0;
z_cur       = x_cur;

% The main loop.
for iter = 1:param.MaxIters
    
    % Evaluate the objective value and its gradient.
    [fx_cur, grad_fx, qx] = scopt_logistic_fx_eval(x_cur, Ahat, ...
                            bhat, N, NAThat, param.isFxEval);
	if param.isAccelerating 
        % Evaluate the objective value at the point z_cur.
        fx_cur = scopt_logistic_fx_eval(z_cur, Ahat, ...
                 bhat, N, NAThat, param.isFxEval);
        fx_cur  = fx_cur + rho*sum(abs(z_cur));
    else
        fx_cur  = fx_cur + rho*sum(abs(x_cur));
    end
    
    % Evalute the initial Lipschitz constant by the Barzilai-Borwein rule.
    Lips_cur =  scopt_compute_BBstep(x_old, x_cur, grad_fx_old, ...
                grad_fx, param, L0);

    % Compute the step-size.
    gamma   = 1/Lips_cur;

    % Perform the proximal gradient step.
    x_trial = scopt_soft_threshold_oper(x_cur - gamma*grad_fx, gamma*rho);

    % Count the prox-operations.
    output.prox_cnt = output.prox_cnt + 1;

    % Compute the proximal-gradient search direction dx.
    dx_cur  = x_trial - x_cur;

    % Compute three quantities: r, lambda and beta. 
    [nrm_dx, r_cur, beta, lambda] = compute_step_quantities(...
                                    dx_cur, Ahat, qx, N, Mf, Lips_cur);

    % Check the condition.
    if param.isBtkLS
        lbd_dx  = (lambda/nrm_dx)^2;
        for itbtk = 1:param.BtkMaxIters

            % Terminate if the condition is satisfied.
            if param.isAccelerating
                if Lips_cur >= 2*(exp(r_cur) - 1 - r_cur)/r_cur^2*lbd_dx
                    break;
                end
                % Back-track the Lipschitz constant by the factor BETA.
                Lips_cur = Lips_cur*param.btkSlop;
            else
                ratio    = beta^2*r_cur/lambda^2;
                alpha    = log(1 + ratio)/r_cur;
                if beta <= lambda || alpha <= 1
                    break;
                end
                % Back-track the Lipschitz constant by the factor BETA.
                Lips_cur    = Lips_cur/param.btkSlop;
            end

            % Compute the step-size.
            gamma   = 1/Lips_cur;

            % Compute the proximal gradient step.
            x_trial = scopt_soft_threshold_oper(x_cur - gamma*grad_fx, gamma*rho);

            % Count the prox-operations.
            output.prox_cnt = output.prox_cnt + 1;

            % Compute the proximal-gradient search direction dx.
            dx_cur  = x_trial - x_cur;

            % Recompute three quantities: r, lambda and beta. 
            [nrm_dx, r_cur, beta, lambda] = compute_step_quantities(...
                                            dx_cur, Ahat, qx, N, Mf, Lips_cur);
        end
    end
    
    % Compute the step size.
    ratio   = beta^2*r_cur/lambda^2;
    alpha   = log(1 + ratio)/max(r_cur, eps);
    
    % Make sure that alphak <= 1.
    alpha   = min(alpha, 1.0);
    
    % Update the step z.
    z_next  = x_cur + alpha*dx_cur;
    
    % Compute the (theoretical) decrease of the objective function.
    fx_dec  = beta^2*((1 + 1/ratio)*log(1 + ratio) - 1)/max(r_cur, eps);
    
    % Compute the relative error.
    err_cur = nrm_dx / max(1.0, sqrt(sum(x_cur.*x_cur)));
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, fx_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, beta, ...
                              r_cur, fx_dec, Lips_cur, x_cur);

    % Print the iteration.
    scopt_prints(param, iter, fx_cur, err_cur, lambda, beta, alpha, ...
                 fx_dec, output.prox_cnt, Lips_cur);
    
    % Update new iterative points.
    if ~param.isAccelerating
        x_next     = z_next;
        theta_next = theta;
    else
        % Updating the next iteration.
        theta_next = 0.5*(1 + sqrt(4*theta^2 + 1));
        factor     = (theta - 1)/theta_next;
        x_next     = z_next + factor*(z_next - z_cur);
    end
        
    % Check stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, fx_list] = scopt_fx_terminate(fx_cur, fx_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop
            output.msg = 'The objective values does not improve significantly';
            % Keep the last iteration.
            z_cur = z_next;
            break;
        end
    end
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX
        output.msg = 'Convergence achieved!';
        % Keep the last iteration.
        z_cur = z_next;
        break;
    end
    
    % Save the iteration.
    x_old       = x_cur;
    grad_fx_old = grad_fx;
    fx_old      = fx_cur;
    
    % Move to the next step.
    x_cur = x_next;
    z_cur = z_next;
    theta = theta_next;
    
    % Restart if necessary.
    if param.isAccelerating
        if strcmpi(param.isRestart, 'fixed')
            if mod(iter, param.nRestartIters) == 0
                theta = 1;
                x_cur = z_cur;
            end
        elseif strcmpi(param.isRestart, 'fx')
            if iter > 1 && fx_cur > fx_old
                theta = 1;
                x_cur = z_cur;
            end
        end
    end
end

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', char(13), ...
                  'Increase the number of iterations to improve ...', ...
                  'Or change the tolerance param.abs_tol.'];
end

% Evaluate the objective function at the last iteration.
fx_cur = scopt_logistic_fx_eval(x_cur, Ahat, bhat, N, NAThat, param.isFxEval);
fx_cur = fx_cur + rho*sum(abs(x_cur));
                        
% The final output.
output.time     = toc(output.time);
output.iter     = iter;
output.fval     = fx_cur;
output.error    = err_cur;
output.lbd      = lambda;
output.r        = r_cur;
output.beta     = beta;
output.x0       = x0;
output.Mf       = Mf;

% The final solution.
final_sol.x     = z_cur;
final_sol.f     = fx_cur;
final_sol.error = err_cur;

% Print the final output.
scopt_finaloutput(param, output, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [nrm_dx, r_cur, beta, lambda] = compute_step_quantities(...
%%%                                           dx_cur, Ahat, qx, N, Mf)
%%% PURPOSE:  Compute norm(dx), r, beta and lambda.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [nrm_dx, r_cur, beta, lambda] = compute_step_quantities(...
                                         dx, Ahat, qx, N, Mf, Lips)

% Compute three quantities: r, lambda and beta. 
nrm_dx  = sqrt( sum(dx.*dx) );
r_cur   = Mf*nrm_dx;
beta    = sqrt(Lips)*nrm_dx;

% Compute lambda.
Adx    = Ahat*dx;
nrmHd2 = sum(qx.*Adx.*Adx); 
lambda = sqrt( nrmHd2/N );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%