%
%                     --------------------------------- 
%                     LOGISTIC REGRESSION v.1.0. (2013)
%                     --------------------------------- 
%
%    An implementation of the Proximal Gradient/Newton algorithms for 
%        solving Sparse (Multinomial) Logistic Regression Problems.
%
% INFORMATION:
%    By Quoc Tran Dinh,
%       Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher.
%    Date: 16.04.2013.
%    Last modified: 16.04.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Quoc Tran Dinh, Volkan Cevher and Yen-Huan Li: Minimization of 
%         self-concordant-like functions, Tech. Report., LIONS, EPFL, 2013.
%    [2]. Krishnapuram, B., Figueiredo, M., Carin, L., & Hartemink, A.
%         (2005) ?Sparse Multinomial Logistic Regression: Fast Algorithms 
%         and Generalization Bounds.? IEEE Transactions on Pattern Analysis 
%         and Machine Intelligence (PAMI), 27, June 2005. pp. 957?968.
%
% FUNCTION: [xsol, output] = scopt_logistic_pna(What, yhat, bias, ...
%                                               rho, x0, param, varargin)
%
% PURPOSE: 
%    An implementation of the proximal-(quasi-)Newton algorithm for solving 
%    the logistic regression problem of the form:
%
%                min{ L(x, z) + tau*|x|_1 }
%
%    where 
%          L(x, z) = (1/N)*sum_{i=1}^Nlog(1 + exp(-(ai'*x + bi))).
%    Here, ai = yi*wi and bi = yi*bias.    
%          
% INPUTS:
%     What     : Training matrix (p x N - matrix).
%     Yhat     : Classification matrix (m x N - matrix).
%     rho      : The penalty parameter in the regularization term.
%     X0       : An initial point (p x m - matrix) for vector X.
%     param    : optional structure to control the algorithm,
%                see scopt_optimset() for more details.
%     varargin : user defined inputs (varargin{1} is a scaling factor
%                TN, default value is TN = N).  
%     
% OUTPUTS:
%     fsol     : a structure containing the final solution.
%     output   : a structure containing the output information.
%
% NOTE: See scopt_mnlogistic_exam.m for the use of this solver.
%

function [fsol, output] = scopt_logistic_pna(What, yhat, bias, rho, ...
                                             x0, param, varargin)

% Check the inputs.
if nargin < 6,     param = scopt_optimset([]);      end
if isempty(param), param = scopt_optimset([]);      end
if nargin < 5,     x0    = zeros(size(What, 2), 1); end
if isempty(x0),    x0    = zeros(size(What, 2), 1); end
if rho < 0 || isnan(rho) || isinf(rho)
    error('The penalty parameter should be positive'); 
end

N = size(What, 1);    % Number of training samples/examples.
p = size(What, 2);    % Number of variables.
if length(yhat) ~= N, error('Size is inconsistent'); end;

fx_list         = [];
param           = scopt_optimset(param);
param.Algorithm = 'PN';
output.nsize    = [p, N];
output.msg      = 'start';
output.info     = scopt_trace(param, []);
output.time     = tic;
output.prox_cnt = 0;

if param.PrintLevel == 1, nsp = 65; else nsp = 91; end
scopt_copyright(param, nsp, 'logistic');

% Generate the input matrix Ahat and vector bhat.
Ahat   = repmat(yhat, 1, p).*What;
NAThat = Ahat'/N; 
bhat   = bias*yhat;

% Compute the parameter Mf.
Mf     = full(max(sqrt(sum(Ahat.*Ahat, 2))));

% Intialization step.
x_old   = []; 
dfx_old = [];
H_old   = [];
lBFGS   = [];
x_cur   = x0(:);

% The main loop.
for iter = 1:param.MaxIters
    
    [dx_cur, lambda, H_cur, dfx_cur, fx_cur, lBFGS, output_cur] = ...
          scopt_logistic_ntdir(x_cur, Ahat, bhat, NAThat, N, rho, ...
                               dfx_old, x_old, H_old, lBFGS, param);

    % Evaluate the objective value at xk, if required.
    if param.isFxEval
        fx_cur = fx_cur + rho*sum(abs(x_cur));
    end
    
    % Accumulate the number of prox-operations.
    if isfield(output_cur, 'prox_cnt')
        output.prox_cnt = output.prox_cnt + output_cur.prox_cnt;
    end
    
	% Compute norm(dx) and r_cur.
    nrm_dx  = sqrt( sum(dx_cur.*dx_cur) );
    r_cur   = Mf*nrm_dx;
    
    % Compute the step size.
    alpha   = log(1 + r_cur)/max(r_cur, eps);
    
    % Make sure that alpha <= 1.
    alpha   = min(alpha, 1.0);
    
    % Update the next step.
    x_next  = x_cur + alpha*dx_cur;
    
    % Compute the decrease of the objective function.
    dec_fx  = lambda^2*((1 + 1/r_cur)*log(1 + r_cur) - 1)/max(r_cur, eps);
    
    % Compute the relative error.
    err_cur = nrm_dx / max(1.0, sqrt(sum(x_cur.*x_cur)));
   
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, fx_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, 0, 0, ...
                              dec_fx, 0, output_cur.iter, output_cur.time, ...
                              output_cur.error, output_cur.Lips, x_cur);
    % Print the iteration.
    scopt_prints(param, iter, fx_cur, err_cur, alpha, dec_fx, ...
                 output.prox_cnt, output_cur.error, output_cur.iter, ...
                 output_cur.time);
    
    % Check stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, fx_list] = scopt_fx_terminate(fx_cur, fx_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop
            output.msg = 'The objective values does not improve significantly';
            % Keep the last iteration.
            x_cur = x_next;
            break;
        end
    end
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX 
        output.msg  = 'Convergence achieved!';
        % Keep the last iteration before leaving the loop.
        x_cur = x_next;
        break;
    end
    
    % Go to the next iteration.
    x_old   = x_cur;
    x_cur   = x_next;
    H_old   = H_cur;
    dfx_old = dfx_cur;
end

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', char(13), ...
                  'Increase the number of iterations to improve ...', ...
                  'Or change the tolerance param.abs_tol.'];
end

% Evaluate the objective function at the last iteration.
fx_cur = scopt_logistic_fx_eval(x_cur, Ahat, bhat, N, NAThat, param.isFxEval);
fx_cur = fx_cur + rho*sum(abs(x_cur));
                        
% The final output.
output.time = toc(output.time);
output.iter = iter;

% The final solution.
fsol.x      = x_cur;
fsol.f      = fx_cur;
fsol.error  = lambda;

% Print the final outputs.
outf = output; outf.fval = fx_cur; 
scopt_finaloutput(param, outf, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%