% 
%                   ------------------------------------ 
%                   MAX-NORM IN CLUSTERING v.1.0. (2013)
%                   ------------------------------------ 
%
%    An implementation of the Path-following Proximal-Newton algorithm for 
%                  solving Max-Norm Clustering Problems.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 16.04.2013
%    Last modified: 16.04.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: No heavy lifting 
%         required: A proximal path following scheme for conic programming.
%    	  Tech. Report, LIONS, EPFL, Lausanne, 2013.
%    [2]. A. Jalali and N. Srebro. Clustering using max-norm constrained 
%         optimization. In Proc. of International Conference on Machine 
%         Learning (ICML2012), pages 1?17, 2012.
%
% FUNCTION: [Ksol, Lsol, Rsol, output] = scopt_maxnorm_pfpna(AK, ...
%                                        K0, L0, R0, ...
%                                        tau0, uLi, uRi, param)
%
% PURPOSE: This solver solves the following conic program arising from 
%          Clustering problem:
%
%                   min_{K, L, R} ||vec(K ? AK)||_1
%                   s.t. [L,   K] 
%                        |      | >= 0,
%                        [K^T, R]      
%                        L(i,i) <= 1, R(i, i) <= 1, i = 1, 2, ..., p.
%    
% INPUTS:
%   AK           : An affinity matrix.
%   uLi, uRi     : The upper bound of the diagonal entries of L and R 
%                  (default = 1).
%   K0, L0, R0   : Initial points for K, L and R, resp.
%   tau0         : An initial value of the penalty parameter
%                  (default = 0.1).
%   param        : optional structure to control the algorithm
%                  (see scopt_optimset() for more details).
%
% OUTPUTS:
%   final_sol    : A structure containing the final solution.
%   output       : A structure containing the output information.
%
% NOTE: For using this solver, see scopt_maxnorm_exam.m
%
%
function [final_sol, output] = scopt_maxnorm_pfpna(AK, uLi, uRi, ...
                               K0, L0, R0, tau0, param)

% Check the inputs.
if nargin < 8,     param = scopt_optimset([]); end
if isempty(param), param = scopt_optimset([]); end
if nargin < 7,     tau0  = 0.1;                end
if isempty(tau0),  tau0  = 0.1;                end
if tau0 < 0
    error('The barrier parameter should be positive'); 
end
if nargin < 1, error('At least one input is required'); end
p1 = size(AK, 1);
if nargin < 6,   R0  = eye(p1);   end
if isempty(R0),  R0  = eye(p1);   end
if nargin < 5,   L0  = eye(p1);   end
if isempty(L0),  L0  = eye(p1);   end
if nargin < 4,   K0  = zeros(p1); end
if isempty(K0),  K0  = zeros(p1); end
if nargin < 3,   uRi = 1;         end
if isempty(uRi), uRi = 1;         end
if nargin < 2,   uLi = 1;         end
if isempty(uLi), uLi = 1;         end

% Initialize the output parameter.
param              = scopt_optimset(param);
param.isGradNorm   = true;
param.Algorithm    = 'PFPN';
output1.msg        = 'Phase 1 is starting';
output1.prox_cnt   = 0;
output1.info       = scopt_trace(param, []);
output1.time       = tic;
p                  = size(AK, 1) + size(AK, 2);

if param.PrintLevel == 1, nsp = 75; else nsp = 102; end
scopt_copyright(param, nsp, 'PFPN');

% Define of the value of beta in (0, 0.2).
beta       = 0.045551;
Delta_beta = (sqrt(beta) - sqrt(5)*beta)/(sqrt(5) + sqrt(beta)); 
sigma      = Delta_beta/((1 + Delta_beta)*sqrt(p));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PHASE I: Perform the Phase I to find and appropriate 
% initial point. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('%s\n', repmat('*', 1, nsp));
fprintf('*** PERFORM PHASE I ... \n');
fprintf('%s\n', repmat('*', 1, nsp));

% Initialize the main loop.
K     = K0;
L     = L0;
R     = R0;
fvalj = nan;
tau   = tau0;

% The main loop of Phase I.
for iter1 = 1:param.MaxItersP1
    
    % Evaluate the objective value if required.
    if param.isFxEval
        fvalj = scopt_maxnorm_fx_eval(K, L, R, AK, tau);
    end

    % Form matrix Ax and compute its inverse, evaluate the Lips const.
    [Ax, iAx, cholfailed] = compute_inverse_Ax(K, L, R, param);
    
    % If the Choleskey factorization is failed, then terminate.
    if cholfailed
        error('The Choleskey decomposition is failed');
    end
    
    % Solve the dual subproblem and recover the primal solution.
    [DK, DL, DR, outputj] = scopt_maxnorm_pFISTA(Ax, iAx, ...
                            K, L, R, K, L, R, tau, AK, uLi, uRi, param);
                        
    % Compute lambdak (the norm of the proximal-gradient mapping).
    lambdaj = compute_PN_decrement(iAx, DK, DL, DR);
    
    % The objective value decreases at least.
    decfj   = lambdaj - log(1 + lambdaj);
    
    % Compute the step size.
    alphaj  = 1.0/(1.0 + lambdaj);
    
    % Update the next iteration.
    K_next = K + alphaj*DK;
    L_next = L + alphaj*DL;
    R_next = R + alphaj*DR;
    
    % Count the number of prox-operations.
    if isfield(outputj, 'prox_cnt')
        n_proxj = outputj.prox_cnt;
        output1.prox_cnt = output1.prox_cnt + n_proxj;
    else
        n_proxj = 0;
    end
    
    % Record the iteration information.
    output1.info = scopt_trace(param, output1.info, fvalj, lambdaj, n_proxj, ...
                               alphaj, lambdaj, 0, 0, decfj, tau0, ...
                               outputj.iter, outputj.time, outputj.error, ...
                               outputj.Lips, {K, L, R});
    % Print the iteration.
    scopt_prints(param, iter1, fvalj, lambdaj, alphaj, decfj, ...
                 output1.prox_cnt, outputj.error, outputj.iter, ...
                 outputj.time, tau0);
             
    % Check the stopping criterion.
    if lambdaj <= 0.75*beta
        output1.msg  = 'Phase I: convergence acheived!';
        % Keep the last iteration.
        K = K_next; L = L_next; R = R_next;
        break;
    end
    
    % Update the next iteration.
    K = K_next; L = L_next; R = R_next;

end
% End of the main loop

% Exceed the maximum number of iterations.
if iter1 >= param.MaxItersP1
    output1.msg = 'Phase I: Exceed the maxinum number of iterations';
end

% Get the output information.
output1.iter     = iter1;
output1.time     = toc(output1.time);
output1.fval     = fvalj;
final_sol.error1 = lambdaj;
final_sol.K0     = K;
final_sol.L0     = L;
final_sol.R0     = R;
final_sol.f1     = fvalj;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PHASE II: Perform the Phase II to find the final 
% solution.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('%s\n', repmat('*', 1, nsp));
fprintf('*** PERFORM PHASE II ... \n');
fprintf('%s\n', repmat('*', 1, nsp));

output2.msg      = 'Start phase 2';
output2.info     = scopt_trace(param, []);
output2.prox_cnt = 0;
output2.time     = tic;

% Initialize the main loop of Phase II.
fvalk = nan;

% The main loop of Phase II.
for iter2 = 1:param.MaxItersP2
    
    % Evaluate the objective value if required.
    if param.isFxEval
        fvalk = scopt_maxnorm_fx_eval(K, L, R, AK, tau);
    end
    
    % Update the penalty parameter tau.
    tau = (1 - sigma)*tau;
    
    % Form matrix Ax and compute its inverse, evaluate the Lips const.
    [Ax, iAx, cholfailed] = compute_inverse_Ax(K, L, R, param);
    
    % If the Choleskey factorization fails, then terminate.
    if cholfailed
        error('The Choleskey decomposition failed');
    end
    
    % Solve the dual subproblem and recover the primal solution.
    [DK, DL, DR, outputk] = scopt_maxnorm_pFISTA(Ax, iAx, ...
                            K, L, R, K, L, R, tau, AK, uLi, uRi, param);
                        
    % Compute lambdak (the norm of the proximal-gradient mapping).
    lambdak = compute_PN_decrement(iAx, DK, DL, DR);
    
    decfk   = lambdak - log(1 + lambdak);
    
    % Perform the full-step size.
    alphak  = 1;
    
    % Count the number of prox-operations.
    if isfield(outputk, 'prox_cnt')
        n_proxk = outputk.prox_cnt;
        output2.prox_cnt = output2.prox_cnt + n_proxk;
    else
        n_proxk = 0;
    end
    
    % Record the iteration information.
    output2.info = scopt_trace(param, output2.info, fvalk, lambdak, ...
                               n_proxk, alphak, lambdak, 0, 0, decfk, ...
                               tau, outputk.iter, outputk.time, ...
                               outputk.error, outputk.Lips, {K, L, R});
    % Print the iteration.
    scopt_prints(param, iter2, fvalk, lambdak, alphak, decfk, ...
                 output2.prox_cnt, outputk.error, outputk.iter, ...
                 outputk.time, tau);

    % Check the stopping criterion.
    if tau <= param.TolParam && lambdak <= param.AbsTol
        output2.msg  = 'Phase II: convergence acheived!';
        break;
    end
    
    % Update the next iteration by Full-Proximal-Newton iteration.
    K = K + DK;
    L = L + DL;
    R = R + DR;

end
% End of the main loop.

% Exceed the maximum number of iterations.
if iter2 >= param.MaxItersP2
    output2.msg = 'Phase II: Exceed the maxinum number of iterations';
end

% Get the final information.
output2.iter    = iter2;
output2.fval    = fvalk;
output2.lambda  = lambdak;
output2.time    = toc(output2.time);
output2.tau     = tau;

output.p1 = output1;
output.p2 = output2;

% The final solution.
final_sol.K     = K;
final_sol.L     = L;
final_sol.R     = R;
final_sol.f     = fvalk;
final_sol.error = lambdak;

% Print the final outputs.
outf.iter1    = output1.iter; outf.iter2    = output2.iter; 
outf.prox_cnt = output1.iter; outf.prox_cnt = output2.iter; 
outf.time1    = output1.time; outf.time2    = output2.time;
outf.msg      = output2.msg;  outf.fval     = output2.fval;
scopt_finaloutput(param, outf, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: lambda = compute_PN_decrement(iAx, DK, DL, DR) 
%%% PURPOSE:  Compute the proximal Newton decrement lambda.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function lambda = compute_PN_decrement(iAx, DK, DL, DR)

Md     = [DL, DK; DK', DR];
W      = Md*iAx;
lambda = sqrt( trace(W*W) );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [Ax, iAx, cholfailed] = compute_inverse_Ax(K, L, R, param)
%%% PURPOSE:  Form matrix Ax, compute its inverse and then evaluate the 
%%%           Lipschitz constant of the Hessian of f.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Ax, iAx, cholfailed] = compute_inverse_Ax(K, L, R, param)

% Form matrix Ax. 
Ax   = [L, K; K', R];

% Cholesky decomposition.
cholfailed = 0;
[Lo, ret]  = chol(Ax);

% Check if Cholesky fails.
if ret ~= 0 
    [V_cur, D_cur] = eig(Ax);
    Dp_cur         = max( real(diag(D_cur)), 1 - 0.9999);
    Ax             = V_cur*diag(Dp_cur)*V_cur';
    Lo             = V_cur*diag(sqrt(Dp_cur));
end

% Compute inverse of Ax.
Ip   = eye(size(Ax));
iLo  = Lo\Ip;
iAx  = iLo*iLo';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%