%
%                     --------------------------------- 
%                     LOGISTIC REGRESSION v.1.0. (2013)
%                     --------------------------------- 
%
%    An implementation of the Proximal Gradient/Newton algorithms for 
%        solving Sparse (Multinomial) Logistic Regression Problems.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher.
%    Date: 16.04.2013.
%    Last modified: 16.04.2013.
%    Contact: quoc.trandinh@epfl.ch
%
% REFERENCES:
%    [1]. Quoc Tran Dinh, Volkan Cevher and Yen-Huan Li: Minimization of 
%         self-concordant-like functions, Tech. Report., LIONS, EPFL, 2013.
%    [2]. Krishnapuram, B., Figueiredo, M., Carin, L., & Hartemink, A.
%         (2005) ?Sparse Multinomial Logistic Regression: Fast Algorithms 
%         and Generalization Bounds.? IEEE Transactions on Pattern Analysis 
%         and Machine Intelligence (PAMI), 27, June 2005. pp. 957?968.
%
% FUNCTION: [Xsol, output] = scopt_mnlogistic_pga(What, Yhat, rho, X0, ...
%                                                 param, varargin)
%
% PURPOSE: 
%    An implementation of the proximal-gradient algorithm for solving 
%    the multinomial logistic regression problem of the form:
%
%                min{ (1/TN)*L(x, z) + rho/(sqrt(TN))*|x|_1 }
%    where 
%         L(x, z) = sum_{j=1}^N log(1 + sum_{i=1}^m exp(wj'*xi))) 
%                 - sum_{j=1}^N sum_{i=1}^mYij*. 
%          
% INPUTS:
%     What     : Training matrix (p x N - matrix).
%     Yhat     : Classification matrix (m x N - matrix).
%     rho      : The penalty parameter in the regularization term.
%     X0       : An initial point (p x m - matrix) for vector X.
%     param    : optional structure to control the algorithm,
%                see scopt_optimset() for more details.
%     varargin : user defined inputs (varargin{1} is a scaling factor
%                TN, default value is TN = N).  
%     
% OUTPUTS:
%     final_sol: a structure containing the final solution.
%     output   : a structure containing the output information.
%
% NOTE: See scopt_mnlogistic_exam.m for the use of this solver.
%
%
function [final_sol, output] = scopt_mnlogistic_pga(What, Yhat, rho, ...
                                                    X0, param, varargin)

% Check the inputs.
if nargin < 5,     param = scopt_optimset([]);                end
if isempty(param), param = scopt_optimset([]);                end
if nargin < 4,     X0    = zeros(size(What,1), size(Yhat,1)); end
if isempty(X0),    X0    = zeros(size(What,1), size(Yhat,1)); end
if rho < 0 || isnan(rho) || isinf(rho)
    error('The penalty parameter should be positive'); 
end

p  = size(What, 1); % Number of variables.
N  = size(What, 2); % Number of training samples/examples.
mx = size(Yhat, 1);

% Define the scaling factor TN (default value is N).
if nargin < 6
    TN = N; 
else
    if isempty(varargin{1}) 
        TN = N;
    else
        TN = varargin{1};
    end
end

% Initialize the output.
FX_list         = [];
param           = scopt_optimset(param);
param.Algorithm = 'PG';
if param.isAccelerating, param.Algorithm = 'FPG'; end
output.nsize    = [p, N];
output.msg      = 'start';
output.info     = scopt_trace(param, []);
output.time     = tic;
output.prox_cnt = 0;

% Print the header message.
if param.PrintLevel == 1, nsp = 65; else nsp = 91; end
scopt_copyright(param, nsp, 'logistic');

% Compute the parameter Mf.
Mf      = sqrt(6)*full(max(sqrt(sum(What.*What, 1))));

% Intialization step.
mWhat   = repmat(What, mx, 1);
L0      = 0.05;
X_old   = []; 
DFX_old = [];
FX_cur  = inf;
theta   = 1;
X_cur   = X0;
Z_cur   = X_cur;

% The main loop.
for iter = 1:param.MaxIters
    
    % Evaluate the objective value and its gradient.
    [FX_cur, DFX_cur, Phat] = scopt_mnlogistic_fx_eval(X_cur, What, ...
                              Yhat, TN, mWhat);
    
    if param.isAccelerating 
        % Evaluate the objective value at the point z_cur.
        FX_cur = scopt_mnlogistic_fx_eval(Z_cur, What, ...
                 Yhat, TN, mWhat);
        FX_cur  = FX_cur + rho*sum(sum(abs(Z_cur)));
    else
        FX_cur  = FX_cur + rho*sum(sum(abs(X_cur)));
    end
    
    % Evalute an initial Lipschitz constant by using Barzilai-Borwein
    % rule.
    Lips_cur = scopt_compute_BBstep(X_old, X_cur, DFX_old, DFX_cur, param, L0);

    % Compute the first step-size.
    gamma   = 1/Lips_cur;

    % Perform the proximal gradient step.
    X_trial = scopt_soft_threshold_oper(X_cur - gamma*DFX_cur, gamma*rho);

    % Count the prox-operations.
    output.prox_cnt = output.prox_cnt + 1;

    % Compute the proximal-gradient search direction DX.
    DX_cur  = X_trial - X_cur;

    % Compute three quantities: r, lambda and beta. 
    [nrm_DX, r_cur, beta, lambda] = compute_step_quantities(...
                                    DX_cur, What, Phat, TN, Mf, Lips_cur);

	% Check the condition.
    if param.isBtkLS
        lbd_DX  = (lambda/nrm_DX)^2;
        for itbtk = 1:param.BtkMaxIters

            % Terminate if the condition is satisfied.
            if param.isAccelerating
                if Lips_cur >= 2*(exp(r_cur) - 1 - r_cur)/r_cur^2*lbd_DX
                    break;
                end
                % Back-track the Lipschitz constant by the factor BETA.
                Lips_cur = Lips_cur*param.btkSlop;
            else
                ratio    = beta^2*r_cur/lambda^2;
                alpha    = log(1 + ratio)/r_cur;
                if beta <= lambda || alpha <= 1
                    break;
                end
                % Back-track the Lipschitz constant by the factor BETA.
                Lips_cur    = Lips_cur/param.btkSlop;
            end

            % Compute the step-size.
            gamma   = 1/Lips_cur;

            % Compute the proximal gradient step.
            X_trial = scopt_soft_threshold_oper(X_cur - gamma*DFX_cur, ...
                                                gamma*rho);

            % Count the prox-operations.
            output.prox_cnt = output.prox_cnt + 1;

            % Compute the proximal-gradient search direction DX.
            DX_cur  = X_trial - X_cur;

            % Recompute three quantities: r, lambda and beta. 
            [nrm_DX, r_cur, beta, lambda] = compute_step_quantities(...
                                            DX_cur, What, Phat, TN, Mf, ...
                                            Lips_cur);
        end
    end
    
    % Compute the step size.
    ratio   = beta^2*r_cur/lambda^2;
    alpha   = log(1 + ratio)/max(r_cur, eps);
    
    % Make sure that alphak <= 1.
    alpha   = min(alpha, 1.0);
    
    % Update the step Z.
    Z_next  = X_cur + alpha*DX_cur;
    
    % Compute the decrease of the objective function (theoretically).
    FX_dec  = beta^2*((1 + 1/ratio)*log(1 + ratio) - 1)/max(r_cur, eps);
    
    % Compute the relative error.
    err_cur = nrm_DX / max(1.0, sqrt(sum(sum(X_cur.*X_cur))));
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, FX_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, beta, ...
                              r_cur, FX_dec, Lips_cur, X_cur);

    % Print the iteration.
    scopt_prints(param, iter, FX_cur, err_cur, lambda, beta, alpha, ...
                 FX_dec, output.prox_cnt, Lips_cur);
    
    % Update new iterative points.
    if ~param.isAccelerating
        X_next     = Z_next;
        theta_next = theta;
    else
        % Updating the next iteration.
        theta_next = 0.5*(1 + sqrt(4*theta^2 + 1));
        factor     = (theta - 1)/theta_next;
        X_next     = Z_next + factor*(Z_next - Z_cur);
    end
        
    % Check stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, FX_list] = scopt_fx_terminate(FX_cur, FX_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop
            output.msg = 'The objective values does not improve significantly';
            % Keep the last iteration.
            Z_cur = Z_next;
            break;
        end
    end
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX
        output.msg = 'Convergence achieved!';
        % Keep the last iteration.
        Z_cur = Z_next;
        break;
    end
    
    % Save the iteration.
    X_old   = X_cur;
    DFX_old = DFX_cur;
    FX_old  = FX_cur;
    
    % Move to the next step.
    X_cur = X_next;
    Z_cur = Z_next;
    theta = theta_next;
    
    % Restart if necessary.
    if param.isAccelerating
        if strcmpi(param.isRestart, 'fixed')
            if mod(iter, param.nRestartIters) == 0
                theta = 1;
                X_cur = Z_cur;
            end
        elseif strcmpi(param.isRestart, 'fx')
            if iter > 1 && FX_cur > FX_old
                theta = 1;
                X_cur = Z_cur;
            end
        end
    end
end

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', char(13), ...
                  'Increase the number of iterations to improve ...', ...
                  'Or change the tolerance param.abs_tol.'];
end

% Finalization phase.
output.time     = toc(output.time);
output.iter     = iter;
output.fval     = FX_cur;
output.error    = err_cur;
output.lbd      = lambda;
output.r        = r_cur;
output.beta     = beta;
output.x0       = X0;
output.Mf       = Mf;

% The final solution.
final_sol.x     = Z_cur;
final_sol.f     = FX_cur;
final_sol.error = err_cur;

% Print the final outputs.
scopt_finaloutput(param, output, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [nrm_DX, r_cur, beta, lambda] = compute_step_quantities(...
%%%                                       DX_cur, What, Phat, TN, Mf, Lips)
%%% PURPOSE:  Compute norm(DX), r, beta and lambda.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [nrm_DX, r_cur, beta, lambda] = compute_step_quantities(...
                                         DX, What, Phat, TN, Mf, Lips)

% Compute three quantities: r, lambda and beta. 
nrm_DX  = sqrt( sum(sum(DX.*DX)) );
r_cur   = Mf*nrm_DX;
beta    = sqrt(Lips)*nrm_DX;

% Compute lambda.
QDX    = DX'*What;
QPhat  = sum(QDX.*Phat, 1);
nrmHd2 = sum(sum(Phat.*QDX.^2)) - sum( QPhat.^2);
lambda = sqrt(nrmHd2/TN);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%