%
%                     --------------------------------- 
%                     LOGISTIC REGRESSION v.1.0. (2013)
%                     --------------------------------- 
%
%    An implementation of the Proximal Gradient/Newton algorithms for 
%        solving Sparse (Multinomial) Logistic Regression Problems.
%
% INFORMATION:
%    By Quoc Tran Dinh,
%       Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher.
%    Date: 16.04.2013.
%    Last modified: 16.04.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1]. Quoc Tran Dinh, Volkan Cevher and Yen-Huan Li: Minimization of 
%         self-concordant-like functions, Tech. Report., LIONS, EPFL, 2013.
%    [2]. Krishnapuram, B., Figueiredo, M., Carin, L., & Hartemink, A.
%         (2005) ?Sparse Multinomial Logistic Regression: Fast Algorithms 
%         and Generalization Bounds.? IEEE Transactions on Pattern Analysis 
%         and Machine Intelligence (PAMI), 27, June 2005. pp. 957?968.
%
% FUNCTION: [fsol, output] = scopt_mnlogistic_pna(What, Yhat, rho, ...
%                                                 X0, param, varargin)
%
% PURPOSE: 
%    An implementation of the proximal-(quasi-)Newton algorithm for solving 
%    the logistic regression problem of the form:
%
%                min{ (1/TN)*L(x, z) + tau/(sqrt(TN))*|x|_1 }
%
%    where 
%       L(x, z) = (1/N)*sum_{i=1}^Nlog(1 + exp(-(ai'*x + bi))).
%    Here, ai = yi*wi and bi = yi*bias.
%          
% INPUTS:
%     What     : Training matrix (p x N - matrix).
%     Yhat     : Classification matrix (m x N - matrix).
%     rho      : The penalty parameter in the regularization term.
%     X0       : An initial point (p x m - matrix) for vector X.
%     param    : optional structure to control the algorithm,
%                see scopt_optimset() for more details.
%     varargin : user defined inputs (varargin{1} is a scaling factor
%                TN, default value is TN = N).  
%     
% OUTPUTS:
%     fsol     : a structure containing the final solution.
%     output   : a structure containing the output information.
%
% NOTE: See scopt_mnlogistic_exam.m for the use of this solver.
%
%
function [fsol, output] = scopt_mnlogistic_pna( What, Yhat, rho, X0, ...
                                                param, varargin)

% Check the inputs.
if nargin < 5,     param = scopt_optimset([]);                end
if isempty(param), param = scopt_optimset([]);                end
if nargin < 4,     X0    = zeros(size(What,1), size(Yhat,1)); end
if isempty(X0),    X0    = zeros(size(What,1), size(Yhat,1)); end
if rho < 0 || isnan(rho) || isinf(rho)
    error('The penalty parameter should be positive'); 
end


% The size of the problem.
p  = size(What, 1); % Number of variables.
N  = size(What, 2); % Number of training samples/examples.
mx = size(Yhat, 1);

% Define the scaling factor TN (default value is N).
if nargin < 6
    TN = N; 
else
    if isempty(varargin{1}) 
        TN = N;
    else
        TN = varargin{1};
    end
end

% Initialize the output.
FX_list         = [];
param           = scopt_optimset(param);
param.Algorithm = 'PN';
output.nsize    = [p, N];
output.msg      = 'start';
output.info     = scopt_trace(param, []);
output.time     = tic;
output.prox_cnt = 0;

% Print the header message.
if param.PrintLevel == 1, nsp = 65; else nsp = 91; end
scopt_copyright(param, nsp, 'mnlogistic');

% Compute the parameter Mf.
Mf      = sqrt(6)*full(max(sqrt(sum(What.*What, 1))));

% Intialization step.
mWhat   = repmat(What, mx, 1);
X_old   = []; 
DFX_old = [];
H_old   = [];
lBFGS   = [];
X_cur   = X0(:);

% The main loop.
for iter = 1:param.MaxIters
    
    % Solve the primal subproblem and compute the proximal-Newton
    % decrement.
    [DX_cur, lambda, H_cur, DFX_cur, FX_cur, lBFGS, output_cur] = ...
          scopt_mnlogistic_ntdir(X_cur, What, Yhat, mWhat, TN, rho, ...
                                 DFX_old, X_old, H_old, lBFGS, param);
    
    % Count the number of prox-operators.
    if isfield(output_cur, 'prox_cnt')
        output.prox_cnt = output.prox_cnt + output_cur.prox_cnt;
    end
    
	% Compute the norm of DX and r_cur.
    nrm_DX = sqrt( sum(sum(DX_cur.*DX_cur)) );
    r_cur  = Mf*nrm_DX;
    
    % Compute the step size.
    alpha  = log(1 + r_cur)/max(r_cur, eps);
    
    % Update the next step.
    X_next  = X_cur + alpha*DX_cur;
    
    % Compute the decrease of the objective function.
    dec_FX = lambda^2*((1 + 1/r_cur)*log(1 + r_cur) - 1)/max(r_cur, eps);
    
    % Compute the relative error.
    err_cur = nrm_DX / max(1.0, sqrt(sum(sum(X_cur.*X_cur))));
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, FX_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, 0, 0, ...
                              dec_FX, 0, output_cur.iter, output_cur.time, ...
                              output_cur.error, output_cur.Lips, X_cur);
    % Print the iteration.
    scopt_prints(param, iter, FX_cur, err_cur, alpha, dec_FX, ...
                 output.prox_cnt, output_cur.error, output_cur.iter, ...
                 output_cur.time);
   
    % Check stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, FX_list] = scopt_fx_terminate(FX_cur, FX_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop
            output.msg = 'The objective values does not improve significantly';
            % Keep the last iteration.
            X_cur = X_next;
            break;
        end
    end
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX 
        output.msg  = 'Convergence achieved!';
        % Keep the last iteration before leaving the loop.
        X_cur = X_next;
        break;
    end
    
    % Go to the next iteration.
    X_old   = X_cur;
    X_cur   = X_next;
    H_old   = H_cur;
    DFX_old = DFX_cur;
end

% Exceed the maximum number of iterations.
if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', char(13), ...
                  'Increase the number of iterations to improve ...', ...
                  'Or change the tolerance param.abs_tol.'];
end

% Finalization phase.
output.time = toc(output.time);
output.iter = iter;
fsol.x      = reshape(X_cur, p, mx);
fsol.f      = FX_cur;
fsol.error  = lambda;

% Print the final outputs.
outf = output; outf.fval = FX_cur; 
scopt_finaloutput(param, outf, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%