%
%                ----------------------------------------- 
%                POISSON IMAGE RECONSTRUCTION v.1.0. (2013)
%                ----------------------------------------- 
%
%    An implementation of the Proximal Gradient/Newton 
%    algorithms for solving Poisson image reconstruction problems.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 22.05.2013.
%    Last modified: 01.07.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1] Q. Tran Dinh, A. Kyrillidis and V. Cevher: Self-Concordant 
%        Minimization, Tech. Report, LIONS, EPFL, 2013.
%    [2] Z.T. Harmany, R.F. Marcia, and R. M. Willett: This is spiral-tap: 
%        Sparse poisson intensity reconstruction algorithms theory and 
%        practice. IEEE Transactions on Image Processing, 1-13, 2012.
%
% FUNCTION:  [final_sol, output] = scopt_poisson_main( prob, param, ...
%                                  algchoice, varargin)
%
% PURPOSE: An implementation of the proximal-gradient algorithm for solving 
%          the Poisson denoising image problem of the form:
%
%                min{ F(x) + tau*TV(x) : x >= 0}
%
%          where 
%                F(x) = 1'*A(x) - sum_i=1^m y_i*log(e_i'*A(x)) 
%          and
%                TV is a given TV-norm (monotonic or nonmonotonic).
% 
% INPUTS:
%     prob      : The structure that contains problem definition.
%                 This structure contains the following fields:
%                   - prob.tau    : the regularization parameter 
%                   - prob.nsize  : the size of image
%                   - prob.Aopr   : the linear operator
%                   - prob.ATopr  : the adjoint operator of A.
%                   - prob.Yf     : the original image
%                   - prob.logeps : ground error, default = 0.
%     param     : optional structure to control the algorithm
%                 (see scopt_optimset() for more details).
%     algchoice : the choice of the algorithms (PG or FBPG).
%     varargin  : user defined inputs.
%     
% OUTPUTS:
%     final_sol : a structure containing the final solution.
%     output    : a structure containing the output information.
%

function [final_sol, output] = scopt_poisson_main(prob, param, ...
                               algchoice, varargin)

% Check the inputs.
if nargin < 1
    error('At least one input is required!'); 
end
if nargin < 2
    param = scopt_optimset([]);    
end
if isempty(param)
    param = scopt_optimset([]);
end
if nargin < 3
    algchoice = 'PG'; % Perform the proximal-gradient method.
end
if isempty(algchoice)
    algchoice = 'PG'; 
end

% Process the problem information.
if ~isstruct(prob) 
    error( 'The first argument must be a structure!');
end
 
% The size of the problem.
if ~isreal(prob.nsize) || length(prob.nsize) ~= 2
    error('The problem size is not correctly define!');
end
m = prob.nsize(1);
n = prob.nsize(2);

% If prob is a structure, then process it.
if ~isfield(prob, 'Aopr') || ~isfield(prob, 'Yf') || ...
   ~isfield(prob, 'tau')  || ~isfield(prob, 'nsize' )
    error('The problem is not correctly defined!');
end

% Check the operator A and its conjugate operator.
if isa(prob.Aopr, 'function_handle')
    Aopr = prob.Aopr;
    if ~isfield(prob, 'ATopr')
        error('The conjugate operator AT of A is not defined!');
    end
    if isa(prob.ATopr, 'function_handle')
        ATopr = prob.ATopr;
    elseif isnumeric( prob.ATopr )
        if isreal(prob.ATopr)
            ATopr = @(X)( prob.ATopr'*X );
        else
            ATopr = @(X)( transpose(prob.ATopr)*X );
        end
    else
        error('The conjugate operator AT of A is not correctly defined!');
    end
elseif isnumeric(prob.Aopr)
    Aopr = @(X)( prob.Aopr*X );
    if isnumeric(prob.ATopr)
        ATopr = @(X)( prob.ATopr*X );
    else
        if isreal(prob.Aopr)
            ATopr = @(X)( prob.Aopr'*X );
        else
            ATopr = @(X)( transpose(prob.Aopr)*X );
        end
    end
else 
    error( 'The linear operator A is not correctly defined!');
end

% Check the value of tau.
if ~isreal(prob.tau)
    error('The regularization parameter must be real!');
end
if prob.tau <= 0 || isnan(prob.tau) || isinf(prob.tau)
    error('The regularization parameter must be positive!');
end
tau = prob.tau;

if ~isreal(prob.tauf)
    error('The regularization parameter must be real!');
end
if prob.tauf <= 0 || isnan(prob.tauf) || isinf(prob.tauf)
    error('The regularization parameter must be positive!');
end

% Check the values of Yf.
if ~isreal( prob.Yf )
    error('The coefficient Yf must be real numbers!');
end
if any(any(prob.Yf < 0))
    error('The coefficient Yf must be positive!');
end
Yf = double(prob.Yf);
if size(Yf, 1) ~= m || size(Yf, 2) ~= n
    error( 'The size of the coefficient Yf is not consistent!' );
end

% Check the value of logeps.
if ~isfield(prob, 'logeps')
    logeps = 0; 
elseif ~isreal(prob.logeps)
    logeps = 0;
else
    logeps = prob.logeps;
end
if logeps < 0
    warning('The ground value "logeps" must be nonnegative!');
    logeps = 0;
end

% Check the initial point.
if isfield(prob, 'X0')
    if isreal(prob.X0) && size(prob.X0, 1) == m && size(prob.X0, 2) == n
        X0 = prob.X0;
    else
        disp('The initial point is not well defined!');
        X0 = [];
    end
else
    X0 = [];
end

% Check optional parameters.
param = scopt_optimset(param);

% Call the normal proximal gradient/function based proximal gradient algorithm.
if strcmpi(algchoice, 'FBPG')
    [final_sol, output] = scopt_poisson_fb_pga( m, n, Aopr, ATopr, Yf, ...
                          tau, logeps, X0, param, varargin{:});
else
     [final_sol, output] = scopt_poisson_pga( m, n, Aopr, ATopr, Yf, ...
                          tau, logeps, X0, param, varargin{:});
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
