%
%                ----------------------------------------- 
%                POISSON IMAGE RECONSTRUCTION v.1.0. (2013)
%                ----------------------------------------- 
%
%    An implementation of the Proximal Gradient/Newton 
%    algorithms for solving Poisson image reconstruction problems.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 22.05.2013.
%    Last modified: 01.07.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% REFERENCES:
%    [1] Q. Tran Dinh, A. Kyrillidis and V. Cevher: Self-Concordant 
%        Minimization, Tech. Report, LIONS, EPFL, 2013.
%    [2] Z.T. Harmany, R.F. Marcia, and R. M. Willett: This is spiral-tap: 
%        Sparse poisson intensity reconstruction algorithms theory and 
%        practice. IEEE Transactions on Image Processing, 1-13, 2012.
%
% FUNCTION: [Xsol, output] = scopt_poisson_pga( m, n, Aopr, ATopr, Yf, ...
%                            tau, beta, X0, param, varargin)
%
% PURPOSE: An implementation of the proximal-gradient algorithm for solving 
%          the Poisson denoising image problem of the form:
%
%                min{ F(x) + tau*TV(x) : x >= 0}
%
%          where 
%                F(x) = 1'*A(x) - sum_i=1^m y_i*log(e_i'*A(x)) 
%          and
%                TV is a given TV-norm (monotonic or nonmonotonic).
% 
% INPUTS:
%     [m, n]    : The size of the noise image.
%     Aopr      : The linear operator (function handle).
%     ATopr     : The adjoint operator of A.
%     Yf        : The coefficient vector of the function F.
%     tau       : The penalty parameter in the regularization term.
%     logeps    : The ground value, default = 0
%     X0        : An initial guess for X.
%     param     : optional structure to control the algorithm 
%                (see scopt_optimset() for more details).
%     varargin  : user defined inputs.
%     
% OUTPUTS:
%     final_sol : a structure containing the final solution.
%     output    : a structure containing the output information.
%
%
function [final_sol, output] = scopt_poisson_pga( m, n, Aopr, ATopr, ...
                               Yf, tau, logeps, X0, param, varargin)

% Check the inputs.
if nargin < 9,     param = scopt_optimset([]);       end
if isempty(param), param = scopt_optimset([]);       end
if nargin < 8,     X0    = generate_init(ATopr, Yf); end
if isempty(X0),    X0    = generate_init(ATopr, Yf); end
if tau < 0 || isinf(tau) || isnan(tau)
    error('The penalty parameter should be positive');  
end

% Initialize the output parameter.
output.nsize = [m, n];
output.msg   = 'start';

% Check the algorithmic parameters.
param           = scopt_optimset(param);
param.Algorithm = 'PG';

if param.PrintLevel == 1, nsp = 42; else nsp = 97; end
scopt_copyright(param, nsp, 'Poisson');

% Initialize the output parameter.
output.nsize    = [m, n];
output.msg      = 'start';
output.prox_cnt = 0;
output.info     = scopt_trace(param, []);
output.time     = tic;

% Compute the parameter Mf.
Yfpos = Yf(Yf > 0);
Mf    = ( max( 1./sqrt(Yfpos(:)) ) )^2;

% Intialization step.
fx_list   = [];
fx_cur    = inf;
X_cur     = X0;
W0_cur{1} = zeros(m-1, n);
W0_cur{2} = zeros(m, n-1);
X0_cur    = X0;
Lips_cur  = 4;

% The main loop.
for iter = 1:param.MaxIters
    
    % Evaluate the vector AX = A*X;
    AX_cur = Aopr(X_cur);
    
    % Evaluate the objective value at X_cur, if required.
    if param.isFxEval
        fx_cur  = scopt_poisson_fx_eval(X_cur, AX_cur, Yf, tau, logeps);
    end
    
    % Evaluate the gradient of Fhat.
    grad_fx = grad_fx_eval(AX_cur, ATopr, Yf, logeps, Mf);
    
    % Compute the Lipschitz constant.
    if mod(iter, 1000) == 1 
        Lips_cur = max(1e-12, 1.1*scopt_poisson_l2norm(X_cur, AX_cur, ...
                   Yf, logeps, Aopr, ATopr, Mf, param.PwMaxIters));
        safe_guard_Lip  = Lips_cur;
    end
    
    % Compute the trial step.
    [DX_cur, lambda, nrm_dx, beta, X0_next, W0_next, out] ...
            = compute_trial_step( X_cur, grad_fx, Lips_cur, tau, Mf, ...
              AX_cur, Aopr, logeps, Yf, X0_cur, W0_cur, param, iter);
	
	% Count the number of prox-operations.
    output.prox_cnt = output.prox_cnt + out.prox_cnt;

    % Back-tracking line-search if required.
    if param.isBtkLS
        for btkit = 1:param.BtkMaxIters
            % If the back-tracking condition meets, then break.
            if Lips_cur <= lambda^2/nrm_dx^2;
                break; 
            end
            
            % Otherwise, half the Lipschitz constant.
            Lips_cur = Lips_cur/param.btkSlop;
            
            % And recompute the trial step.
            [DX_cur, lambda, nrm_dx, beta, X0_next, W0_next, out1] ...
                = compute_trial_step( X_cur, grad_fx, Lips_cur, tau, Mf, ...
                  AX_cur, Aopr, logeps, Yf, X0_cur, W0_cur, param, iter);
          
            % Count the number of prox-operations.
            output.prox_cnt = output.prox_cnt + out.prox_cnt;
            out.time        = out.time + out1.time;
            out.iter        = out.iter + out1.iter;
        end
    end
    
    % Compute the step size.
    alpha   = beta/(lambda*(lambda + beta));
    
    % Make sure that alpha <= 1.
    alpha   = min(alpha, 1.0);
    
    % Update the next step.
    X_next  = X_cur + alpha*DX_cur;
    
    % Compute the decrease of the objective function.
    ratio   = beta/lambda;
    fx_dec  = abs(ratio - log(1 + ratio));
    
    % Compute the relative error.
    err_cur = nrm_dx / max(1.0, sqrt(sum(sum(X_cur.*X_cur))));
    
    % Checking stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, fx_list] = scopt_fx_terminate(fx_cur, fx_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop && err_cur <= param.acptRelTolX
            output.msg = 'The objective values does not improve significantly';
            % Keep the last iteration.
            X_cur = X_next;
            break;
        end
    end
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX
        output.msg = 'Convergence achieved! Error is below the tolerance!';
        % Keep the last iteration.
        X_cur = X_next;
        break;
    end
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, fx_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, beta, 0, ...
                              fx_dec, 0, out.iter, out.time, 0.0, ...
                              Lips_cur);

    % Print the iteration.
    scopt_prints(param, iter, fx_cur, err_cur, lambda, beta, alpha, ...
                 fx_dec, output.prox_cnt, Lips_cur);
    
    % Jump to the next iteration.
    X_cur  = X_next;
    X0_cur = X0_next;
    W0_cur = W0_next;
    
    % Update the Lipschitz constant whenever lambda < 1.
    if ~param.isBtkLS
        if lambda < 1.0
            Lips_cur = 0.1*lambda^2/nrm_dx^2;
        else
            Lips_cur = safe_guard_Lip;
        end
    end
end

if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', char(13), ...
                  'Increase the number of iterations to improve ...', ...
                  char(13), 'Or change the tolerance param.abs_tol.'];
end

% Get the output information.
output.iter  = iter;
output.time  = toc(output.time);
output.fval  = fx_cur;
output.error = err_cur;
output.X0    = X0;
output.Mf    = Mf;

final_sol.x     = X_cur;
final_sol.f     = fx_cur;
final_sol.error = err_cur;

scopt_finaloutput(param, output, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DX, lambda, nrm_DX, beta, X0, W0, out] = compute_trial_step( 
%%%                                                   X_cur, grad_fx, L, ...
%%%                                                   tau, Mf, AX_cur, ...
%%%                                                   Aopr, logeps, Yf, ...
%%%                                                   X0_cur, W0_cur, ...
%%%                                                   param, oiter)
%%% PURPOSE:  Compute the trial step.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX, lambda, nrm_DX, beta, X0, W0, out] = compute_trial_step(...
                                                   X_cur, grad_fx, L, ...
                                                   tau, Mf, AX_cur, ...
                                                   Aopr, logeps, Yf, ...
                                                   X0_cur, W0_cur, ...
                                                   param, oiter)
% Compute the first-step size.
gamma   = 1/L;

% Compute S = X - gamma*grad_fx.
S_cur   = X_cur - gamma*grad_fx;

% Update the next iteration.
[X_trial, X0, W0, out] = subprob_solver(S_cur, tau, gamma, Mf, X0_cur, ...
                                        W0_cur, param, oiter);

% Compute the search direction DX.
DX      = X_trial - X_cur;

% Compute beta. 
nrm_DX  = sqrt( sum(sum(DX.*DX) ) );
beta    = L*nrm_DX^2;

% Compute the Newton decrement.
Ad      = Aopr(DX);
dHd     = Mf*Yf.*(Ad./(AX_cur + logeps)).^2;
lambda  = sqrt( sum(sum(dHd)) );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: X0 = generate_init(ATopr, Yf)
%%% PURPOSE:  Generate an initial point for the algorithm.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function X0 = generate_init(ATopr, Yf)

ATy  = ATopr(Yf);
Xtmp = sum(sum((ATopr(ones(size(Yf))))));
X0   = (sum(sum(Yf)).*numel(ATy))./(sum(sum(ATy)).* Xtmp).*ATy;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: gradf = grad_fx_eval(Axk, ATopr, yf, logeps, Mf)
%%% PURPOSE:  Evaluate the gradient of F:
%%%           F'(x) = Mf*(A'*1 - sum_i=1^m(y./(A*x + logeps))*A').
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function gradf = grad_fx_eval(Ax, ATopr, yf, beta, Mf)

gradf = Mf*ATopr(1.0 - (yf./(Ax + beta)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [Xout, X0out, W0out, output] = subprob_solver(sk, tau, gamma, 
%%%                                          Mf, X0, W0, param, oiter)
%%% PURPOSE:  Solve the subproblem with the denoising solver.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Xout, X0out, W0out, output] = subprob_solver(s_cur, tau, ...
                                        gamma, Mf, X0, W0, param, oiter)

% The regularization parameter.
reg_param = Mf*tau*gamma;

% Call the solver.
[Xout, X0out, W0out, output] = scopt_poisson_dnFISTA(s_cur, reg_param, ...
                               0.0, X0, W0, param, oiter); 
             
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%