% FUNCTION: [fsol, output] = scopt_scheds_fb_pga(Xhat, yhat, rho, x0, ...
%                            sig0, param, varargin)
%
% PURPOSE: An implementation of the function-based proximal-gradient 
%          algorithm for solving the following LASSO problem with unknown 
%          variance:
%
%              min{ -sum_i=1^nlog(sigma_i) + 1/(2*n)*|sigma.*y-X*phi|^2
%                   + rho*|phi|_1 : sigma > 0, phi}
%          
% REFERENCES: 
%    [1]. Q. Tran Dinh, A. Kyrillidis and V. Cevher: A proximal Newton 
%    framework for composite minimization: Graph learning without 
%    Cholesky decompositions and matrix inversions. ICML 2013, Alanta,
%    USA.
%    [2]. N. Stadler, P. Bulmann, and S. Van de Geer: l1-penalization for 
%    mixture regression models. Tech. Report., pages 1?35, 2012.
%
% INPUTS: 
%     Xhat    : an m x n - input matrix.
%     yhat    : an m x 1 - observation vector.
%     rho     : the penalty parameter (regularization parameter)
%               it is a positive number for tuning the sparsity.
%     x0      : An initial guess for vector x (or Phi).
%     sig0    : An initial guess for sigma.
%     param   : optional parameter, see scopt_optimset() for more details.
%     varargin: user-defined parameters.
%
% OUTPUTS:
%     fsol    : A structure containing the final solution.
%     output  : A structure containing the output information.
%
% INFORMATION:
%     Created date: 28.03.2013.
%     Implemented by Quoc Tran Dinh,
%            LIONS/STL/IEL, EPFL, Lausanne, Switzerland.
%     Joint work with Anastasios Kyrillidis and Volkan Cevher.
%     Last modified: 01.07.2013.
%
% NOTE: For more information about the problem formulation, we refer to the
%       reference [2]. 
%
%
function [fsol, output] = scopt_scheds_fb_pga(Xhat, yhat, rho, x0, ...
                                              sig0, param, varargin)

% Check the inputs.
if nargin < 6,     param = scopt_optimset([]); end
if isempty(param), param = scopt_optimset([]); end

mx = size(Xhat, 1);
nx = size(Xhat, 2);

% Check the initial point.
if nargin < 4
    [x0, sig0] = generate_init_point(Xhat, yhat); 
end
if isempty(x0) || isempty(sig0)
    [x0, sig0] = generate_init_point(Xhat, yhat);
end

% Weighted matrix.
Wi = [];
if nargin > 6, Wi = varargin{1}; end
if ~isempty(Wi)
    if size(Wi, 2) ~= nx
        error('The size of the weighted matrix is inconsistent!');
    end
    Wi = sqrt(sum(Wi.*Wi));
    Wi = Wi(:);
end

% Check other input variables.
if rho < 0 || isnan(rho) || isinf(rho)
    error('The penalty parameter should be positive!'); 
end

% Check the algorithmic parameters.
param = scopt_optimset(param);
param.Algorithm = 'PG';

if param.PrintLevel == 1, nsp = 42; else nsp = 97; end
scopt_copyright(param, nsp, 'glearn');

% Initialize the output parameter.
output.nsize    = [mx, nx];
output.msg      = 'start';
output.prox_cnt = 0;
output.info     = scopt_trace(param, []);
output.time     = tic;

% Intialize the parameter.
fx_list     = [];
X_cur       = x0(:);
S_cur       = sig0;
W_old       = [];
gradFw_old  = [];

% Evaluate the Lipschitz constant.
Lips_cur    = l2_norm_eval(sig0, Xhat, yhat, mx, param.PwMaxIters);

% Evaluate the objective value, and its gradient vector.
XhatX                      = Xhat*X_cur;
[DFX_cur, DFS_cur, FW_cur] = fx_eval(X_cur, S_cur, XhatX, Xhat, yhat, ...
                                     mx, rho, false);
% Concatenate the vectors.
if param.BBStepType ~= 0
    W_cur                      = [S_cur; X_cur];
    gradFw_cur                 = [DFS_cur; DFX_cur];
end

% The main loop.
for iter = 1:param.MaxIters
    
    % Predict the constant Lips_cur by the Barzilai-Borwein rule.
    if param.BBStepType ~= 0
        Lips_cur = scopt_compute_BBstep(W_old, W_cur, gradFw_old, ...
                   gradFw_cur, param, Lips_cur);
    end
    
    % Compute the first trial step.
    [DX_cur, DS_cur, X_trial, S_trial, XD, lambda, nrm_Dw, beta] = ...
                                             compute_trial_step(X_cur, ...
                                             S_cur, DFX_cur, DFS_cur, ...
                                             rho, Lips_cur, Xhat, yhat, ...
                                             mx, Wi);
    % Count the number of prox-operations.
    output.prox_cnt = output.prox_cnt + 1;
    
    % Back-tracking if required.
    if param.isBtkLS
        for btkit = 1:param.BtkMaxIters
            
            % Check the condition
            if Lips_cur <= lambda^2/nrm_Dw^2;
                break; 
            end
            
            % Decrease the Lipschitz constant.
            Lips_cur = Lips_cur/param.btkSlop;
            
            % Compute the second trial step.
            [DX_cur, DS_cur, X_trial, S_trial, XD, lambda, nrm_Dw, beta] = ...
                                                     compute_trial_step(...
                                                     X_cur, S_cur, ...
                                                     DFX_cur, DFS_cur, ...
                                                     rho, Lips_cur, Xhat, ...
                                                     yhat, mx, Wi);
            % Count the number of prox-operations.
            output.prox_cnt = output.prox_cnt + 1;
        end
    end
    
    % Compute the step size.
    alpha   = beta/(lambda*(lambda + beta));
    
    % Make sure that alpha <= 1.
    alpha   = min(alpha, 1.0);
    
    % Update the next step.
    X_next  = X_cur + alpha*DX_cur;
    S_next  = S_cur + alpha*DS_cur;
    
    % Evaluate the objective value at X_trial and S_trial
    FW_trial    = inf;
    XhatX_trial = XhatX + XD;
    if S_cur >= 1e-8
        FW_trial = fx_eval(X_trial, S_trial, XhatX_trial, Xhat, ...
                   yhat, mx, rho, true);
    end
    
    % Evaluate the objective value at X_next and S_next.
    XhatX_next = XhatX + alpha*XD;
    FW_next    = fx_eval(X_next, S_next, XhatX_next, Xhat, yhat, ...
                 mx, rho, true);

    % Check the objective value of X_trial and X_next
    if FW_trial < FW_next
        FW_next     = FW_trial;
        X_next      = X_trial;
        S_next      = S_trial;
        alpha       = 1;
        XhatX_next  = XhatX_trial;
    end
    
    % Compute the decrease of the objective function.
    ratio   = beta/lambda;
    dec_fx  = ratio - log(1 + ratio);
    
    % Compute the relative error.
    err_cur = nrm_Dw/max(1.0, sqrt(S_cur*S_cur + sum(sum(X_cur.*X_cur))));
    
    % Check stopping criterion based on the norm of the search direction.
    if err_cur <= param.RelTolX
        output.msg = 'Convergence achieved!';
        % Keep the last iteration.
        X_cur   = X_next;
        S_cur   = S_next;
        FW_cur  = FW_next;
        break;
    end
    
    % Checking stopping criterion based on the objective values if required.
    if param.isFxEval && param.isFxStopCheck
        [fstop, fx_list] = scopt_fx_terminate(FW_cur, fx_list, ...
                           param.numFxCheck, param.RelTolFun);
        if fstop && err_cur <= param.acptRelTolX
            output.msg = 'The objective values does not improve significantly';
            % Keep the last iteration.
            X_cur   = X_next;
            S_cur   = S_next;
            FW_cur  = FW_next;
            break;
        end
    end
    
    % Record the iteration information.
    output.info = scopt_trace(param, output.info, FW_cur, err_cur, ...
                              output.prox_cnt, alpha, lambda, beta, 0, ...
                              dec_fx, Lips_cur);

    % Print the iteration.
    scopt_prints(param, iter, FW_cur, err_cur, lambda, beta, alpha, ...
                 dec_fx, output.prox_cnt, Lips_cur);
    
	% Save the previous iteration.
    if param.BBStepType ~= 0
        W_old       = W_cur;
        gradFw_old  = gradFw_cur;
    end
    
    % Jump to the next iteration.
    X_cur   = X_next;
    S_cur   = S_next;
    FW_cur  = FW_next;
    XhatX   = XhatX_next;
    
    % Evalute the gradient for the next iteration.
    [DFX_cur, DFS_cur] = fx_eval(X_cur, S_cur, XhatX, Xhat, yhat, ...
                         mx, rho, false);
                     
    % Form the gradient vector and the next iteration.
    if param.BBStepType ~= 0
        W_cur      = [S_next; X_next];
        gradFw_cur = [DFS_cur; DFX_cur];
    end
end

if iter >= param.MaxIters
    output.msg = ['Exceed the maxinum number of iterations.', ...
                  char(13), 'Increase the number of iterations to ', ...
                  'improve or change the tolerance param.abs_tol.'];
end

% Get the output information.
output.time  = toc(output.time);
output.iter  = iter;
output.fval  = FW_cur;
output.error = err_cur;
output.X0    = [sig0; x0(:)];

sig          = 1/S_cur;
fsol.sig     = sig;
fsol.x       = sig*X_cur;
fsol.f       = FW_cur;
fsol.error   = err_cur;

scopt_finaloutput(param, output, nsp);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DX_cur, DS_cur, X_trial, S_trial, XD, lambda, nrm_Dw, beta] =
%%%                                            compute_trial_step( ...
%%%                                            X_cur, S_cur, DFX_cur, ...
%%%                                            DFS_cur, rho, L, Xhat, ...
%%%                                            yhat, mx, Wi)
%%% PURPOSE:  Compute the quantities for step size generation.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DX_cur, DS_cur, X_trial, S_trial, XD, lambda, nrm_Dw, beta] = ...
                                                  compute_trial_step(...
                                                  X_cur, S_cur, DFX_cur, ...
                                                  DFS_cur, rho, L, Xhat, ...
                                                  yhat, mx, Wi)
% Compute the step size.
gamma   = 1./L;

% Update the next iteration by prox-operation.
[X_trial, S_trial] = perform_prox_oper(X_cur, S_cur, DFX_cur, DFS_cur, ...
                     gamma, rho, 0.0, inf, Wi);

% Compute the search direction DW_cur.
DX_cur  = X_trial - X_cur;
DS_cur  = S_trial - S_cur;

% Compute beta.
nrm_Dw  = sqrt(sum(DX_cur.*DX_cur) + DS_cur*DS_cur);
beta    = L*nrm_Dw^2;

% Compute the proximal Newton decrement.
[lambda, XD] = compute_PN_decrement(DX_cur, DS_cur, S_cur, Xhat, yhat, mx);
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [x0, sig0] = generate_init_point(Xhat, yhat)
%%% PURPOSE:  Generate an initial point for the algorithm.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [x0, sig0] = generate_init_point(Xhat, yhat)

sig0 = exp(1);
x0   = zeros(size(Xhat,2), 1); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [DFX, DFS, fx] = fx_eval(X, sig, Xhat, yhat, mx, rho, isFx)
%%% PURPOSE:  Evaluate the gradient vector of Fw.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [DFX, DFS, fx] = fx_eval(X, sig, XhatX, Xhat, yhat, mx, rho, isFx)

if nargin < 8, isFx = false; end

zhat = XhatX - sig*yhat;

if ~isFx
    DFX  = Xhat'*zhat/mx;
    DFS  = -1/sig - yhat'*zhat/mx;
    if nargout > 2
        fx = -log(sig) + 0.5*sum( zhat.*zhat )/mx + rho*sum(abs(X));
    end
else
    DFX = -log(sig) + 0.5*sum( zhat.*zhat )/mx + rho*sum(abs(X));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [lambda, z] = compute_PN_decrement(dx, ds, sig, Xhat, yhat, mx)
%%% PURPOSE:  Evaluate the Proximal-Newton decrement.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [lambda, z] = compute_PN_decrement(dx, ds, sig, Xhat, yhat, mx)

% Compute lambda
z       = Xhat*dx;
lambda  = ( 1./sig^2 + yhat'*yhat/mx )*ds^2 + z'*z/mx - 2*ds*yhat'*z/mx;
lambda  = sqrt(lambda);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: [Xout, Sout] = perform_prox_oper(X, S, DFX, DFS, gamma, ...
%%%                          rho, lbx, ubx, Wi)
%%% PURPOSE:  Compute the prox-operator of W
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Xout, Sout] = perform_prox_oper(X, S, DFX, DFS, gamma, rho, ...
                                          lbx, ubx, Wi)

% Form the necessary vector.
zX   = X - gamma*DFX;
zS   = S - gamma*DFS;

% Projection step.
if lbx >= 0 && lbx < inf,   zS = max(zS, lbx); end
if ubx >= lbx && ubx < inf, zS = min(zS, ubx); end

% Soft-thresholding step.
if isempty(Wi), WzX = zX; else WzX = Wi.*zX; end
Sout = zS;
Xout = sign(WzX).*max( abs(WzX) - rho*gamma, 0);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: l2norm = l2_norm_eval(sig, Xhat, yhat, mx, maxiter)
%%% PURPOSE:  Evaluate the Lipschitz constant by a simple power method.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function l2norm = l2_norm_eval(sig, Xhat, yhat, mx, maxiter)

if nargin < 2, maxiter = 20; end

dsig = 1; 
dx   = ones(size(Xhat,2), 1);
asig = 1/sig^2 + yhat'*yhat/mx;
for iter = 1:maxiter
    nrmdw = sqrt(dsig^2 + dx'*dx); 
    Xdx   = Xhat*dx;
    dx    = Xhat'*(Xdx - dsig*yhat)/mx;
    dsig  = asig*dsig - yhat'*Xdx/mx;
    dsig  = dsig/nrmdw;
    dx    = dx/nrmdw;
end

Ad     = Xhat*dx;
nrm2dw = dsig^2 + dx'*dx;
l2norm = ( asig*dsig^2 + Ad'*Ad/mx - 2*dsig*yhat'*Ad/mx )/nrm2dw;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%