% FUNCTION: options = scopt_optimset(varargin)
% PURPOSE:  Generate optional parameters for algorithms.
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       (LIONS), EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 03.06.2013.
%    Last modified: 03.06.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% CALL: 
%    help sccomopt_optimset: see optional parameters.   
%    options = sccomopt_optimset() to see the default values.
%    options = sccomopt_optimset('Name', Value, ...) to set new options.
%    options = sccomopt_optimset(options, 'Name', Value, ...)
%              to add new options.
%
% NOTES:
%    MaxIters      : The maximum number of iterations.
%    MaxItersP1    : The maximum number of iterations in Phase I.
%    MaxItersP2    : The maximum number of iterations in Phase II.
%    SubMaxIters   : The maximum number of iterations in the inner loop.
%    ubSubIters    : The upper bound of iterations in the subsolvers.
%    Algorithm     : The algorithm (PG, FPG, PN and PFPN). 
%    PwMaxIters    : The maximum number of iterations in the power method
%                    for estimating the Lipschitz constant.
%    AbsTol        : The absolute tolerance.
%    TolParam      : The tolerance for the penalty parameter in PFPN.
%    RelTolX       : The relative tolerance for the search direction.
%    RelTolFun     : The relative tolerance for the objective value.
%    acptRelTolX   : Accepted tolerance for the residual.
%    TolStep       : The lower bound of the step size in the subsolver.
%    TerminateType : Type of termination conditions.', char(13), ...
%                    (0 = default, 1 = norm of the search
%                    direction, 2 - objective value change, 3 = 1 + 2).
%    SubRelTol     : The relative tolerance in the inner loops.
%    SubTolOpt     : The tolerance for the optimality condition (inner
%                    loop).
%    DescentSlop   : The descent slop in line-search procedures.
%    isDebug       : Debug mode.
%    isTrace       : Keep trace the information w.r.t. iterations.
%    isFxEval      : Evaluating the objective value if required.
%    isFxStopCheck : Check the stopping criterion based on the f(x) values.
%    numFxCheck    : Number of checking iterations in the objective value.
%    PrintLevel    : Output printing level.
%    isWarning     : Warning message.
%    proxOperType  : Type of prox-operations (l1, iso ...).
%    SubMethod     : The method for solving the subproblem (FISTA or
%                    SPARSA).
%    UpdateH       : The method for updating the Hessian matrix
%                    (BFGS, LBFGS or Newton).
%    LBfgsLength   : The length of the limitted memory BFGS list.
%    TraceMode     : Trace mode (0 = basic and 1 = details, 
%                    3 = include the iterative vector).
%    isGradNorm    : Compute the norm of the gradient vector if required.
%    isRestart     : Perform the restart for the outer loop.
%    nRestartIters : The number of iterations for repeating restart.
%    isSubRestart  : Perform restart technique in fast gradient algorithm.
%    nSubRsIters   : Number of iterations for repeating restart trick.
%    isSubBtkLS    : Do backtracking line-search to estimate a value for
%                    Lipschitz constant in the fast gradient method.
%    nSubFxCheck   : Number of iterations to check f(x) in subproblem.
%    isBtkLS       : Applying back-tracking line-search if required.
%    lsMode        : Line-search mode: BTK or FWLK.
%    minStepSize   : The lower bound of the step size.
%    BtkMaxIters   : The maximum number of iterations in back-tracking 
%                    line-search for the inner loop.
%    BtkMemIters   : The number of memorized iterations in back-tracking 
%                    line-search for the inner loop.
%    btkSlop       : Backtracking slop.
%    BBStepType    : Types of the Barzilai-Borwein step (1, 2 or 3).
%    isEvalStrCvx  : Evaluate the strong convexity parameter if required.
%    QuadCvRg      : The radius of quadratic convergence region.
%    bcdsSubTol    : The tolerance for the block-coordinate descent method.
%    

function options = scopt_optimset(varargin)

% Define the default options.
defaultopt = struct( ...
                 'MaxIters',        5000,     ...
                 'MaxItersP1',      1000,       ... 
                 'MaxItersP2',      1000,       ... 
                 'SubMaxIters',     100,        ...
                 'PwMaxIters',      20,         ...
                 'TerminateType',   0,          ...
                 'Algorithm',       'PG',       ...
                 'TolParam',        1e-3,       ...
                 'AbsTol',          1e-8,       ...
                 'RelTolX',         1e-5,       ...
                 'RelTolFun',       1e-4,       ...
                 'acptRelTolX',     1e-2,       ...
                 'TolStep',         1e-6,       ...
                 'SubRelTol',       1e-5,       ...
                 'SubTolOpt',       1e-5,       ...
                 'DescentSlop',     0.0001,     ...
                 'isDebug',         false,      ...
                 'isFxStopCheck',   true,       ...
                 'isFxEval',        true,       ...
                 'numFxCheck',      5,          ...
                 'isWarning',       'off',      ...
                 'PrintLevel',      1,          ...
                 'SubMethod',       'sparsa',   ...
                 'UpdateH',         'bfgs',     ...
                 'LBfgsLength',     50,         ...
                 'proxOperType',    'l1',       ...
                 'isTrace',         true,       ...
                 'TraceMode',       0,          ...
                 'isGradNorm',      false,      ...
                 'isRestart',       'none',     ...
                 'nRestartIters',   50,         ...
                 'isSubRestart',    'none',     ...
                 'nSubRsIters',     100,        ...
                 'nSubFxCheck',     5,          ...
                 'BBStepType',      3,          ...
                 'isSparseCheck',   false,      ...
                 'BtkMaxIters',     20,         ...
                 'isBtkLS',         false,      ...
                 'BtkMemIters',     10,         ...
                 'isSubBtkLS',      false,      ...
                 'btkSlop',         2,          ...
                 'lsMode',          'BTK',      ...
                 'minStepSize',     1.0e-4,     ...
                 'bcdsSubTol',      0.05,       ...
                 'ubSubIters',      200,        ... % Internal parameter.
                 'isEvalStrCvx',    false,      ...
                 'QuadCvRg',        0.16441800, ... 
                 'Epsilon',         1e-12       ...
                 );
             
% In case no input or empty input.
if nargin < 1,        
    options = defaultopt; 
    printOptions();
    return; 
end
if isempty(varargin{1})
    options = defaultopt; 
    return; 
end

% In case varargin{1} is a structure.
if isstruct(varargin{1})
    options = varargin{1};
    if nargin > 1
        inputs = varargin(2:end);
    else
        inputs = [];
    end
else
   options = defaultopt; 
   inputs = varargin;
end
if mod(length(inputs), 2) ~= 0
    error('Optional parameters must be in pair (Name, Value)!');
end

% Now, add the new options.
options = addNewOptions(options, inputs);

% Check if inputs are correct.
options = checkfield(options, defaultopt, 'MaxIters' );
options = checkfield(options, defaultopt, 'MaxItersP1' );
options = checkfield(options, defaultopt, 'MaxItersP2' );
options = checkfield(options, defaultopt, 'SubMaxIters' );
options = checkfield(options, defaultopt, 'BtkMaxIters' );
options = checkfield(options, defaultopt, 'BtkMemIters' );
options = checkfield(options, defaultopt, 'PwMaxIters' );
options = checkfield(options, defaultopt, 'AbsTol' );
options = checkfield(options, defaultopt, 'TolParam' );
options = checkfield(options, defaultopt, 'RelTolX' );
options = checkfield(options, defaultopt, 'RelTolFun' );
options = checkfield(options, defaultopt, 'TolStep' );
options = checkfield(options, defaultopt, 'SubRelTol' );
options = checkfield(options, defaultopt, 'SubTolOpt' );
options = checkfield(options, defaultopt, 'DescentSlop' );
options = checkfield(options, defaultopt, 'isDebug' );
options = checkfield(options, defaultopt, 'isFxEval' );
options = checkfield(options, defaultopt, 'numFxCheck' );
options = checkfield(options, defaultopt, 'PrintLevel' );
options = checkfield(options, defaultopt, 'isWarning' );
options = checkfield(options, defaultopt, 'SubMethod' );
options = checkfield(options, defaultopt, 'UpdateH' );
options = checkfield(options, defaultopt, 'LBfgsLength' );
options = checkfield(options, defaultopt, 'TerminateType');
options = checkfield(options, defaultopt, 'Algorithm');
options = checkfield(options, defaultopt, 'isTrace');
options = checkfield(options, defaultopt, 'TraceMode');
options = checkfield(options, defaultopt, 'isGradNorm');
options = checkfield(options, defaultopt, 'isSubRestart');
options = checkfield(options, defaultopt, 'nSubRsIters');
options = checkfield(options, defaultopt, 'isSparseCheck');
options = checkfield(options, defaultopt, 'BtkMaxIters' );
options = checkfield(options, defaultopt, 'BtkMemIters' );
options = checkfield(options, defaultopt, 'isSubBtkLS' );
options = checkfield(options, defaultopt, 'btkSlop' );
options = checkfield(options, defaultopt, 'isEvalStrCvx' );
options = checkfield(options, defaultopt, 'QuadCvRg' );
options = checkfield(options, defaultopt, 'isFxStopCheck' );
options = checkfield(options, defaultopt, 'isBtkLS' );
options = checkfield(options, defaultopt, 'nRestartIters' );
options = checkfield(options, defaultopt, 'isRestart' );
options = checkfield(options, defaultopt, 'BBStepType' );
options = checkfield(options, defaultopt, 'acptRelTolX' );
options = checkfield(options, defaultopt, 'nSubFxCheck' );
options = checkfield(options, defaultopt, 'proxOperType' );
options = checkfield(options, defaultopt, 'bcdsSubTol' );
options = checkfield(options, defaultopt, 'lsMode' );
options = checkfield(options, defaultopt, 'minStepSize' );
options = checkfield(options, defaultopt, 'ubSubIters' );
options = checkfield(options, defaultopt, 'Epsilon' );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: outopt = checkfield(outopt, default, fieldname) 
%%% PURPOSE:  This function checks the field of optional parameters.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function outopt = checkfield(outopt, default, fieldname)
    
    fieldvalue = getfield( default, fieldname );
    if isfield(outopt, fieldname )
        optfieldval = getfield(outopt, fieldname );
        if isempty( optfieldval )
            outopt = setfield( outopt, fieldname, fieldvalue );
        end;
    else
        outopt = setfield( outopt, fieldname, fieldvalue );
    end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: options = addNewOptions(options, inputs)
%%% PURPOSE:  Add new options to the structure.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function options = addNewOptions(defin, inputs)

options = defin;
if isempty(inputs), return; end

for k=1:length(inputs)/2
    name  = inputs{2*k-1};
    value = inputs{2*k};
    if ~ischar(name)
        error('Name of an option must be a string!');
    end
    if ~isfield(options, name)
        warning('This optional parameter does not exist! Add a new one.');
    end
    options = setfield(options, name, value);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION: printOptions()
%%% PURPOSE:  Print the options.
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function printOptions()

fprintf('  MaxIters      : The maximum number of iterations.\n');
fprintf('  MaxItersP1    : The maximum number of iterations in Phase I.\n');
fprintf('  MaxItersP2    : The maximum number of iterations in Phase II.\n');
fprintf(['  Algorithm     : The algorithm (PG, FPG, PN and PFPN).\n']); 
fprintf(['  SubMaxIters   : The maximum number of iterations in the', ...
         ' inner loop.\n']);
fprintf('  ubSubIters    : The upper bound of iterations in the subsolvers.\n');
fprintf(['  PwMaxIters    : The maximum number of iterations in the ', ...
         'power method', char(13), ...
         '                  for estimating the Lipschitz constant.\n']);
fprintf(['  TolParam      : The tolerance of the penalty parameter', ...
         ' in the path-following method.\n']);
fprintf('  AbsTol        : The absolute tolerance.\n');
fprintf(['  RelTolX       : The relative tolerance for the search ', ...
         'direction.\n']);
fprintf('  acptRelTolX   : Accepted tolerance for the residual.\n');
fprintf(['  RelTolFun     : The relative tolerance for the objective ', ...
         'values.\n']);
fprintf( '  TolStep       : The lower bound of the step size in subsolvers.\n');
fprintf( '  SubRelTol     : The relative tolerance in the inner loop.\n');
fprintf(['  SubTolOpt     : The tolerance for the optimality ', ...
         'condition in the inner loop.\n']);
fprintf('  DescentSlop   : The descent slop in line-search procedures.\n');
fprintf('  proxOperType  : Type of prox-operations (l1, iso, ...).\n');
fprintf('  isDebug       : Debug mode.\n');
fprintf(['  isFxStopCheck : Check the stopping criterion based on', ...
         ' the objective values.\n']);
fprintf('  isFxEval      : Evaluating the objective value if required.\n');
fprintf(['  numFxCheck    : Number of checking iterations in the ', ...
         'objective value (using in the stopping criterion).\n']);
fprintf( '  PrintLevel    : Option for printing output.\n');
fprintf( '  isWarning     : Warning message.\n');
fprintf(['  SubMethod     : The method for solving the subproblems', ...
        char(13), '                  (FISTA, SPARSA and Active Set).\n']);
fprintf(['  UpdateH       : The method for updating the Hessian matrix', ...
        char(13), ...
        '                  (Exact Hessian, BFGS or limitted-memory BFGS).\n']);
fprintf('  LBfgsLength   : The length of the limitted memory BFGS list.\n');
fprintf(['  TerminateType : Type of termination conditions.', char(13), ...
         '                  (0 = default, 1 = norm of the search', ...
         ' direction, 2 - objective value change, 3 = 1 + 2).\n']);
fprintf( ['  isTrace       : Keep trace the information w.r.t.', ...
          ' the iterations.\n']);
fprintf( ['  TraceMode     : Trace mode (0 = basic, 1 = details,', ...
          ' 3 - iterative vector).\n']);
fprintf( ['  isGradNorm    : Compute the norm of the gradient vector', ...
          ' if required.\n']);
fprintf( '  isRestart     : Perform the restart for the outer loop.\n');
fprintf( ['  nRestartIters : Number of repeating iterations for', ...
          ' restart in the outer loop.\n']);
fprintf( ['  isSubRestart  : Perform the restart technique in fast', ...
          ' gradient algorithms of the inner loop.\n']);
fprintf( ['  nSubRsIters   : Number of iterations for repeating', ...
          ' restart in the inner loop.\n']);
fprintf( ['  nSubFxCheck   : Number of iterations to check the', ...
          ' objective values in subproblems.\n']);
fprintf( '  isBtkLS       : Applying back-tracking line-search.\n');
fprintf( '  lsMode        : Line-search mode: BTK or FWLK.\n');
fprintf( '  minStepSize   : The lower bound of the step size.\n');
fprintf(['  isSubBtkLS    : Applying backtracking line-search to find ', ...
         'a value of Lipschitz constant in the fast gradient methods.\n']);
fprintf(['  BtkMaxIters   : The maximum number of iterations of ', ...
         'back-tracking line-search in the inner loop.\n']);
fprintf(['  BtkMemIters   : The number of memorized iterations in ', ...
         'back-tracking line-search in the inner loop.\n']);
fprintf('  btkSlop       : Back-tracking slop.\n');
fprintf('  BBStepType    : Type of Barzilai-Borwein step (1, 2 or 3).\n');
fprintf(['  isEvalStrCvx  : Evaluate the strong convexity parameter', ...
         ' if required.\n']);
fprintf('  QuadCvRg      : The radius of quadratic convergence region.\n');
fprintf(['  bcdsSubTol    : The tolerance for the block-coordinate ', ...
         'descent method.\n']);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%