% FUNCTION: scopt_prints(param, varargin)
% PURPOSE:  Print the output information at each iteration.
%
% INFORMATION:
%    By Quoc Tran Dinh, Laboratory for Informations and Inference Systems,
%       (LIONS), EPFL, Lausanne, Switzerland.
%    Joint work with Volkan Cevher and Anastasios Kyrillidis.
%    Date: 03.06.2013.
%    Last modified: 03.06.2013.
%    Contact: quoc.trandinh@epfl.ch
% 
% DETAILS:
%   This procedure prints the output information at each iteration:
%    iter           : Iteration counter.
%    lambda         : The Proximal-Newton decrement.
%    alpha          : Step size.
%    prox-g/prox    : Number of prox-operations upto the current iter.
%    f(x) dec/ decf : The theoretical decrease of the objective value.
%    fval           : The objective value.
%    param          : The penalty parameter (in path-following framework).
%    error          : The relative error measuring the norm of the search 
%                     direction.
%    serror         : The relative error of the subproblems.
%    siter          : Number of iterations for solving subproblems.
%    stime          : Computational time for solving the subproblems.
%    beta           : The quantity ||d||_2 (norm of the search direction).
%    Lips           : The Lipschitz constant of the quadratic approximation
%                     in the subproblem.
%
%
function scopt_prints(param, varargin)

% Print the iteration in the proximal-Newton-type algorithms. 
if strcmpi(param.Algorithm(1:2), 'PN');
    
    % Print the iteration with short description.
    if param.PrintLevel == 1
        if mod(varargin{1}, 100) == 1
            fprintf([repmat('*', 1, 65 ), '\n']); 
            fprintf(['iter |   lambda |   alpha  |   serror |', ...
                     'prox-g|      fval    \n']);         
            fprintf([repmat('*', 1, 65 ), '\n']);      
        end

        if mod(varargin{1}, 10) == 1
            fprintf(['%5d| %3.3e| %3.3e| %3.3e|%6d| %3.12e\n'], ...
                    varargin{1}, varargin{3}, varargin{4}, varargin{7}, ...
                    varargin{6}, varargin{2});
        end
    end

    % Print the iteration in details.
    if param.PrintLevel >= 2 
        if mod(varargin{1}, 10) == 1
            fprintf([repmat('*', 1, 91 ), '\n']); 
            fprintf(['iter |  lambda  |   alpha  |', ...
                 '   decf   |  serror  |siter|  stime  | prox-g|     fval\n']);         
            fprintf([repmat('*', 1, 91 ), '\n']);      
        end

        fprintf(['%5d| %3.3e| %3.3e| %3.3e| %3.3e| %4d|', ...
                 ' %3.2e|  %5d| %3.10e \n'], ...
                 varargin{1}, varargin{3}, varargin{4}, varargin{5}, ...
                 varargin{7}, varargin{8}, varargin{9}, varargin{6}, ...
                 varargin{2});
    end
end


% Print the iteration in the path-following proximal-Newton algorithms. 
if strcmpi(param.Algorithm, 'PFPN');
    
    % Print the iteration with short description.
    if param.PrintLevel == 1
        if mod(varargin{1}, 100) == 1
            fprintf([repmat('*', 1, 75 ), '\n']); 
            fprintf(['iter |   lambda |   alpha  |   serror |  param  |', ...
                     'prox-g|      fval    \n']);         
            fprintf([repmat('*', 1, 75 ), '\n']);      
        end

        if mod(varargin{1}, 10) == 1
            fprintf(['%5d| %3.3e| %3.3e| %3.3e| %3.2e|%6d| %3.12e\n'], ...
                    varargin{1}, varargin{3}, varargin{4}, varargin{7}, ...
                    varargin{10}, varargin{6}, varargin{2});
        end
    end

    % Print the iteration in details.
    if param.PrintLevel >= 2 
        if mod(varargin{1}, 10) == 1
            fprintf([repmat('*', 1, 102 ), '\n']); 
            fprintf(['iter |  lambda  |   alpha  |   param  |', ...
                 '   decf   |  serror  |siter|  stime  | prox-g|     fval\n']);         
            fprintf([repmat('*', 1, 102 ), '\n']);      
        end

        fprintf(['%5d| %3.3e| %3.3e| %3.3e| %3.3e| %3.3e| %4d|', ...
                 ' %3.2e|  %5d| %3.10e \n'], ...
                 varargin{1}, varargin{3}, varargin{4}, varargin{10}, ...
                 varargin{5}, varargin{7}, varargin{8}, varargin{9}, ...
                 varargin{6}, varargin{2});
    end

end

% Print the iteration in the (fast) proximal gradient method.
if strcmpi(param.Algorithm, 'PG') || strcmpi(param.Algorithm, 'FPG')
    
    % Print information after each 50 iterations.
    if param.PrintLevel == 1
        if mod(varargin{1}, 500) == 1
            fprintf([repmat('*', 1, 42 ), '\n']); 
            fprintf('iter |   error  | prox |      fval\n');         
            fprintf([repmat('*', 1, 42 ), '\n']);      
        end

        if mod(varargin{1}, 50) == 1
            fprintf('%5d| %3.3e| %5d| %3.10e \n', ...
                     varargin{1}, varargin{3}, varargin{8}, varargin{2});
        end
    end

    % Print the iteration in details.
    if param.PrintLevel >= 2
        if mod(varargin{1}, 10) == 1
            fprintf([repmat('*', 1, 97 ), '\n']); 
            fprintf(['iter |   error  |  beta    |  lambda  |', ...
                     '  alpha   | f(x) desc|   Lips   | prox |', ...
                     '     fval    \n']);         
            fprintf([repmat('*', 1, 97 ), '\n']);      
        end
        fprintf(['%5d| %3.3e| %3.3e| %3.3e| %3.3e| %3.3e|', ...
                 ' %3.3e| %5d| %3.10e\n'], ...
                 varargin{1}, varargin{3}, varargin{5}, varargin{4}, ...
                 varargin{6}, varargin{7}, varargin{9}, varargin{8}, ...
                 varargin{2});
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% END OF THE IMPLEMENTATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%